/* Wisteria Daifugo Library
 defines.h
 */

#ifndef _DEFINES_H_
#define _DEFINES_H_

//ゲーム以前の定義とか
//時間計測もすぐに使えるようにここに入れておく
//設定ファイルの内容を使うので、
//設定ファイルはこれより先にインクルードする
//ただし設定ファイルのインクルード無しでもコンパイルエラーにならないようにする

//整数定義
typedef long long int int64;
typedef unsigned long long int uint64;
typedef int int32;
typedef unsigned int uint32;
typedef short int16;
typedef unsigned short uint16;
typedef signed char int8;
typedef unsigned char uint8;

typedef unsigned long ulong;


//クロック計測関数(64ビットで値を返す)
static uint64 cputime(void){
	unsigned int ax,dx;
	asm volatile("rdtsc\nmovl %%eax,%0\nmovl %%edx,%1":"=g"(ax),"=g"(dx): :"eax","edx");
	return ((uint64)(dx)<<32)+(uint64)(ax);
}

//タイムスタンプ読み込みにかかる時間
//環境依存しそうなので0にしておく
constexpr uint64 CLOCK_LATENCY=0ULL;

//デバッグ用

uint64 CLOCK_START;

static void tick(void){
	CLOCK_START=cputime();
}

static void tock(void){
	uint64 CLOCK_END=cputime();
	printf("clock=%lld\n",CLOCK_END-CLOCK_START-CLOCK_LATENCY);
}

timeval CLOCKMICS_START;

static void tickMicS(){
	gettimeofday(&CLOCKMICS_START,NULL);
}

static void tockMicS(){
	timeval CLOCKMICS_END;
	gettimeofday(&CLOCKMICS_END,NULL);
	long t=(CLOCKMICS_END.tv_sec-CLOCKMICS_START.tv_sec)*1000000+CLOCKMICS_END.tv_usec-CLOCKMICS_START.tv_usec;
	printf("%ld micsec.\n",t);
}

//腕時計クラス


class Clock{
private:
	uint64 c_start;
public:
	void start(){c_start=cputime();}
	uint64 stop(){return (cputime()-c_start-CLOCK_LATENCY);}
	uint64 restart(){//結果を返し、0から再スタート
		uint64 tmp=cputime();
		uint64 old=c_start;
		c_start=tmp;
		return (tmp-old-CLOCK_LATENCY);
	}
	
	constexpr Clock():
	c_start()
	{}
};

class ClockMicS{
	//microsec単位	
private:
	timeval t_start;
public:
	void start(){gettimeofday(&t_start,NULL);}
	long stop(){
		timeval t_end;
		gettimeofday(&t_end,NULL);
		long t=(t_end.tv_sec-t_start.tv_sec)*1000000+t_end.tv_usec-t_start.tv_usec;
		return t;
	}
	long restart(){//結果を返し、0から再スタート
		timeval t_end;
		gettimeofday(&t_end,NULL);
		long t=(t_end.tv_sec-t_start.tv_sec)*1000000+t_end.tv_usec-t_start.tv_usec;
		t_start=t_end;
		return t;
	}
	ClockMicS(){}
	ClockMicS(int m){start();}
};

class ClockMS{
	//millisec単位	
private:
	timeval t_start;
public:
	void start(){gettimeofday(&t_start,NULL);}
	long stop(){
		timeval t_end;
		gettimeofday(&t_end,NULL);
		long t=(t_end.tv_sec-t_start.tv_sec)*1000+(t_end.tv_usec-t_start.tv_usec)/1000;
		return t;
	}
	long restart(){//結果を返し、0から再スタート
		timeval t_end;
		gettimeofday(&t_end,NULL);
		long t=(t_end.tv_sec-t_start.tv_sec)*1000+(t_end.tv_usec-t_start.tv_usec)/1000;
		t_start=t_end;
		return t;
	}
	ClockMS(){}
	ClockMS(int m){start();}
};

//出力
struct Space{
	int n;
	constexpr Space():n(0){}
	constexpr Space(const int arg):n(arg){}
};

static std::ostream& operator<<(std::ostream& out,const Space& sp){
	for(int i=0;i<sp.n;++i){out<<" ";}
	return out;
}

template<typename T=int>
struct OutXY{
	T x,y;
	constexpr OutXY(T ax,T ay):x(ax),y(ay){}
};

template<typename T>
static std::ostream& operator<<(std::ostream& out,const OutXY<T>& arg){
	out<<"("<<arg.x<<","<<arg.y<<")";
	return out;
}

//3値定数
constexpr int _YES=1;
constexpr int _NO=0;
constexpr int _BOTH=2;

//constexprのmin,max
template<typename T>
constexpr T cmax(const T& a,const T& b){
	return a<b?b:a;
}

template<typename T>
constexpr T cmin(const T& a,const T& b){
	return a>b?b:a;
}

//出力等可能なアサーション

// アサーション用関数
void reportError( const char *functionName,
				 const char *fileName, int lineNumber,
				 std::string message );

//条件x、命令等y
#ifdef NDEBUG

#define ASSERT(X,Y)

#else

#define ASSERT(X,Y)  if(! (X) ) {\
Y\
;\
assert(0);}

#endif /* NDEBUG */

//標準ライブラリ使用
using std::cout;
using std::cerr;
using std::setw;
using std::endl;
using std::ostream;
using std::stringstream;
using std::max;
using std::min;

//出力
#ifdef DEBUG  
#define DOUT cout
#define DERR cerr
#define DWAIT getchar()
#else 
#define DERR 0 && cerr
#define DOUT 0 && cout
#define DWAIT
#endif

#ifndef FOR_TOURNAMENT
#define COUT cout
#define CERR cerr
#define CWAIT getchar()
#else
#define CERR 0 && cerr
#define COUT 0 && cout
#define CWAIT
#endif

#define STATIC_ASSERT(x) static_assert(x, "")

#if defined HAVE_BMI2
#include <immintrin.h>
#endif

#if defined (HAVE_SSE4)
#include <smmintrin.h>
#elif defined (HAVE_SSE2)
#include <emmintrin.h>
#endif

#if !defined(NDEBUG)
// デバッグ時は、ここへ到達してはいけないので、assert でプログラムを止める。
#define UNREACHABLE assert(false)
#elif defined(_MSC_VER) && !defined(__INTEL_COMPILER)
#define UNREACHABLE __assume(false)
#elif defined(__INTEL_COMPILER)
// todo: icc も __assume(false) で良いのか？ 一応ビルド出来るけど。
#define UNREACHABLE __assume(false)
#elif defined(__GNUC__) && (4 < __GNUC__ || (__GNUC__ == 4 && 4 < __GNUC_MINOR__))
#define UNREACHABLE __builtin_unreachable()
#else
#define UNREACHABLE assert(false)
#endif

#if defined(_MSC_VER) && !defined(__INTEL_COMPILER)
#define FORCE_INLINE __forceinline
#elif defined(__INTEL_COMPILER)
#define FORCE_INLINE inline
#elif defined(__GNUC__)
#define FORCE_INLINE __attribute__((always_inline)) inline
#else
#define FORCE_INLINE inline
#endif

// インラインアセンブリのコメントを使用することで、
// C++ コードのどの部分がアセンブラのどの部分に対応するかを
// 分り易くする。
#if defined(_MSC_VER) && !defined(__INTEL_COMPILER)
#define ASMCOMMENT(s)
#elif defined(__INTEL_COMPILER)
#define ASMCOMMENT(s)
#elif defined(__clang__)
#define ASMCOMMENT(s)
#elif defined(__GNUC__)
#define ASMCOMMENT(s) __asm__("#"s)
#else
#define ASMCOMMENT(s)
#endif

//出力ストリーム
/*struct StringStream{
	std::stringstream ss;
	void clear(){
		ss.str("");
		ss.clear(stringstream::goodbit);
	}
};*/


//環境依存にはまらないように色々

#ifndef M_PI
#define M_PI 3.14159265358979343846
#endif

//#ifdef (_MSC_VER)
//#define constexpr
//#endif


#endif  /* _DEFINES_H_ */