//ハッシュによる置換表
//用途により様々なタイプを用意しておくつもり

#include "../analyze/Analyzer.hpp"

struct TwoValuePage{
	uint32 data;
	
	uint32 any()const{return data;}
	void setHash(const uint64 hash){data=(uint32)(hash>>32);}
	void setResult(const uint32 res){
		assert( !(res&(~3U)) );
		data=(data&(~3U))|res;
	}
	uint32 cmpHash(const uint64 hash)const{return ((data^((uint32)(hash>>32)))&(~3U));}
	uint32 getResult()const{return (data&(3U));}

};

template<int SIZE>
class TwoValueBook{
	//2(+中間1)値を保存するハッシュ表
	//値1...1
	//値2...2
	//中間値...0
	//未登録...-1
private:

	HashBookAnalyzer ana;
	
	TwoValuePage page[SIZE];
	
	int convHash_Index(const uint64 hash){
		return ((uint32)(hash))%SIZE;
	}
public:
	
	void init(void){
		memset(page,0,sizeof(TwoValuePage)*SIZE);
	}
	
	TwoValueBook(){
		init();
	}
	
	TwoValueBook(const std::string& argName)
	:ana(argName,SIZE)
	{
		init();
	}
	
	~TwoValueBook(){}
	
	int read(const uint64 hash){
		const TwoValuePage& fpage=page[convHash_Index(hash)];
		if(fpage.any()){//結果が記入されている
			if(fpage.cmpHash(hash)){//Hash値が違う
				ana.addDiffHash();
				return -1;
			}else{
				ana.addHit();
				return fpage.getResult();
			}
		}
		ana.addWhite();
		return -1;
	}
	
	void regist(const int result,const uint64 hash){
		TwoValuePage& fpage=page[convHash_Index(hash)];
		if(!fpage.any()){
			ana.addFill();
		}
		fpage.setHash(hash);
		fpage.setResult(result);
		ana.addRegist();
	}
};

template<int N_BITS>
struct BitPage{
	//Nビットを保存する形式
	//64-Nビットをハッシュ値の確認として利用
	
	const static uint64 MASK=(1ULL<<N_BITS)-1ULL;
	
	uint64 data;
	
	uint64 any()const{return data;}
	void set(const uint64 arg){data=arg;}
	uint64 cmpHash(const uint64 hash)const{return (data^hash)&(~MASK);}
	uint64 get()const{return data;}
};

template<int N_BITS,int SIZE>
class BitBook{
	
private:

	HashBookAnalyzer ana;

	BitPage<N_BITS> page[SIZE];
	
	int convHash_Index(const uint64 hash){
		return (hash)%SIZE;
	}
public:
	
	void init(void){
		memset(page,0,sizeof(BitPage<N_BITS>)*SIZE);
	}
	
	BitBook(){
		init();
	}
	
	BitBook(const std::string& argName)
	:ana(argName,SIZE)
	{
		init();
	}
	
	~BitBook(){}
	
	uint64 read(const uint64 hash){
		const BitPage<N_BITS>& fpage=page[convHash_Index(hash)];
		if(fpage.any()){//結果が記入されている
			if(fpage.cmpHash(hash)){//Hash値が違う
				ana.addDiffHash();
				return 0ULL;
			}else{
				ana.addHit();
				return fpage.get() & (BitPage<N_BITS>::MASK);
			}
		}
		ana.addWhite();
		return 0ULL;
	}
	
	void regist(const uint64 bits,const uint64 hash){
		BitPage<N_BITS>& fpage=page[convHash_Index(hash)];
		if(!fpage.any()){
			ana.addFill();
		}
		fpage.set((hash&(~BitPage<N_BITS>::MASK)) | (bits& BitPage<N_BITS>::MASK));
		ana.addRegist();
	}
};