﻿/*
 Wisteria Daifugo Library 
 bitArray.hpp
 */

#ifndef _BITARRAY_HPP_
#define _BITARRAY_HPP_

#include<iostream>

#include "../defines.h"

#include"container.hpp"

//整数のNビットずつを使ってデータを保存する
//アクセスは配列より遅いだろうが、
//パラレルに処理出来るので速くなる部分もあると期待

template< int N=4,int ARG_SIZE=32/N >
struct BitArray32{
	
	uint32 dat;
	
	const static uint32 MASK_HORIZON_32=(1U<<N)-1U;
	const static int SIZE=cmax(0,cmin(ARG_SIZE,32/N));//指定サイズがおかしい時は勝手に調整
	
	constexpr BitArray32():dat(){}
	constexpr BitArray32(const uint32 arg):dat(arg){}
	constexpr BitArray32(const BitArray32& arg):dat(arg.dat){}
	
	//データをいじりやすいようにオペレータを定義
	constexpr operator uint32()const{return static_cast<uint32>(dat);}
	
	BitArray32& operator=(const uint32 arg){dat=arg;return *this;}
	BitArray32& operator+=(const uint32 arg){dat+=arg;return *this;}
	BitArray32& operator-=(const uint32 arg){dat-=arg;return *this;}
	BitArray32& operator*=(const uint32 arg){dat+=arg;return *this;}
	BitArray32& operator/=(const uint32 arg){dat-=arg;return *this;}
	BitArray32& operator|=(const uint32 arg){dat|=arg;return *this;}
	BitArray32& operator&=(const uint32 arg){dat&=arg;return *this;}
	BitArray32& operator^=(const uint32 arg){dat^=arg;return *this;}
	BitArray32& operator<<=(const uint32 arg){dat<<=arg;return *this;}
	BitArray32& operator>>=(const uint32 arg){dat>>=arg;return *this;}
	
	constexpr uint32 any()const{return dat;}
	
	template<typename size_t=int>
	constexpr uint32 operator[](const size_t n){//境界チェックなし
		return (dat>>getPlace(n))&MASK_HORIZON_32;
	}
	
	void clear(){
		dat=0U;
	}
	
	constexpr uint32 data()const{return dat;}
	
	template<typename size_t=int>
	constexpr size_t size()const{return static_cast<size_t>(SIZE);}
	
	constexpr uint32 getPlace(const int n)const{return (n*N);}
	
	constexpr uint32 getMask(const int n)const{return MASK_HORIZON_32<<getPlace(n);}
	
	constexpr uint32 getMask(const int n0,const int n1)const{
		return ( (1U<<getPlace(n1+1))-1U ) & (~( (1U<<getPlace(n0))-1U ));
	}
	
	constexpr uint32 getPart(const int n)const{return dat&getMask(n);}
	
	template<typename size_t=int>
	uint32& at(const size_t n)const{//境界チェックあり
		if(!(n>=0 && n<SIZE)){cout<<"BitArray32.at() failed."<<endl;exit(1);}
		return (dat>>getPlace(n))&MASK_HORIZON_32;
	}

	void set(const int n,const uint32 argDat){
		dat|=(argDat<<getPlace(n));
	}
	void assign(const int n,const uint32 argDat){
		assert(n>=0 && n<SIZE);
		const uint32 mask=getMask(n);
		
		assert(!(argDat & (~MASK_HORIZON_32)));
		
		dat=(dat&(~mask))|(argDat<<getPlace(n));
	}
	//他のビット配列の同じ位置をコピー
	template<int N_A,int SIZE_A>
	void setPart(const int n,const BitArray32<N_A,SIZE_A> arg){
		assert(n>=0 && n<SIZE);
		dat|=arg.getPart(n);
	}
	template<int N_A,int SIZE_A>
	void assignPart(const int n,const BitArray32<N_A,SIZE_A> arg){
		assert(n>=0 && n<SIZE);
		const uint32 mask=getMask(n);
		dat=(dat&(~mask))|(arg.data()&mask);
	}
	
	//ある基準位置が０となるように回転
	void rotate(const int n){
		assert(n>=0 && n<SIZE);
		//if(SIZE * N == 32){//サイズが一杯のとき
		//	data=ror32(getPlace(n));
		//}else{
		int place=getPlace(n);
		uint32 mask0=(1U<<place)-1U;
		uint32 mask1=(~mask0);
		dat=((dat&mask0)<<((N*SIZE)-place)) | ((dat&mask1)>>place);
		//}
	}
	
	//ある位置のものを抜いて、1つ詰める
	void remove(const int n){
		assert(n>=0 && n<SIZE);
		if( n==0 ){
			dat>>=N;
		}else if( n==SIZE-1 ){
			dat&=(~getMask(n));
		}else{
			uint32 mask=getMask(0,n-1);
			
			//cout<<"n:"<<n<<" mask:"<<mask;
			
			uint32 org=dat&mask;
			
			dat>>=N;
			dat=(dat&(~mask)) | org;
		}
	}
	
	//加減算
	void plus(const int n,const uint32 argDat){
		dat+=(argDat<<getPlace(n));
	}
	void minus(const int n,const uint32 argDat){
		dat-=(argDat<<getPlace(n));
	}
	
	//合計
	uint32 sum()const{
		uint32 tmp=dat;
		uint32 mask=getMask(0);
		uint32 res=tmp&mask;
		for(int i=1;i<SIZE;++i){
			tmp>>=N;
			res+=tmp&mask;
		}
		return res;
	}

};

template<int N,int SIZE>
ostream& operator<<(ostream& out,const BitArray32<N,SIZE>& arg){//出力
	out<<"{";
	if(arg.size() >= 1){
		out<<arg[0];
		for(auto i=1;i<arg.size();++i){
			out<<","<<arg[i];
		}
	}
	out<<"}";
	return out;
}
	
template<int N,int SIZE>
BitArray32<N,SIZE> splitHorizon(const BitArray32<N,SIZE>& c,const int n){//基準をずらす(不可逆)
	return BitArray32<N,SIZE>(c.data()>>c.getPlace(n));
}

template< int N=4,int ARG_SIZE=64/N >
struct BitArray64{
	
	uint64 dat;
	
	const static uint64 MASK_HORIZON_64=(1ULL<<N)-1ULL;
	const static int SIZE=cmax(0,cmin(ARG_SIZE,64/N));//指定サイズがおかしい時は勝手に調整
	
	constexpr BitArray64():dat(){}
	constexpr BitArray64(const uint64 arg):dat(arg){}
	constexpr BitArray64(const BitArray64& arg):dat(arg.dat){}
	
	//データをいじりやすいようにオペレータを定義
	constexpr operator uint64()const{return static_cast<uint64>(dat);}
	
	BitArray64& operator=(const uint64 arg){dat=arg;return *this;}
	BitArray64& operator+=(const uint64 arg){dat+=arg;return *this;}
	BitArray64& operator-=(const uint64 arg){dat-=arg;return *this;}
	BitArray64& operator*=(const uint64 arg){dat+=arg;return *this;}
	BitArray64& operator/=(const uint64 arg){dat-=arg;return *this;}
	BitArray64& operator|=(const uint64 arg){dat|=arg;return *this;}
	BitArray64& operator&=(const uint64 arg){dat&=arg;return *this;}
	BitArray64& operator^=(const uint64 arg){dat^=arg;return *this;}
	BitArray64& operator<<=(const uint64 arg){dat<<=arg;return *this;}
	BitArray64& operator>>=(const uint64 arg){dat>>=arg;return *this;}
	
	constexpr uint64 any()const{return dat;}
	
	template<typename size_t=int>
	constexpr uint64 operator[](const size_t n){//境界チェックなし
		return (dat>>getPlace(n))&MASK_HORIZON_64;
	}
	
	void clear(){
		dat=0ULL;
	}
	
	constexpr uint64 data()const{return dat;}
	
	template<typename size_t=int>
	constexpr size_t size()const{return static_cast<size_t>(SIZE);}
	
	constexpr uint64 getPlace(const int n)const{return (n*N);}
	
	constexpr uint64 getMask(const int n)const{return MASK_HORIZON_64<<getPlace(n);}
	
	constexpr uint32 getMask(const int n0,const int n1)const{
		return ( (1ULL<<getPlace(n1+1))-1ULL ) & (~( (1ULL<<getPlace(n0))-1ULL ));
	}
	
	constexpr uint64 getPart(const int n)const{return dat&getMask(n);}
	
	template<typename size_t=int>
	uint64& at(const size_t n)const{//境界チェックあり
		if(!(n>=0 && n<SIZE)){cout<<"BitArray64.at() failed."<<endl;exit(1);}
		return (dat>>getPlace(n))&MASK_HORIZON_64;
	}
	
	void set(const int n,const uint64 argDat){
		dat|=(argDat<<getPlace(n));
	}
	void assign(const int n,const uint64 argDat){
		assert(n>=0 && n<SIZE);
		const uint64 mask=getMask(n);
		
		assert(!(argDat & (~MASK_HORIZON_64)));
		
		dat=(dat&(~mask))|(argDat<<getPlace(n));
	}
	//他のビット配列の同じ位置をコピー
	template<int N_A,int SIZE_A>
	void setPart(const int n,const BitArray64<N_A,SIZE_A> arg){
		assert(n>=0 && n<SIZE);
		dat|=arg.getPart(n);
	}
	template<int N_A,int SIZE_A>
	void assignPart(const int n,const BitArray64<N_A,SIZE_A> arg){
		assert(n>=0 && n<SIZE);
		const uint64 mask=getMask(n);
		dat=(dat&(~mask))|(arg.data()&mask);
	}
	
	//ある基準位置が０となるように回転
	void rotate(const int n){
		assert(n>=0 && n<SIZE);
		//if(SIZE * N == 64){//サイズが一杯のとき
		//	data=ror64(getPlace(n));
		//}else{
		int place=getPlace(n);
		uint64 mask0=(1ULL<<place)-1ULL;
		uint64 mask1=(~mask0);
		dat=((dat&mask0)<<((N*SIZE)-place)) | ((dat&mask1)>>place);
		//}
	}
	
	//ある位置のものを抜いて、1つ詰める
	void remove(const int n){
		assert(n>=0 && n<SIZE);
		if( n==0 ){
			dat>>=N;
		}else if( n==SIZE-1 ){
			dat&=(~getMask(n));
		}else{
			uint64 mask=getMask(0,n-1);
			uint64 org=dat&mask;
			
			dat>>=N;
			dat=(dat&(~mask)) | org;
		}
	}
	
	//加減算
	void plus(const int n,const uint64 argDat){
		dat+=(argDat<<getPlace(n));
	}
	void minus(const int n,const uint64 argDat){
		dat-=(argDat<<getPlace(n));
	}
	
	//合計
	uint64 sum()const{
		uint64 tmp=dat;
		uint64 mask=getMask(0);
		uint64 res=tmp&mask;
		for(int i=1;i<SIZE;++i){
			tmp>>=N;
			res+=tmp&mask;
		}
		return res;
	}
	
};

template<int N,int SIZE>
ostream& operator<<(ostream& out,const BitArray64<N,SIZE>& arg){//出力
	out<<"{";
	if(arg.size() >= 1){
		out<<arg[0];
		for(auto i=1;i<arg.size();++i){
			out<<","<<arg[i];
		}
	}
	out<<"}";
	return out;
}

template<int N,int SIZE>
BitArray64<N,SIZE> splitHorizon(const BitArray64<N,SIZE>& c,const int n){//基準をずらす(不可逆)
	return BitArray64<N,SIZE>(c.data()>>c.getPlace(n));
}




	
#endif /* _BITARRAY_HPP_ */