﻿/*
 Wisteria Daifugo Library
 bitPartition.hpp
 */

#ifndef _BITPARTITION_HPP_
#define _BITPARTITION_HPP_

#include<cstdio>
#include<cassert>

#include"../defines.h"

#include"container.hpp"
#include"bitArray.hpp"

//ビット列に対するランダム操作

/**************************少数ビット取り出し**************************/
template<class dice64_t>
static uint64 pop1Bit64(uint64 *arg,dice64_t *const dice){
	uint64 tmp=*arg,tmp2;
	while(1){
		tmp2=tmp&dice->rand();
		if(tmp2!=0ULL){
			if(tmp2&(tmp2-1ULL)){//2つ以上
				tmp=tmp2;
			}else{
				break;
			}
		}
	}
	(*arg)-=tmp2;
	return tmp2;
}

template<class dice64_t>
static uint64 pick1Bit64(uint64 arg,dice64_t *const dice){
	uint64 tmp;
	while(1){
		tmp=arg&dice->rand();
		if(tmp!=0ULL){
			if(tmp&(tmp-1ULL)){//2つ以上
				arg=tmp;
			}else{
				break;
			}
		}
	}
	return tmp;
}

/**************************2分割**************************/

template<class dice64_t>
static uint64 pickNBits64(uint64 arg,const int argN0,const int argN1,dice64_t *const dice){
	//argからランダムにN0ビット抽出する
	//最初はN0+N1ビットある必要あり
	assert( (int)countBits(arg) == argN0+argN1 );

	uint64 res=0ULL;
	int N0=argN0;
	int N1=argN1;
	
	if( N0 < N1 ){
		if(N0==0){return res;}
		if(N0==1){return pick1Bit64(arg,dice);}
	}else{
		if(N1==0){return arg;}
		if(N1==1){return arg-pick1Bit64(arg,dice);}
	}
	
	int NDist;
	uint64 dist;
	
	while(1){
		dist=arg&dice->rand();
		NDist=countBits64(dist);
		
		//まずは一致チェック
		if(NDist == N0){
			res|=dist;
			assert( (int)countBits(res) == argN0 );
			return res;
		}
		if(NDist == N1){
			res|=(arg-dist);
			assert( (int)countBits(res) == argN0 );
			return res;
		}
		
		if(NDist < N0){
			if(NDist < N1){
				//NDistが小さく、鋭い
				if(N0 > N1){
					res|=dist;
					N0-=NDist;
				}else{
					N1-=NDist;
				}
				arg-=dist;
			}else{
				//鈍い
				int NRDist=N0+N1-NDist;
				if(NDist > NRDist){
					//NDistが大きい
					N0-=NDist;
					res|=dist;
					arg-=dist;
				}else{
					//NRDistが大きい
					N0-=NRDist;
					res|=(arg-dist);
					arg=dist;
				}
			}
		}else{
			if(NDist < N1){
				//鈍い
				int NRDist=N0+N1-NDist;
				if(NDist > NRDist){
					//NDistが大きい
					N1-=NDist;
					arg-=dist;
				}else{
					//NRDistが大きい
					N1-=NRDist;
					arg=dist;
				}
			}else{
				//NDistが大きく、鋭い
				if(N0 > N1){
					res|=(arg-dist);
					N0=NDist-N1;
				}else{
					N1=NDist-N0;
				}
				arg=dist;
			}
		}
		assert( (int)countBits(arg) == N0+N1 );
	}
	assert(0);
}

template<class dice64_t>
static void dist2_64(uint64 *goal0,uint64 *goal1,uint64 arg,int N0,int N1,dice64_t *const dice){
	assert( (int)countBits64(arg) == N0+N1 );
	
	uint64 tmp=pickNBits64(arg,N0,N1,dice);
	*goal0|=tmp;
	*goal1|=(arg-tmp);
}

/**************************n分割**************************/

template<class dice64_t>
static void dist3_64(uint64 *goal0,uint64 *goal1,uint64 *goal2,uint64 arg,int N0,int N1,int N2,dice64_t *const dice){
	uint64 goal1_2=0ULL;
	dist2_64(&goal1_2,goal0,arg,N1+N2,N0,dice);
	dist2_64(goal1,goal2,goal1_2,N1,N2,dice);
}

template<class dice64_t>
static void dist4_64(uint64 *goal0,uint64 *goal1,uint64 *goal2,uint64 *goal3,uint64 arg,int N0,int N1,int N2,int N3,dice64_t *const dice){
	//std::cout<<arg;
	uint64 goal0_1=0ULL,goal2_3=0ULL;
	dist2_64(&goal0_1,&goal2_3,arg,N0+N1,N2+N3,dice);
	dist2_64(goal0,goal1,goal0_1,N0,N1,dice);
	dist2_64(goal2,goal3,goal2_3,N2,N3,dice);
}

template<int N=2,class size_t,class dice64_t>
static void dist64(uint64 *const dst,const uint64 arg,const size_t *const argNum,dice64_t *const dice){
	//ビット集合をN分割
	//N<=0には未対応
	//DOUT<<"pointer mode. "<<N<<endl;
	if(N>0){
		switch(N){
			case 1:
				dst[0]|=arg;
				break;
			case 2:
				dist2_64(&dst[0],&dst[1],arg,argNum[0],argNum[1],dice);
				break;
			case 3:
				dist3_64(&dst[0],&dst[1],&dst[2],arg,argNum[0],argNum[1],argNum[2],dice);
				break;
			default:
				const int NH=N/2;
				
				int num[2]={0};
				int i=0;
				for(;i<NH;++i){
					num[0]+=argNum[i];
				}
				for(;i<N;++i){
					num[1]+=argNum[i];
				}
				
				assert( num[0]+num[1] == (int)countBits(arg) );
				
				uint64 half[2]={0ULL};
				dist64<2>(half,arg,num,dice);
				dist64<NH>(dst,half[0],argNum,dice);
				dist64<N-NH>(dst+NH,half[1],splitHorizon(argNum,NH),dice);
				break;
		}
	}
}

template<int N=2,int SN,int SSIZE,class dice64_t>
static void dist64(uint64 *const dst,const uint64 arg,const BitArray32<SN,SSIZE>& argNum,dice64_t *const dice){
	//ビット集合をN分割
	//N<=0には未対応
	//DOUT<<"reference mode. "<<N<<endl;
	if(N>0){
		switch(N){
			case 1:
				dst[0]|=arg;
				break;
			case 2:
				dist2_64(&dst[0],&dst[1],arg,argNum[0],argNum[1],dice);
				break;
			case 3:
				dist3_64(&dst[0],&dst[1],&dst[2],arg,argNum[0],argNum[1],argNum[2],dice);
				break;
			default:	
				const int NH=N/2;
				
				int num[2]={0};
				int i=0;
				for(;i<NH;++i){
					num[0]+=argNum[i];
				}
				for(;i<N;++i){
					num[1]+=argNum[i];
				}
				
				assert( num[0]+num[1] == (int)countBits(arg) );
				
				uint64 half[2]={0ULL};
				dist64<2>(half,arg,num,dice);
				dist64<NH>(dst,half[0],argNum,dice);
				dist64<N-NH>(dst+NH,half[1],splitHorizon(argNum,NH),dice);//BitArray32<SN,SSIZE>(argNum>>(NH*SN)));
				break;
		}
	}
}

#endif  /* _BITPARTITION_HPP_ */
