﻿/*
 Wisteria Daifugo Library
 bitSet.hpp
 */

#ifndef _BITSET_HPP_
#define _BITSET_HPP_

//ビットセットを定義
//サイズ固定
//std::bitsetより色々機能が多い

#include<iostream>

#include"../defines.h"

#include"container.hpp"

struct BitSet32{
	
	uint32 dat;
	
	constexpr BitSet32():dat(){}
	constexpr BitSet32(const uint32 arg):dat(arg){}
	constexpr BitSet32(const uint64 arg):dat((uint32)arg){}
	constexpr BitSet32(const BitSet32& arg):dat(arg.dat){}
	
	//データをいじりやすいようにオペレータを定義
	constexpr operator uint32()const{return static_cast<uint32>(dat);}
	
	//uint32 operator ~()const{return ~dat;}
	
	BitSet32& operator=(const uint32 arg){dat=arg;return *this;}
	BitSet32& operator+=(const uint32 arg){dat+=arg;return *this;}
	BitSet32& operator-=(const uint32 arg){dat-=arg;return *this;}
	BitSet32& operator*=(const uint32 arg){dat+=arg;return *this;}
	BitSet32& operator/=(const uint32 arg){dat-=arg;return *this;}
	BitSet32& operator|=(const uint32 arg){dat|=arg;return *this;}
	BitSet32& operator&=(const uint32 arg){dat&=arg;return *this;}
	BitSet32& operator^=(const uint32 arg){dat^=arg;return *this;}
	BitSet32& operator<<=(const uint32 arg){dat<<=arg;return *this;}
	BitSet32& operator>>=(const uint32 arg){dat>>=arg;return *this;}

	void set(int p){
		assert( 0<=p && p<32 );
		dat|=(1U<<p);
	}
	void flip(int p){
		assert( 0<=p && p<32 );
		dat^=(1U<<p);
	}
	
	void push0(int p){
		assert( 0<=p && p<32 );
		uint32 mask=(1U<<p)-1U;
		dat=(dat&mask) | ((dat&(~mask))<<1);
	}
	
	void push1(int p){
		assert( 0<=p && p<32 );
		uint32 bit=(1U<<p);
		uint32 mask=bit-1U;
		dat=(dat&mask) | ((dat&(~mask))<<1) | bit;
	}
	
	void reset(int p){
		assert( 0<=p && p<32 );
		dat&=~(1U<<p);
	}
	
	void reset(){dat=0U;}
	void flip(){dat=~dat;}
	
	uint32 test(int p)const{
		assert( 0<=p && p<32 );
		return dat & (1U<<p);
	}
	uint32 get(int p)const{
		assert( 0<=p && p<32 );
		return (dat >> p) & 1U;
	}
	
	bool holds(const BitSet32& arg)const{
		return holdsBits<uint32>(dat,arg.dat);
	}
	
	constexpr uint32 any()const{return dat;}
	int count()const{return countBits32(dat);}
	int bsf()const{return bsf32(dat);}
	int bsr()const{return bsr32(dat);}
	
	void pop_lsb(){ dat&=(dat-1U); }
	
	
	void init(){dat=0U;}
	
	constexpr uint32 data()const{return dat;}
};
/*
uint32 operator&(const BitSet32& a,const uint32 b){
	return a.data()&b;
}
uint32 operator|(const BitSet32& a,const uint32 b){
	return a.data()|b;
}
uint32 operator^(const BitSet32& a,const uint32 b){
	return a.data()^b;
}
uint32 operator+(const BitSet32& a,const uint32 b){
	return a.data()+b;
}
uint32 operator-(const BitSet32& a,const uint32 b){
	return a.data()-b;
}
uint32 operator*(const BitSet32& a,const uint32 b){
	return a.data()*b;
}
uint32 operator/(const BitSet32& a,const uint32 b){
	return a.data()/b;
}
uint32 operator<<(const BitSet32& a,const uint32 b){
	return a.data()<<b;
}
uint32 operator>>(const BitSet32& a,const uint32 b){
	return a.data()>>b;
}*/

struct BitSet64{
	
	uint64 dat;
	
	constexpr BitSet64():dat(){}
	//constexpr BitSet64(const uint32 arg):dat((uint64)arg){}
	constexpr BitSet64(const uint64 arg):dat(arg){}
	constexpr BitSet64(const BitSet64& arg):dat(arg.dat){}
	
	//データをいじりやすいようにオペレータを定義
	constexpr operator uint64()const{return static_cast<uint64>(dat);}
	
	constexpr uint32 toU32()const{return static_cast<uint32>(dat);}
	
	//uint32 operator ~()const{return ~dat;}
	
	BitSet64& operator=(const uint64 arg){dat=arg;return *this;}
	BitSet64& operator+=(const uint64 arg){dat+=arg;return *this;}
	BitSet64& operator-=(const uint64 arg){dat-=arg;return *this;}
	BitSet64& operator*=(const uint64 arg){dat+=arg;return *this;}
	BitSet64& operator/=(const uint64 arg){dat-=arg;return *this;}
	BitSet64& operator|=(const uint64 arg){dat|=arg;return *this;}
	BitSet64& operator&=(const uint64 arg){dat&=arg;return *this;}
	BitSet64& operator^=(const uint64 arg){dat^=arg;return *this;}
	BitSet64& operator<<=(const uint64 arg){dat<<=arg;return *this;}
	BitSet64& operator>>=(const uint64 arg){dat>>=arg;return *this;}
	
	constexpr uint64 any()const{return dat;}
	int64 count()const{return countBits64(dat);}
	
	void set(int p){
		assert( 0<=p && p<64 );
		dat|=(1ULL<<p);
	}
	void flip(int p){
		assert( 0<=p && p<64 );
		dat^=(1ULL<<p);
	}
	void push0(int p){
		assert( 0<=p && p<64 );
		uint32 mask=(1ULL<<p)-1ULL;
		dat=(dat&mask) | ((dat&(~mask))<<1);
	}
	void push1(int p){
		assert( 0<=p && p<64 );
		uint32 bit=(1ULL<<p);
		uint32 mask=bit-1ULL;
		dat=(dat&mask) | ((dat&(~mask))<<1) | bit;
	}
	void reset(int p){
		assert( 0<=p && p<64 );
		dat&=~(1ULL<<p);
	}
	
	void reset(){dat=0ULL;}
	void flip(){dat=~dat;}
	
	void set(int p0,int p1){
		dat|=(1ULL<<p0)|(1ULL<<p1);
	}
	
	void set(int p0,int p1,int p2){
		dat|=(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2);
	}
	
	void set(int p0,int p1,int p2,int p3){
		dat|=(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2)|(1ULL<<p3);
	}
	
	void set(int p0,int p1,int p2,int p3,int p4){
		dat|=(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2)|(1ULL<<p3)|(1ULL<<p4);
	}
	
	uint64 test(int p)const{
		assert( 0<=p && p<64 );
		return dat & (1ULL<<p);
	}
	uint64 get(int p)const{
		assert( 0<=p && p<64 );
		return (dat >> p) & 1ULL;
	}
	
	bool holds(int p0)const{
		return holdsBits<uint64>(dat,(1ULL<<p0));}
	bool holds(int p0,int p1)const{
		return holdsBits<uint64>(dat,(1ULL<<p0)|(1ULL<<p1));}
	bool holds(int p0,int p1,int p2)const{
		return holdsBits<uint64>(dat,(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2));}
	bool holds(int p0,int p1,int p2,int p3)const{
		return holdsBits<uint64>(dat,(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2)|(1ULL<<p3));}
	bool holds(int p0,int p1,int p2,int p3,int p4)const{
		return holdsBits<uint64>(dat,(1ULL<<p0)|(1ULL<<p1)|(1ULL<<p2)|(1ULL<<p3)|(1ULL<<p4));
	}
	
	int bsf()const{return bsf64(dat);}
	int bsr()const{return bsr64(dat);}
	
	void pop_lsb(){ dat&=(dat-1ULL); }
	
	
	void init(){dat=0ULL;}
	
	
	constexpr uint64 data()const{return dat;}
};

#endif /* _BITSET_HPP_ */