#include <cmath>

#if defined(__GNUC__) && ( defined(__i386__) || defined(__x86_64__) ) && !defined(_WIN32)
#include <quadmath.h>
#endif

//数学的定数
const unsigned int FactorialTable[12]={
	1,1,2,6,24,120,720,5040,40320,362880,3628800,39916800
};

double dFactorial(int n){
	assert( n >= 0 );
	
	double ans;
	if( n < 12 ){
		ans=(double)FactorialTable[n];
	}else{
		ans=1.0;
		while(n>1){
			ans*=(double)n;
			n--;
		}
	}
	return ans;
}

double dPermutation(int n,int r){
	double ans=1.0;
	while(r--){
		ans*=(n-r);
	}
	return ans;
}

double dCombination(int n,int r){
	if(n>=r){
		return dPermutation(n,r)/dFactorial(r);
	}else{
		return 0;
	}
}

double dCombination4(int n,int r1,int r2,int r3,int r4){
	if(n>=(r1+r2+r3+r4)){
		return dFactorial(n)/dFactorial(r1)/dFactorial(r2)/dFactorial(r3)/dFactorial(r4);
	}else{
		return 0;
	}
}

double dCombination4(int r1,int r2,int r3,int r4){
	if(r1+r2+r3+r4>=0){
		return dFactorial(r1+r2+r3+r4)/dFactorial(r1)/dFactorial(r2)/dFactorial(r3)/dFactorial(r4);
	}else{
		return 0;
	}
}

double dCombination5(int n,int r1,int r2,int r3,int r4,int r5){
	if(n>=(r1+r2+r3+r4+r5)){
		return dFactorial(n)/dFactorial(r1)/dFactorial(r2)/dFactorial(r3)/dFactorial(r4)/dFactorial(r5);
	}else{
		return 0;
	}
}

int nC2(const int n){
	int res=1;
	int nn=n;
	while(nn>2){
		res*=nn;
		--nn;
	}
	nn=n-2;
	while(nn){
		res/=nn;
		--nn;
	}
	return res;
}

constexpr int ipow(int m,int n){
	return (n==0)?1:(m*ipow(m,n-1));
}

/**************************座標変換**************************/

struct dXY{double x,y;};
struct dRT{double r,t;};

dRT XYtoRT(double x,double y){
	dRT ret;
	ret.r=std::sqrt(x*x + y*y);
	assert( x*y != 0 );
	ret.r=atan2(y,x);
	return ret;
}

dXY RTtoXY(double r,double t){
	dXY ret;
	ret.x=r*std::cos(t);
	ret.y=r*std::sin(t);
	return ret;
}

constexpr uint32 XYtoR2(uint32 x,uint32 y){
	return x*x+y*y;
}

constexpr float XYtoR2(float x,float y){
	return x*x+y*y;
}
float XYtoR(float x,float y){
	return hypotf(x,y);
}
constexpr float XYtoT(float x,float y){
	return atan2f(y,x);
}
float XYtoCosT(float x,float y){
	assert( x!=0 || y!=0 );
	return x / XYtoR(x,y);
}
float XYtoSinT(float x,float y){
	assert( x!=0 || y!=0 );
	return y / XYtoR(x,y);
}

constexpr double XYtoR2(double x,double y){
	return x*x+y*y;
}
constexpr double XYtoR(double x,double y){
	return hypot(x,y);
}
constexpr double XYtoT(double x,double y){
	return atan2(y,x);
}
double XYtoCosT(double x,double y){
	assert( x!=0 || y!=0 );
	return x / XYtoR(x,y);
}
double XYtoSinT(double x,double y){
	assert( x!=0 || y!=0 );
	return y / XYtoR(x,y);
}

constexpr long double XYtoR2(long double x,long double y){
	return x*x+y*y;
}
constexpr long double XYtoR(long double x,long double y){
	return hypotl(x,y);
}
constexpr long double XYtoT(long double x,long double y){
	return atan2l(y,x);
}
long double XYtoCosT(long double x,long double y){
	assert( x!=0 || y!=0 );
	return x / XYtoR(x,y);
}
long double XYtoSinT(long double x,long double y){
	assert( x!=0 || y!=0 );
	return y / XYtoR(x,y);
}

#if defined(__GNUC__) && ( defined(__i386__) || defined(__x86_64__) ) && !defined(_WIN32)
constexpr __float128 XYtoR2(__float128 x,__float128 y){
	return x*x+y*y;
}
__float128 XYtoR(__float128 x,__float128 y){
	return hypotq(x,y);
}
__float128 XYtoT(__float128 x,__float128 y){
	return atan2q(y,x);
}
__float128 XYtoCosT(__float128 x,__float128 y){
	assert( x!=0 || y!=0 );
	return x / XYtoR(x,y);
}
__float128 XYtoSinT(__float128 x,__float128 y){
	assert( x!=0 || y!=0 );
	return y / XYtoR(x,y);
}
#endif

/************************** trigonometric function **************************/

template<typename T>
T sinAaddB( T sinA, T cosA, T sinB, T cosB ){
	return sinA * cosB + cosA * sinB;
}

template<typename T>
T sinAsubB( T sinA, T cosA, T sinB, T cosB ){
	return sinA * cosB - cosA * sinB;
}

template<typename T>
T cosAaddB( T sinA, T cosA, T sinB, T cosB ){
	return cosA * cosB - sinA * sinB;
}

template<typename T>
T cosAsubB( T sinA, T cosA, T sinB, T cosB ){
	return cosA * cosB + sinA * sinB;
}

/************************** rotation **************************/
template<class xy0_t,class xy1_t>
void rotate(const xy0_t& arg,const double theta,xy1_t *const dst){
	dst->setX( arg.getX() * cos(theta) - arg.getY() * sin(theta) );
	dst->setY( arg.getY() * cos(theta) + arg.getX() * sin(theta) );
}

template<class xy0_t,class xy1_t>
void rotate(const xy0_t& arg,const float theta,xy1_t *const dst){
	dst->setX( arg.getX() * cosf(theta) - arg.getY() * sinf(theta) );
	dst->setY( arg.getY() * cosf(theta) + arg.getX() * sinf(theta) );
}

template<class xy0_t>
void rotate(xy0_t *const dst,const double theta){
	const double ax=dst->getX();
	const double ay=dst->getY();
	dst->setX( ax * cos(theta) - ay * sin(theta) );
	dst->setY( ay * cos(theta) + ax * sin(theta) );
}

template<class xy0_t>
void rotate(xy0_t *const dst,const float theta){
	const float ax=dst->getX();
	const float ay=dst->getY();
	dst->setX( ax * cos(theta) - ay * sin(theta) );
	dst->setY( ay * cos(theta) + ax * sin(theta) );
}

template<class xy0_t,class xy1_t>
void rotateToAdd(const xy0_t& arg,const double theta,xy1_t *const dst){
	dst->addX( arg.getX() * cos(theta) - arg.getY() * sin(theta) );
	dst->addY( arg.getY() * cos(theta) + arg.getX() * sin(theta) );
}

template<class xy0_t,class xy1_t>
void rotateToAdd(const xy0_t& arg,const float theta,xy1_t *const dst){
	dst->addX( arg.getX() * cosf(theta) - arg.getY() * sinf(theta) );
	dst->addY( arg.getY() * cosf(theta) + arg.getX() * sinf(theta) );
}

void rotate(const float ox,const float oy,const float theta,float *const x,float *const y){
	*x= ox * cosf(theta) - oy * sinf(theta);
	*y= oy * cosf(theta) + ox * sinf(theta);
}

void rotateToAdd(const float ox,const float oy,const float theta,float *const x,float *const y){
	*x+= ox * cosf(theta) - oy * sinf(theta);
	*y+= oy * cosf(theta) + ox * sinf(theta);
}

void rotate(const double ox,const double oy,const double theta,double *const x,double *const y){
	*x= ox * cos(theta) - oy * sin(theta);
	*y= oy * cos(theta) + ox * sin(theta);
}

void rotateToAdd(const double ox,const double oy,const double theta,double *const x,double *const y){
	*x+= ox * cos(theta) - oy * sin(theta);
	*y+= oy * cos(theta) + ox * sin(theta);
}

void rotate(const long double ox,const long double oy,const long double theta,long double *const x,long double *const y){
	*x= ox * cosl(theta) - oy * sinl(theta);
	*y= oy * cosl(theta) + ox * sinl(theta);
}

void rotateToAdd(const long double ox,const long double oy,const long double theta,long double *const x,long double *const y){
	*x+= ox * cosl(theta) - oy * sinl(theta);
	*y+= oy * cosl(theta) + ox * sinl(theta);
}

#if defined(__GNUC__) && ( defined(__i386__) || defined(__x86_64__) ) && !defined(_WIN32)
void rotate(const __float128 ox,const __float128 oy,const __float128 theta,__float128 *const x,__float128 *const y){
	*x= ox * cosq(theta) - oy * sinq(theta);
	*y= oy * cosq(theta) + ox * sinq(theta);
}

void rotateToAdd(const __float128 ox,const __float128 oy,const __float128 theta,__float128 *const x,__float128 *const y){
	*x+= ox * cosq(theta) - oy * sinq(theta);
	*y+= oy * cosq(theta) + ox * sinq(theta);
}
#endif


/************************** rotation by trigonometric function **************************/

template<typename T>
void rotate(const T ox,const T oy,const T sinA, const T cosA, T *const x,T *const y){
	*x= ox * cosA - oy * sinA;
	*y= oy * cosA + ox * sinA;
}
template<typename T>
void rotateToAdd(const T ox,const T oy,const T sinA, const T cosA, T *const x,T *const y){
	*x+= ox * cosA - oy * sinA;
	*y+= oy * cosA + ox * sinA;
}








