//デジタルカーリングクライアント
//歩-Ayumu-
//Katsuki Ohto

namespace DigitalCurling{
	
	struct ThreadTools{
		//各スレッドの持ち物
		typedef XorShift64 dice64_t;
		typedef DSFMT ddice_t;
		
		//サイコロ
		dice64_t dice;
		ddice_t ddice;
		
		ThreadTools()
		{}
	};
	
	ThreadTools thread_tools[ N_THREADS ];
	
}

#include "mc/thinkboard.hpp"
#include "mc/leaf.hpp"
#include "mc/root.hpp"
#include "mc/mcplayer.hpp"

using namespace DigitalCurling::Ayumu;

namespace DigitalCurling{
	
	class Client{
	private:
		XorShift64 dice;
		DSFMT ddice;
	public:
		void initAll(){
			//各スレッドの持ち物を初期化
			dice.srand((uint32)time(NULL));
			ddice.srand((uint32)dice.rand());
			for( int th=0; th<N_THREADS; ++th ){
				thread_tools[th].dice.srand( dice.rand() * (th+111) );
				thread_tools[th].ddice.srand( dice.rand() * (th+111) );
			}
			
			initRandToMove(&ddice);
			
#ifndef FOR_TOURNAMENT 
			using namespace Simulator;
			using namespace FastSimulator;
			
			//ステップ <-> 逆ステップ
			{
				CERR<<"Simulator::step() <-> Simulator::invStep() check"<<endl;
				fMStone<> mst( 2.3, 4.9, -0.8, 1.7, +FANGV_ORG );
				CERR<<mst.x<<" , "<<mst.y<<" , "<<mst.vx<<" , "<<mst.vy<<" , "<<mst.w;
				stepSolo( &mst, 0.0001 );
				CERR<<" -> ";
				CERR<<mst.x<<" , "<<mst.y<<" , "<<mst.vx<<" , "<<mst.vy<<" , "<<mst.w;
				CERR<<endl;
				CERR<<mst.x<<" , "<<mst.y<<" , "<<mst.vx<<" , "<<mst.vy<<" , "<<mst.w;
				invStepSolo( &mst, 0.0001 );
				CERR<<" -> ";
				CERR<<mst.x<<" , "<<mst.y<<" , "<<mst.vx<<" , "<<mst.vy<<" , "<<mst.w;
				CERR<<endl;
			}
			//V_R変換
			{
				CERR<<"FV -> FR check"<<endl;
				fMStone<> org,dst;
				for(int i=0;i<4;++i){
					double v=ddice.drand() * FV_MAX;
					CERR<<" array : "<<v<<" -> "<<FVtoFR(v)<<" real : "<<v<<" -> ";
					org.set( 0, 0, 0, v, FANGV_ORG );
					Simulator::simulateSolo( org, &dst, 0.0001 );
					CERR<<XYtoR(dst.x,dst.y)<<endl;
				}
			}
			//R_V変換
			{
				CERR<<"FR -> FV check"<<endl;
				fMStone<> org,dst;
				for(int i=0;i<4;++i){
					double r=ddice.drand() * 49;
					CERR<<" array : "<<r<<" <- "<<FRtoFV(r)<<" real : ";
					org.set( 0, 0, 0, FRtoFV(r), FANGV_ORG );
					Simulator::simulateSolo( org, &dst, 0.0001 );
					CERR<<XYtoR(dst.x,dst.y)<<" <- "<<FRtoFV(r)<<endl;
				}
			}
			//ALL 変換
			{
				CERR<<"FALL check"<<endl;
				FMoment aorg;
				fMStone<> sorg,sdst;
				for(int i=0;i<4;++i){
					double v=ddice.drand() * FV_MAX;
					double theta=ddice.drand() * 2*M_PI;
					double w=dice.rand()%2 ? FANGV_ORG : (-FANGV_ORG);
					sorg.set( 0, 0, v*sin(theta), v*cos(theta), w );
					aorg.set( 0, 0, theta, 0, v , w );
					stepToStop( FVtoID(v), &aorg );//高速シミュレート
					simulateSolo( sorg, &sdst, 0.0001 );//通常シミュレート
					CERR<<" array : ";
					CERR<<aorg.x<<","<<aorg.y<<","<<aorg.v*sin(aorg.theta)<<","<<aorg.v*cos(aorg.theta)<<","<<aorg.t<<endl;
					CERR<<" real : ";
					CERR<<sdst.x<<","<<sdst.y<<","<<sdst.vx<<","<<sdst.vy<<","<<sdst.w<<endl;
				}
			}
			
			//いろいろやりたいことをやる
			fPosXY<> pos;
			fMoveXY<> vec;
			fRMoveXY<> rvec;
			
			/*
			//CERR<<"corner move"<<endl;
			pos.set(FX_RINK_MIN,FY_RINK_MIN);
			vec.setSpin(0);
			calcDeparture(pos,&vec);
			CERR<<"RINK LU 0 ("<<vec.x<<","<<vec.y<<")"<<endl;
			pos.set(FX_PA_MAX,FY_PA_MAX);
			vec.setSpin(0);
			calcDeparture(pos,&vec);
			CERR<<"PA RD 0 ("<<vec.x<<","<<vec.y<<")"<<endl;
			pos.set(FX_RINK_MIN,FY_RINK_MIN);
			vec.setSpin(0);
			CreateShot(pos,16,&vec);
			CERR<<"RINK LU 0 PowMAX("<<vec.x<<","<<vec.y<<") "<<XYtoR(vec.x,vec.y)<<endl;
			pos.set(FX_PA_MAX,FY_PA_MAX);
			vec.setSpin(0);
			CreateShot(pos,16,&vec);
			CERR<<"PA RD 0 PowMAX("<<vec.x<<","<<vec.y<<") "<<XYtoR(vec.x,vec.y)<<endl;
			pos.set(FX_RINK_MIN,FY_RINK_MIN);
			vec.setSpin(0);
			CreateShot(pos,0,&vec);
			CERR<<"RINK LU 0 PowMIN("<<vec.x<<","<<vec.y<<") "<<XYtoR(vec.x,vec.y)<<endl;
			pos.set(FX_PA_MAX,FY_PA_MAX);
			vec.setSpin(0);
			CreateShot(pos,0,&vec);
			CERR<<"PA RD 0 PowMIN("<<vec.x<<","<<vec.y<<") "<<XYtoR(vec.x,vec.y)<<endl;*/
			
			CERR<<"size of b2Vec2 ... "<<sizeof(b2Vec2)<<endl;
			CERR<<"size of b2Body ... "<<sizeof(b2Body)<<endl;
			CERR<<"size of b2World ... "<<sizeof(b2World)<<endl;
			
			CERR<<atan2(0,0)<<endl;
			
			//逆関数のはたらき調査
			
			/*fPosXY<> v(2.43,54);
			fMoveXY<> v2;
			DIRtoVEC( v, &v2 );
			CERR<<"DIRtoVEC"<<OutXY<float>(v.x,v.y)<<" -> "<<OutXY<float>(v2.x,v2.y)<<endl;
			VECtoDIR( v2, &v );
			CERR<<"VECtoDIR"<<OutXY<float>(v2.x,v2.y)<<" -> "<<OutXY<float>(v.x,v.y)<<endl;
			*/
			pos.set(FX_TEE,FY_TEE);
			vec.set(1,2,0);
			
			calcDeparture( pos, &vec );
			CERR<<"calcDeparture"<<pos<<" -> "<<vec<<endl;
			calcDestination( vec, &pos );
			CERR<<"calcDestination"<<vec<<" -> "<<pos<<endl;
			
			//着手作成の精度調査
			ThinkBoard bd;
			int suc[16][8];
			const int N=200;
			
			//ティーショット
			suc[0][0]=0;
			for(int i=0;i<N;++i){
				bd.init();
				vec.setSpin(dice.rand()%2);
				genTeeShot(&vec);
				bd.makeMove<1>(vec,&ddice);
				
				//CERR<<bd.stone[WHITE][0]<<endl;
				
				if( bd.NInHouse[WHITE]>0 && isNearTee(bd.stone[WHITE][0],1.0) ){
					++suc[0][0];
				}
			}
			
			//ショートティーショット
			suc[1][0]=0;
			for(int i=0;i<N;++i){
				bd.init();
				vec.setSpin(dice.rand()%2);
				genShortTeeShot(&vec);
				bd.makeMove<1>(vec,&ddice);
				
				//CERR<<bd.stone[WHITE][0]<<endl;
				
				if( bd.NInHouse[WHITE]>0
				   && isNearTee(bd.stone[WHITE][0],F_HOUSE_RAD+F_STONE_RAD)
				   && isThisSideOfTeeLine(bd.stone[WHITE][0].y) ){
					++suc[1][0];
				}
			}
			
			//レンジティーショット
			suc[2][0]=0;
			for(int i=0;i<N;++i){
				bd.init();
				rvec.setSpin(dice.rand()%2);
				genTeeRShot(&rvec);
				vec=rvec.genNR(&ddice);
				bd.makeMove<1>(vec,&ddice);
				
				//CERR<<bd.stone[WHITE][0]<<endl;
				
				if( bd.NInHouse[WHITE]>0 && isNearTee(bd.stone[WHITE][0],1.0) ){
					++suc[2][0];
				}
			}
			
			//ストロングヒットによるテイクアウト
			suc[3][0]=0;
			suc[3][1]=0;
			for(int i=0;i<N;++i){
				bd.init();
				
				double r0=ddice.drand();double r1=ddice.drand();
				fpn_t x=FX_TEE + F_HOUSE_RAD *r0*cos(r1*M_PI);fpn_t y=FY_TEE + F_HOUSE_RAD*r0*sin(r1*M_PI);
				
				bd.pushStone(BLACK,x,y);
				
				vec.setSpin(dice.rand()%2);
				genStrongHit(bd.stone[BLACK][0],&vec,1);
				
				bd.makeMove<1>(vec,&ddice);
				
				if( bd.NInHouse[BLACK]==0 ){//TakeOut
					++suc[3][0];
				}
				if( bd.NInHouse[WHITE]>0 ){
					++suc[3][1];
				}
			}
			
			//ミドルヒットによるテイクアウト
			suc[4][0]=0;
			suc[4][1]=0;
			for(int i=0;i<N;++i){
				bd.init();
				
				double r0=ddice.drand();double r1=ddice.drand();
				fpn_t x=FX_TEE + F_HOUSE_RAD*r0*cos(r1*M_PI);fpn_t y=FY_TEE + F_HOUSE_RAD*r0*sin(r1*M_PI);
				
				bd.pushStone(BLACK,x,y);
				
				vec.setSpin(dice.rand()%2);
				genMiddleHit(bd.stone[BLACK][0],&vec,1);
				
				bd.makeMove<1>(vec,&ddice);
				
				if( bd.NInHouse[BLACK]==0 ){//TakeOut
					++suc[4][0];
				}
				if( bd.NInHouse[WHITE]>0 ){
					++suc[4][1];
				}
			}
			
			//ウィークヒットによるテイクアウト
			suc[5][0]=0;
			suc[5][1]=0;
			for(int i=0;i<N;++i){
				bd.init();
				
				double r0=ddice.drand();double r1=ddice.drand();
				fpn_t x=FX_TEE + F_HOUSE_RAD*r0*cos(r1*M_PI);fpn_t y=FY_TEE + F_HOUSE_RAD*r0*sin(r1*M_PI);
				
				bd.pushStone(BLACK,x,y);
				
				vec.setSpin(dice.rand()%2);
				genWeakHit(bd.stone[BLACK][0],&vec,1);
				
				bd.makeMove<1>(vec,&ddice);
				
				if( bd.NInHouse[BLACK]==0 ){//TakeOut
					++suc[5][0];
				}
				if( bd.NInHouse[WHITE]>0 ){
					++suc[5][1];
				}
			}
			
			
			CERR<<"TeeShot      "<<suc[0][0]<<"/"<<N<<endl;
			CERR<<"ShortTeeShot "<<suc[1][0]<<"/"<<N<<endl;
			CERR<<"RTeeShot     "<<suc[2][0]<<"/"<<N<<endl;
			CERR<<"StrongHit    "<<suc[3][0]<<"/"<<N<<" "<<suc[3][1]<<"/"<<N<<endl;
			CERR<<"MiddleHit    "<<suc[4][0]<<"/"<<N<<" "<<suc[4][1]<<"/"<<N<<endl;
			CERR<<"WeakHit      "<<suc[5][0]<<"/"<<N<<" "<<suc[5][1]<<"/"<<N<<endl;
			//CERR<<"TakeOut      "<<suc[1]<<"/"<<N<<endl;
			//CERR<<"TakeOut      "<<suc[1]<<"/"<<N<<endl;
			//testSimulator();
			
			//シミュレータの精度テスト
			//ティーショットがどの程度ティーに近づくかを判定
			/*
			fpn_t dist;//距離
			fMStone<> org,dst;
			fPosXY<> mato;
			fMoveXY<> mv;
			
			const fpn_t ltar=fabs( FY_THROW - FY_TEE );
			
			mato.set( FX_TEE + 0.5, FY_TEE + 1.4 );
			
			mv.setSpin( Spin::RIGHT );
			genDrawShot(mato,&mv);
			
			org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), FANGV_ORG );
			
			for(int i=0;i<16;++i){
				Simulator::simulateSolo( org, &dst, 1/double(1<<i) );
				dist=XYtoR( dst.x-mato.x, dst.y-mato.y );
				CERR<<"step "<<(1<<i)<<" "<<dist<<endl;
			}
			
			//ヒットショットの最高速度を調べる
			*/
			
			
			//角速度による運動の変化を求める
			{
				fMStone<> org;
				fMoveXY<> mv;
				for(int i=0;i<4;++i){
					bd.init();
					bd.pushStone(BLACK,FX_TEE,FY_TEE+1);
					//mv.setSpin(Spin::RIGHT);
					genStrongHit(bd.stone[BLACK][0],&mv);
					org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), -FANGV_ORG*i );
					Simulator::simulateb2d<0>( &bd, org, 0.0001 );
					//bd.makeMove<1>(mv,&ddice);
					bd.printStones();
				}
			}
			
			//ここから自作シミュレータの精度チェック
			
			//衝突処理チェック
			{
				//動-静
				CERR<<"AW_AS Collision check"<<endl;
				FMoment aw,as,taw,tas;
				fPosXY<> pos_aw,pos_as;
				fpn_t dtheta,w,v;
				StoneState state;
				for(int i=0;i<4;++i){
					pos_aw.set(0,0);
					double r=ddice.drand();
					pos_as.set( pos_aw.getX()+2*F_STONE_RAD*cos(2*r*M_PI), pos_aw.getX()+2*F_STONE_RAD*sin(2*r*M_PI) );
					dtheta=XYtoT( pos_as.getY()-pos_aw.getY(), pos_as.getX()-pos_aw.getX() ) + (M_PI-0.5)*( ddice.drand()-0.5 );
					w=ddice.drand()-0.5;
					v=FV_MAX * ddice.drand();
					
					state.init();
					state.setColor(BLACK);
					
					aw.set( pos_aw.getX(), pos_aw.getY(), dtheta, 0, v, w );
					as.set( pos_as.getX(), pos_as.getY(), 0, 0, 0, 0 );
					
					//高速シミュレート
					taw=aw;tas=as;
					tick();
					FastSimulator::collisionAw_As( &taw, &tas );
					tock();
					
					CERR<<"AW ";
					CERR<<" vx,vy "<<fPosXY<>(taw.getVX(),taw.getVY())
					<<" "<<taw.getW()<<endl;
					CERR<<"AS ";
					CERR<<" vx,vy "<<fPosXY<>(tas.getVX(),tas.getVY())
					<<" "<<tas.getW()<<endl;
					
					//通常シミュレート
					b2Board bd;
					bd.setAwakeStone( aw );
					bd.setAsleepStone( as, state ); 
					
					tick();
					while( bd.body[1]->GetLinearVelocity().x==0 && bd.body[1]->GetLinearVelocity().y==0 ){
						Simulator::step( &bd, 0.00001 );
					}
					tock();
					
					CERR<<"AW ";
					CERR<<" vx,vy "<<fPosXY<>(bd.body[0]->GetLinearVelocity().x,bd.body[0]->GetLinearVelocity().y)
					<<" "<<bd.body[0]->GetAngularVelocity()<<endl;
					CERR<<"AS ";
					CERR<<" vx,vy "<<fPosXY<>(bd.body[1]->GetLinearVelocity().x,bd.body[1]->GetLinearVelocity().y)
					<<" "<<bd.body[1]->GetAngularVelocity()<<endl;
				}
			}
			{
				//動-動
				CERR<<"2AW Collision check"<<endl;
				FMoment aw[2],taw[2];
				fPosXY<> pos[2];
				fpn_t dtheta[2],w[2],v[2];
				for(int i=0;i<4;++i){
					pos[0].set(0,0);
					double r=ddice.drand();
					pos[1].set( pos[0].getX()+2*F_STONE_RAD*cos(2*r*M_PI), pos[0].getX()+2*F_STONE_RAD*sin(2*r*M_PI) );
					dtheta[0]=XYtoT( pos[1].getY()-pos[0].getY(), pos[1].getX()-pos[0].getX() ) + (M_PI-0.5)*( ddice.drand()-0.5 );
					dtheta[1]=XYtoT( pos[0].getY()-pos[1].getY(), pos[0].getX()-pos[1].getX() ) + (M_PI-0.5)*( ddice.drand()-0.5 );
					w[0]=ddice.drand()-0.5;
					w[1]=ddice.drand()-0.5;
					v[0]=FV_MAX * ddice.drand();
					v[1]=FV_MAX * ddice.drand();
					
					aw[0].set( pos[0].getX(), pos[0].getY(), dtheta[0], 0, v[0], w[0] );
					aw[1].set( pos[1].getX(), pos[1].getY(), dtheta[1], 0, v[1], w[1] );
					
					//高速シミュレート
					taw[0]=aw[0];taw[1]=aw[1];
					tick();
					FastSimulator::collision2Aw( &taw[0], &taw[1] );
					tock();
					
					CERR<<"AW0 ";
					CERR<<" vx,vy "<<fPosXY<>(taw[0].getVX(),taw[0].getVY())
					<<" "<<taw[0].getW()<<endl;
					CERR<<"AW1 ";
					CERR<<" vx,vy "<<fPosXY<>(taw[1].getVX(),taw[1].getVY())
					<<" "<<taw[1].getW()<<endl;
					
					//通常シミュレート
					b2Board bd;
					bd.setAwakeStone( aw[0], BLACK );
					bd.setAwakeStone( aw[1], WHITE ); 
					
					tick();
					
					fpn_t orgw0=w[0];
					while( fabs(bd.body[0]->GetAngularVelocity() - orgw0 ) < 0.001 ){
						//速度内積が負 -> 互いに近づいている
						Simulator::step( &bd, 0.00001 );
					}
					tock();
					
					CERR<<"AW0 ";
					CERR<<" vx,vy "<<fPosXY<>(bd.body[0]->GetLinearVelocity().x,bd.body[0]->GetLinearVelocity().y)
					<<" "<<bd.body[0]->GetAngularVelocity()<<endl;
					CERR<<"AW1 ";
					CERR<<" vx,vy "<<fPosXY<>(bd.body[1]->GetLinearVelocity().x,bd.body[1]->GetLinearVelocity().y)
					<<" "<<bd.body[1]->GetAngularVelocity()<<endl;
				}
			}
			
			//1_0シミュレーション
			{
				CERR<<"1_0 Fast Simulation check"<<endl;
				fMStone<> org;
				fMoveXY<> mv;
				fpn_t x,y;
				for(int i=0;i<8;++i){
					if( i%2==0 ){
						mv.setSpin(ddice.rand()%2);
						double r0=ddice.drand();double r1=ddice.drand();
						x=FX_TEE + 3.0*r0*cos(2*r1*M_PI);y=FY_TEE + 3.0*r0*sin(2*r1*M_PI);
						CERR<<"dst : ( "<<x<<", "<<y<<" )"<<endl;
					}
					bd.init();
					genDrawShot(fPosXY<>(x,y),&mv);
					org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), (mv.getSpin()==Spin::RIGHT) ? FANGV_ORG : (-FANGV_ORG) );
					if( i%2==1 ){
						tick();
						Simulator::simulateb2d<0>( &bd, org, 0.0001 );//通常シミュレート
						tock();
					}else{
						tick();
						if( FastSimulator::simulateSoloF<0>( &bd, mv )==0 ){//高速シミュレート
							CERR<<"no collision."<<endl;
						}
						tock();
					}
					bd.printStones();
				}
			}
			//1_1シミュレーション
			{
				CERR<<"1_1 Fast Simulation check"<<endl;
				fMStone<> org;
				fMoveXY<> mv;
				fpn_t x,y;
				for(int i=0;i<8;++i){
					if( i%2==0 ){
						double r0=ddice.drand();double r1=ddice.drand();
						x=FX_TEE + 2*r0*cos(2*r1*M_PI);y=FY_TEE + 2*r0*sin(2*r1*M_PI);
					}
					bd.init();
					bd.pushStone(BLACK,x,y);
					genMiddleHit(bd.stone[BLACK][0],&mv);
					org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), (mv.getSpin()==Spin::RIGHT) ? FANGV_ORG : (-FANGV_ORG) );
					if( i%2==1 ){
						tick();
						Simulator::simulateb2d<0>( &bd, org, 0.0001 );//通常シミュレート
						tock();
					}else{
						tick();
						if( FastSimulator::loopHitZone1_1<0>( &bd, org )==0 ){//高速シミュレート
							CERR<<"no collision."<<endl;
						}
						tock();
					}
					bd.printStones();
				}
			}
			//1_Nシミュレーション
			{
				CERR<<"1_N Fast Simulation check"<<endl;
				fMStone<> org;
				fMoveXY<> mv;
				int col[4];
				fPosXY<> pos[4];
				int turn;
				for(int i=0;i<8;++i){
					bd.init();
					if( i%2==0 ){
						
						turn=ddice.rand()%8;
						
						for(int j=0;j<4;++j){
							col[j]=(ddice.rand()%2) ? BLACK : WHITE;
							
							while(1){
								double r0=ddice.drand();double r1=ddice.drand();
								fpn_t x=FX_TEE + 2*r0*cos(2*r1*M_PI);fpn_t y=FY_TEE + 2*r0*sin(2*r1*M_PI);
								int ok=1;
								for(int k=0;k<j;++k){
									if( XYtoR( x-pos[k].x, y-pos[k].y ) < F_STONE_RAD*2 ){
										ok=0;break;	
									}
								}
								if( ok ){
									pos[j].set( x, y );
									break;
								}
							}
						}
					}
					bd.turn=turn;
					for(int j=0;j<4;++j){
						bd.pushStone(col[j],pos[j].x,pos[j].y);
					}
					bd.setState();
					if( i%2==0 ){
						CERR<<"col : "<<OutColor(TURNtoCOL(turn))<<" org : ";
						bd.printStones();
					}
					genDrawShot(fPosXY<>(FX_TEE,FY_TEE+5),&mv);
					org.set( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), (mv.getSpin()==Spin::RIGHT) ? FANGV_ORG : (-FANGV_ORG) );
					if( i%2==1 ){
						tick();
						Simulator::simulateb2d<0>( &bd, org, 0.0001 );//通常シミュレート
						tock();
					}else{
						tick();
						if( FastSimulator::loopHitZone1_N<0>( &bd, org )==0 ){//高速シミュレート
							CERR<<"no collision."<<endl;
						}
						tock();
					}
					bd.printStones();
				}
			}
			//getchar();
#endif
		}
		int init1G(const ClientField& cf){
			mvTT.init();
			bdTT.init();
		}
		int play(const ClientField& cf,fMoveXY<> *const dst){
			
			MonteCarloPlayer mcp;
			MoveXY mv;
			int ret=mcp.play(myField,&mv);
			
			//fMoveXYに変換
			convMove(mv,dst);
			
			//公式の座標系に変換
			*dst=convMove_Mine_Official(*dst);
			
			if( ret!=-1 ){
				CERR<<"My Move : "<<*dst<<endl;
				ret=0;
			}
			return ret;
		}
	};
	
}

