//デジタルカーリング用
//シンプルなクライアント

#include <cstdio>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <cstring>
#include <string>
#include <sys/types.h>
#include <errno.h>
#include <random>
#include <bitset>

#include <sys/time.h>

#include <thread>
#include <mutex>
#include <atomic>

#include <boost/lexical_cast.hpp>
//#include <boost/thread.hpp>

#include <Box2D/Box2D.h>

#include "settings.h"

#ifdef _WIN32

#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>

#else

#include <sys/param.h>
#include <sys/socket.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <netdb.h>

#endif

#include "../common/common_all.h"
#include "defines.h"
#include "lib.hpp"
#include "tools/prim.hpp"
#include "tools/hash.hpp"
#include "eval/stat.hpp"
#include "eval/eval.hpp"
#include "strategy/checkmate.hpp"
#include "strategy/killer.hpp"
#include "simulation/simulator.hpp"
#include "simulation/fastsimulator.hpp"
#include "simulation/genshot.hpp"
#include "simulation/rand.hpp"
#include "tools/field.hpp"

using namespace DigitalCurling;

const unsigned short PORT_NUM=9876;

class AdminInfo{
private:
	void close(){
		//管理用ファイルを消す
		if( id_num >= 0 ){//id_numが設定されている
			char str_file[256];
			snprintf( str_file, 256, "conf/%d.txt", id_num );
			if( remove( str_file ) >=0 ){
				CERR<<"succeeded to remove Admin file."<<endl;
			}else{
				CERR<<"failed to remove Admin file."<<endl;
			}
		}
	}
public:
	
	int id_num;
	
	AdminInfo():
	id_num(-1){}
	
	~AdminInfo(){
		close();
	}
};

AdminInfo gadm;
ClientField myField;

#include "ayumu.hpp"

int decideID(){
	//自分が複数ログインするときのために、
	//自分の識別番号を決める
	
	int i;
	FILE *pf;
	char str_file[256];
	//ログの番号を決める
	for ( i = 0; i < 999; i++ )
	{
		snprintf( str_file, 256, "conf/%d.txt", i );
		pf = fopen( str_file, "r" );
		if ( pf == NULL ) { break; }
		int iret = fclose( pf );
		if ( iret < 0 ) { return -1; }
	}
	gadm.id_num=i;
	//最後に新しい番号のファイルを作る
	snprintf( str_file, 256, "conf/%d.txt", i );
	pf = fopen( str_file, "w" );
	if ( pf == NULL ) { return -1; }
	
	return 0;
}

//グローバル変数
Client client;
NETWORKINFO NetworkInfo;

std::string gmy_id;
int gNLimitGames;//試合数限度

void searchToken(const char* str,const int size,int *ret){
	for(int c=*ret;c<size-1;++c){
		if(str[c]=='\0'){
			if(str[c+1]=='\0'){
				*ret=-1;return;
			}else{
				*ret=c+1;return;
			}
		}
	}
	*ret=-1;
	return;
}

int sendCommand(const char* cmd_send){
	if( strlen(cmd_send)>0 ){
		CERR<<"Client << "<<cmd_send<<endl;
	}
#if defined(_WIN32) && defined(FOR_ENGINE)
	DWORD NBytesWritten;
	WriteFile( GetStdHandle( STD_OUTPUT_HANDLE ), cmd_send, strlen( cmd_send ), &NBytesWritten, NULL );
#else
	send( NetworkInfo.dstSocket, cmd_send, strlen(cmd_send)+1, 0 );
#endif
	return 0;
}

int recvCommand(){	
	
	//受信
	
	int games=0;
	
	char cmd_recv[BUFFER_SIZE]={0};
	char cmd_send[BUFFER_SIZE]={0};
	
	while(1){
		
		char *tok;
		int c;
		
		memset(cmd_recv,0,sizeof(char)*BUFFER_SIZE);
		memset(cmd_send,0,sizeof(char)*BUFFER_SIZE);
		
#if defined(_WIN32) && defined(FOR_ENGINE)
		DWORD NBytesRead;
		ReadFile( GetStdHandle( STD_INPUT_HANDLE ), cmd_recv, sizeof( cmd_recv ), &NBytesRead, NULL );
#else
		if( recv( NetworkInfo.dstSocket, cmd_recv, sizeof( cmd_recv ), 0 )<0 ){return -1;}
#endif
		//convert to o-moji
		for(int i=0;i<BUFFER_SIZE;++i){
			cmd_recv[i]=toupper(cmd_recv[i]);
		}
		
		c=0;
		
		while( c>=0 ){
			
			tok=&cmd_recv[c];
			
			CERR<<"Server >> "<<tok<<endl;
			if( strstr(tok,"CONNECTED") ){
			}else if( strstr(tok,"LOGIN OK") ){
			}else if( strstr(tok,"LOGIN NG") ){
				return -1;
			}else if( strstr(tok,"ISREADY") ){
				
				client.init1G(myField);
				
				sendCommand( "READYOK" );
			}else if( strstr(tok,"NEWGAME") ){
				
				myField.initGame();
				//プレーヤー名が送られてくるので記録
				{
					char tmp[32];
					
					if( GetArgument(tmp,sizeof(tmp),tok,1) ){
						myField.setName(0,std::string(tmp));
					}else{
						myField.setName(0,"First");
					}
					if( GetArgument(tmp,sizeof(tmp),tok,2) ){
						myField.setName(1,std::string(tmp));
					}else{
						myField.setName(1,"Second");
					}
				}
				client.init1G(myField);
			}else if( strstr(tok,"POSITION") ){
				//ストーンの位置については毎回初期化
				char tmp[32];
				for( int i=0; i<16; i++ ){
					GetArgument(tmp,sizeof(tmp),tok,2*i+1);
					float x=( float )atof( tmp );
					GetArgument(tmp,sizeof(tmp),tok,2*i+2);
					float y=( float )atof( tmp );
					
					myField.setStonePos(i,x,y);
				}
			}else if( strstr(tok,"SETSTATE") ){
				
				//状態を設定
				char tmp[32];
				GetArgument(tmp,sizeof(tmp),tok,1);
				myField.setTurn( atoi(tmp) );
				GetArgument(tmp,sizeof(tmp),tok,2);
				myField.setEnd( atoi(tmp) );
				GetArgument(tmp,sizeof(tmp),tok,3);
				myField.setToEnd( atoi(tmp) );
				
			}else if( strstr(tok,"GO") ){
				char tmp[32];
				//手を決めて送る
				//まず制限時間の確認
				for(int c=0;c<2;++c){
					GetArgument(tmp,sizeof(tmp),tok,c+1);
					uint64 tl=(uint64)atoi(tmp);
					if( tl==0ULL ){//これが無制限の意味
						myField.setTimeLimit( c, TIME_LIMIT_UNLIMITED );
					}else{
						myField.setTimeLimit( c, tl );
					}
				}
				
				if( myField.my_first_color<0 ){
					//ここで自分の先後がわかる
					if( myField.getTurnColor()==0 ){
						myField.setMyFirstColor(0);
					}else{
						myField.setMyFirstColor(1);
					}
				}
				assert( myField.isMyTurn() );
				
				myField.broadcast();
				fMoveXY<> fmv;
				int ret=client.play(myField,&fmv);
				
				if( ret==-1 ){//投了
					snprintf( cmd_send, sizeof(cmd_send), "CONCEDE" );
				}else{
					snprintf( cmd_send, sizeof(cmd_send), "BESTSHOT %f %f %d", (float)fmv.getVX(), (float)fmv.getVY(), (int)fmv.getSpin() );
				}
				sendCommand( cmd_send );
				
				
				//置換表掃除
				if( mvTT.npages>0 ){
					mvTT.init();
				}
				if( bdTT.npages>16 ){
					bdTT.init();
				}
			}else if( strstr(tok,"TIMEOUT") ){
				//タイムアウト
			}else if( strstr(tok,"CONCEDE") ){
				//投了
			}else if( strstr(tok,"RUNSHOT") ){
				//ショット結果
			}else if( strstr(tok,"TOTALSCORE") ){
				//総合結果
			}else if( strstr(tok,"SCORE") ){
				//SCOREを含むので、TOTALSCOREより後の必要
				char tmp[32];
				GetArgument(tmp,sizeof(tmp),tok,1);
				myField.setEndScore( atoi(tmp) );
				myField.closeEnd();
				myField.initEnd();//次のエンドの初期化までやっておく
				client.init1G(myField);
			}else if( strstr(tok,"GAMEOVER") ){
				char tmp[32];
				GetArgument(tmp,sizeof(tmp),tok,1);
				if( strstr(tmp,"WIN") ){
					myField.closeGame(1);
				}else if( strstr(tmp,"LOSE") ){
					myField.closeGame(-1);
				}else{
					myField.closeGame(0);
				}
				
				++games;
				if( games >= gNLimitGames ){
					return 1;
				}
				
			}else if( strstr(tok,"DISCONNECT") ){
				return 1;
			}else{
				Sleep(1);
			}
			searchToken(cmd_recv,BUFFER_SIZE,&c);
			//CERR<<c<<" , "<<cmd_recv[c]<<endl;
		}
		
	}
	return 0;
	
}

int main(int argc,char* argv[]){
	
	setvbuf(stdout,NULL,_IONBF,0);
	setvbuf(stdin,NULL,_IONBF,0);
	setvbuf(stderr,NULL,_IONBF,0);
	
	std::cerr<<std::setprecision(8);
	
	gNLimitGames=100;
	
	//引数から諸々を指定
	for( int i=1;i<argc;++i){
		if( strstr(argv[i],"-g") ){
			gNLimitGames=atoi(argv[i+1]);
		}
		/*if( strstr(argv[i],"-t") ){
		 tl=atoi(argv[i+1]);
		 }*/
	}
	
	char cmd_send[BUFFER_SIZE]={0};
	char cmd_recv[BUFFER_SIZE]={0};
	
	//ID設定
	//VERSIONモードでは、バージョンをIDに入れる
#ifdef BUILD_VERSION
	const std::string my_name=MY_NAME+MY_VERSION;
	gmy_id=MY_ID+MY_VERSION;
#else
	if( decideID() != 0 ){
		CERR<<"failed to get my ID number."<<endl;
		exit(1);
	}
	const std::string my_name=MY_NAME+boost::lexical_cast<std::string>(gadm.id_num);
	gmy_id=MY_ID+boost::lexical_cast<std::string>(gadm.id_num);
#endif
	
	//const std::string my_name=MY_NAME+MY_VERSION;
	//gmy_id=MY_ID+MY_VERSION;
	
	CERR<<"I'm "<<my_name<<endl;
	
#ifndef FOR_ENGINE
	//設定
	strcpy(NetworkInfo.ID,std::string(gmy_id).c_str());
	strcpy(NetworkInfo.PASS,std::string(MY_PASSWORD).c_str());
	strcpy(NetworkInfo.NAME,std::string(my_name).c_str());
	
	//通信設定
	strcpy(NetworkInfo.IP,std::string(SERVER_IP).c_str());
	NetworkInfo.PORT=PORT_NUM;
	NetworkInfo.IsConnected = FALSE;
	
#ifdef _WIN32
	//Windows 独自の設定
	WSADATA data;
	if( SOCKET_ERROR == WSAStartup( MAKEWORD(2,0), &data ) ){
		cerr<<"failed to initialize WSA-data."<<endl;
		exit(1);
	}
#endif
	
	//open socket
	NetworkInfo.dstSocket = socket( AF_INET, SOCK_STREAM, 0 );
	if( NetworkInfo.dstSocket<0 ){
		cerr<<"failed to open socket."<<endl;
	}
	
	//sockaddr_in 構造体のセット
	memset( &NetworkInfo.dstAddr, 0, sizeof( NetworkInfo.dstAddr ) );
	NetworkInfo.dstAddr.sin_port = htons( NetworkInfo.PORT );
	NetworkInfo.dstAddr.sin_family = AF_INET;
	NetworkInfo.dstAddr.sin_addr.s_addr = inet_addr( NetworkInfo.IP );
	
	//接続
	if( SOCKET_ERROR == connect( NetworkInfo.dstSocket, (struct sockaddr *) &NetworkInfo.dstAddr, sizeof( NetworkInfo.dstAddr )) ){
		cerr<<"failed to connect to server."<<endl;
		exit(1);//ソケット接続失敗
	}
	
	//接続結果受け取り
	int q=recv( NetworkInfo.dstSocket, cmd_recv, sizeof( cmd_recv ), 0 );
	CERR<<q<<" "<<cmd_recv<<endl;
	
	if( !std::strstr(cmd_recv,"CONNECTED") ){
		cerr<<"failed to make logging-in connection"<<endl;
		exit(1);
	}
	
	//接続完了
	NetworkInfo.IsConnected = TRUE;
	CERR<<"connected to server."<<endl;
	
	//ログイン
	snprintf( cmd_send, sizeof(cmd_send), "LOGIN %s %s %s", NetworkInfo.ID, NetworkInfo.PASS, NetworkInfo.NAME );
	CERR<<cmd_send<<endl;
	send( NetworkInfo.dstSocket, cmd_send, strlen( cmd_send )+1, 0 );
	
#endif //FOR_ENGINE
	
	//試合のための初期化処理
	initTransProb();
	initReward();
	initPrim();
	initHash();
	FastSimulator::init();
	
	client.initAll();
	myField.initAll();
	
	//通信スレッド
	recvCommand();
	
#ifndef FOR_ENGINE
	//終了
	sendCommand("LOGOUT");
	
	CERR<<"logeed out."<<endl;
	
	// ソケットを閉じる
#ifdef _WIN32
	closesocket(NetworkInfo.dstSocket);
	WSACleanup();
#else
	close(NetworkInfo.dstSocket);
#endif
	
	NetworkInfo.IsConnected = FALSE;
	CERR<<"closed socket."<<std::endl;
#endif //FOR_ENGINE
	
	return 0;
}