//デジタルカーリング
//勝率ベースのエンド末端評価点

namespace DigitalCurling{
	
	const int SCOREGAP_MIN=-6;
	const int SCOREGAP_MAX=+6;
	
	//開区間の得点差を閉区間に落とし込む
	constexpr int RSGtoSG(int rsd){
		return (rsd>SCOREGAP_MAX)?SCOREGAP_MAX:((rsd<SCOREGAP_MIN)?SCOREGAP_MIN:rsd);
	}
	
	//必要な配列サイズ（最大インデックス+1）
	constexpr int SCOREGAP_LENGTH=SCOREGAP_MAX - SCOREGAP_MIN +1;
	
	//閉区間の得点差からインデックスを出す
	constexpr int SGtoIDX(int sd){return sd-SCOREGAP_MIN;}
	
	//インデックスから閉区間の得点差を出す
	constexpr int IDXtoSG(int idx){return idx+SCOREGAP_MIN;}
	
	/**************************得点差遷移確率**************************/
	
	fpn_t dScoreProb_EASY[ NTURNS+1 ]={
		//先手が+s点を取る確率
		0.0000001,
		0.000001,
		0.00001,
		0.0005,
		0.009,
		0.03,
		0.18,
		0.34,
		0.12,//0
		0.12,
		0.04,
		0.012,
		0.003,
		0.0003,
		0.00001,
		0.000001,
		0.0000001,
	};
	
	const fpn_t dWinningPercentage_EX=0.2;//延長戦の先手予測勝率
	fpn_t dScoreProb[ MAX_NENDS ][ SCOREGAP_LENGTH ][ SCOREGAP_LENGTH ];
	
	void initTransProb(){
		//遷移確率初期化
		
		//dScoreProb_EASYを正規化
		fpn_t sum=0;
		for(int s=-8;s<=+8;++s){
			sum+=dScoreProb_EASY[s+8];
		}
		for(int s=-8;s<=+8;++s){
			dScoreProb_EASY[s+8]/=sum;
		}
		
		
		//得点確率
		for(int e=0;e<MAX_NENDS;++e){
			for(int s=SCOREGAP_MIN;s<=SCOREGAP_MAX;++s){
				for(int ss=SCOREGAP_MIN;ss<=SCOREGAP_MAX;++ss){
					dScoreProb[e][SGtoIDX(s)][SGtoIDX(ss)]=dScoreProb_EASY[ss+8];
				}
			}
		}
	}
	
	/**************************エンド後勝率**************************/
	
	//残りエンド数×次エンドの手番×点差の報酬を設定
	fpn_t dWinningPercentage[ MAX_NENDS+1 ][ SCOREGAP_LENGTH ];

	fpn_t getEndWP(int c,int e,int s,int next_sg){
		fpn_t ev;
		if( s>=0 ){
			int idx=SGtoIDX(RSGtoSG(next_sg));
			if( c==BLACK ){
				ev=dWinningPercentage[e][idx];
			}else{
				ev=1.0-dWinningPercentage[e][idx];
			}
		}else{
			int idx=SGtoIDX(RSGtoSG(-next_sg));
			if( c==BLACK ){
				ev=1.0-dWinningPercentage[e][idx];
			}else{
				ev=dWinningPercentage[e][idx];
			}
		}
		return ev;
	}
	
	/**************************序盤用のエンド報酬**************************/
	
	//序盤に勝率ベースの報酬を考えても差がつきにくいため緩い評価になる
	//そのため序盤用のエンド報酬テーブルを用意しておく
	
	constexpr fpn_t DOPENING_REWARD_W1=0;
	constexpr fpn_t DOPENING_REWARD_0=-0.33333;
	
	fpn_t dOpeningReward[ 2 ][ NTURNS+1 ];
	
	fpn_t getOpeningEndReward(int c,int s){
		return dOpeningReward[c][s+8];
	}
	
	
	void initReward(){
		//報酬系初期化
		
		//最終エンド後の勝率
		for(int s=SCOREGAP_MIN;s<=-1;++s){//先手負け
			dWinningPercentage[0][SGtoIDX(s)]=0;
		}
		dWinningPercentage[0][SGtoIDX(0)]=dWinningPercentage_EX;//延長戦突入
		for(int s=1;s<=SCOREGAP_MAX;++s){//先手勝ち
			dWinningPercentage[0][SGtoIDX(s)]=1;
		}
		
		//それまでのエンド後の勝率
		for(int e=1;e<MAX_NENDS+1;++e){
			for(int s=SCOREGAP_MIN;s<=SCOREGAP_MAX;++s){
				dWinningPercentage[e][SGtoIDX(s)]=0;
				for(int ss=SCOREGAP_MIN;ss<=-1;++ss){//手番交代
					dWinningPercentage[e][SGtoIDX(s)]+=dScoreProb[e-1][SGtoIDX(s)][SGtoIDX(ss)] * ( 1.0-dWinningPercentage[e-1][SGtoIDX(RSGtoSG(-(s+ss)))] );
				}
				for(int ss=0;ss<=SCOREGAP_MAX;++ss){//手番そのまま
					dWinningPercentage[e][SGtoIDX(s)]+=dScoreProb[e-1][SGtoIDX(s)][SGtoIDX(ss)] * dWinningPercentage[e-1][SGtoIDX(RSGtoSG(s+ss))];
				}
			}
		}
		
		//序盤報酬
		for(int i=-8;i<=-1;++i){//後手が得点し先後交代
			dOpeningReward[ 0 ][ i+8 ]= DOPENING_REWARD_W1 + i -(-1);
			dOpeningReward[ 1 ][ i+8 ]= -(DOPENING_REWARD_W1 + i -(-1));
		}
		for(int i=0;i<=8;++i){//先後そのまま
			dOpeningReward[ 0 ][ i+8 ]= DOPENING_REWARD_0 + i -(0);
			dOpeningReward[ 1 ][ i+8 ]= -(DOPENING_REWARD_0 + i -(0));
		}
		
		//報酬表表示
#ifdef MONITOR
		CERR<<"dWinningPercentage = {"<<endl;
		for(int e=0;e<MAX_NENDS+1;++e){
			for(int s=SCOREGAP_MIN;s<=SCOREGAP_MAX;++s){
				CERR<<" "<<dWinningPercentage[e][SGtoIDX(s)]<<",";
			}
			CERR<<endl;
		}
		CERR<<"};"<<endl;
		
		CERR<<"dOpeningReward = {"<<endl;
		for(int c=0;c<2;++c){
			for(int s=-8;s<=+8;++s){
				CERR<<" "<<dOpeningReward[c][s+8]<<",";
			}
			CERR<<endl;
		}
		CERR<<"};"<<endl;
#endif
	}
	
	/**************************一時的エンド報酬テーブル**************************/
	
	//色々調整があるので、毎プレー前にこの配列をセット
	
	fpn_t dTmpEndReward[ 2 ][ NTURNS+1 ];
	
	template<class board_t>
	void setTmpEndReward(const board_t& bd,const int my_col){
		
		const int ops_col=flipColor(my_col);
		const int e=bd.getEnd();
		const int t=bd.getTurn();
		const int rs=bd.getRelScore();
		
		int conv[ NTURNS+1 ];//得点換算テーブル(256が1点に相当)
		
		for(int s=-8;s<=+8;++s){
			conv[s+8]=s*256;
		}
		
		int myLead=(my_col==BLACK)?bd.getRelScore():(-bd.getRelScore());
		if( myLead > 0 ){
			//自分がリードしている
			//自分の得点の丸め込みを行う
			int marume_line=8*256;
			if( my_col==BLACK ){
				switch( myLead ){
					case 0:assert(0);break;
					case 1:marume_line=256+170;break;
					case 2:marume_line=205;break;
					default:marume_line=3;break;
				}
				for(int s=(marume_line+255)/256;s<=8;++s){
					conv[s+8]=marume_line;
				}
			}else{
				if( t!=TURN_LAST ){//ラストストーンならやらない
					switch( myLead ){
						case 0:assert(0);break;
						case 1:marume_line=4*256;break;
						case 2:marume_line=3*256;break;
						default:marume_line=2*256;break;
					}
					for(int s=-8;s<=(-marume_line-255)/256;++s){
						conv[s+8]=-marume_line;
					}
				}
			}
		}else if( myLead==0 ){//同点
			
		}
		
		if( e>=END_JOBAN ){
			for(int c=0;c<2;++c){
				for(int s=-8;s<=-1;++s){
					int sc=conv[s+8]/256;
					int hasu=conv[s+8]%256;
					dTmpEndReward[c][s+8]=dOpeningReward[c][sc+8]*(256-hasu) + dOpeningReward[c][max(-8,sc-1)+8]*hasu;
				}
				for(int s=0;s<=+8;++s){
					int sc=conv[s+8]/256;
					int hasu=conv[s+8]%256;
					dTmpEndReward[c][s+8]=dOpeningReward[c][sc+8]*(256-hasu) + dOpeningReward[c][min(+8,sc+1)+8]*hasu;
				}
			}
		}else{
			for(int c=0;c<2;++c){
				for(int s=-8;s<=-1;++s){
					int sc=conv[s+8]/256;
					int hasu=conv[s+8]%256;
					dTmpEndReward[c][s+8]=getEndWP(c,e,sc,rs+sc)*(256-hasu) + getEndWP(c,e,max(-8,sc-1),rs+max(-8,sc-1))*hasu -128;
				}
				for(int s=0;s<=+8;++s){
					int sc=conv[s+8]/256;
					int hasu=conv[s+8]%256;
					dTmpEndReward[c][s+8]=getEndWP(c,e,sc,rs+sc)*(256-hasu) + getEndWP(c,e,min(+8,sc+1),rs+min(+8,sc+1))*hasu -128;
				}
			}
		}
		
		//平均、分散調整
		//平均が0、標準偏差が0.5になるように
		fpn_t sum(0),sum2(0);
		
		for(int s=-8;s<=+8;++s){
			fpn_t r=dTmpEndReward[BLACK][s+8];
			sum+=r * dScoreProb_EASY[s+8];
			sum2+=r*r * dScoreProb_EASY[s+8];
		}
		
		fpn_t sigma=sqrt(max(0.000001,sum2 - sum*sum));
		fpn_t mu=sum;
		
		CERR<<"mu : "<<sum<<" sigma : "<<sigma<<endl;
		
		for(int s=-8;s<=+8;++s){
			dTmpEndReward[BLACK][s+8]=(dTmpEndReward[BLACK][s+8]-mu)/sigma*0.5;
		}
		for(int s=-8;s<=+8;++s){
			dTmpEndReward[WHITE][s+8]=(dTmpEndReward[WHITE][s+8]-(-mu))/sigma*0.5;
		}
		
		CERR<<"dTmpEndReward ( my color )= {"<<endl;
		for(int s=-8;s<=-1;++s){
			CERR<<" "<<dTmpEndReward[my_col][s+8]<<",";
		}
		CERR<<endl;
		CERR<<" "<<dTmpEndReward[my_col][0+8]<<",";
		CERR<<endl;
		for(int s=1;s<=+8;++s){
			CERR<<" "<<dTmpEndReward[my_col][s+8]<<",";
		}
		CERR<<endl;
		CERR<<"};"<<endl;
	}
	
	fpn_t getTmpEndReward(const int col,const int sc){
		return dTmpEndReward[col][sc+8];
	}
}
