//デジタルカーリング
//モンテカルロ関連

namespace DigitalCurling{
	namespace Ayumu{
		constexpr int N_DIV_TRIALS_LEAFNODE=4;//局面分岐閾値
		constexpr fpn_t ATT_RATE_LEAFCHILD=0.995;//減衰率
		
		constexpr int TDEPTH_MAX=3;
	
		
		struct KillerChild{
			//キラー手の検討データ
			KillerMove mv;
			
			float size;
			float eval_sum;
			float eval2_sum;
			
			void init(){
				size=0;
				eval_sum=0;
				eval2_sum=0;
			}
			
			void set(int at){
				init();
				mv.set(at);
			}
			
			void feed_eval(fpn_t eval){
				size+=1.0;
				eval_sum+=eval;
				eval2_sum+=eval*eval;
			}
		};
		
		struct RecursiveNode{
			//再帰局面データ
			
			//typede child_t;
			const static int N_CHILDS=2;
			const static int N_KILLER_CHILDS=8;
			
			const RecursiveNode *mother;//母局面
			
			uint64 hash;//局面ハッシュ
			KillerChild kchild[N_KILLER_CHILDS];//キラー着手			
			int killers;//キラー手の数
			int range;//認識レンジ
			int turn;//ターン
			//uint32 trials;
			fpn_t size;
			
#ifdef MULTI_THREADING
			volatile int _lock[1];//spin lock
#endif
			
			void init(){
				unlock();
			}
			
			template<class board_t>
			void set(const board_t& abd,int ar,uint64 ahash){
				init();
				hash=ahash;
				range=ar;
				turn=abd.getTurn();
				//no range moveセット
				killers=pushKiller(kchild,abd);
				assert( killers>0 );
				size=0;
				mother=(RecursiveNode*)0xffffffffffffffff;
			}
			
			template<class board_t,class node_t>
			void succeed(const board_t& abd,const node_t& anode,uint64 ahash){
				init();
				hash=ahash;
				range=anode.getRange()-1;
				turn=abd.getTurn();
				killers=pushKiller(kchild,abd);//killer moveセット
				assert( killers>0 );
				size=0;
				for(int c=0;c<killers;++c){//killer moveの評価継承
					for(int cc=0;cc<anode.killers;++cc){
						if( kchild[c].mv.type==anode.kchild[cc].mv.type ){
							kchild[c].size=anode.kchild[cc].size/3;
							kchild[c].eval_sum=anode.kchild[cc].eval_sum/3;
							kchild[c].eval2_sum=anode.kchild[cc].eval2_sum/3;
							
							//kchild[c].init();
							size+=kchild[c].size;
							break;
						}
					}
				}
				mother=&anode;
			}
			
			void setNull(int t){
				//ターンtでの空場を設定
				init();
				hash=(uint64)t;
				range=0;
				turn=t;
				killers=pushNullKiller(kchild,t);
				assert( killers>0 );
				size=0;
				mother=(RecursiveNode*)0xffffffffffffffff;
			}

			int getRange()const{return range;}
			
			void feed_killer_eval(const KillerMove& akill,fpn_t ev){
				//キラー手探索
				for(int c=0;c<killers;++c){
					if( kchild[c].mv.type == akill.type ){//同じキラー手を見つけた
						kchild[c].feed_eval(ev);
						size+=1.0;
						break;
					}
					//assert( c<killers-1 );
				}
			}
			
			void lock(){
#ifdef MULTI_THREADING
				::lock(_lock);
#endif
			}
			void unlock(){
#ifdef MULTI_THREADING
				::unlock(_lock);
#endif
			}
			
		};

		ostream& operator<<(ostream& out, const RecursiveNode& arg){
			out<<"(t"<<arg.turn<<",r"<<arg.range<<",m"
			<<((((uint64)arg.mother)==0xffffffffffffffff)?0:((uint64)arg.mother))<<")"<<" : ";
			for(int c=0;c<arg.killers;++c){
				out<<arg.kchild[c].size<<"("<<arg.kchild[c].eval_sum/(arg.kchild[c].size+0.0001)<<") ";
			}
			out<<endl;
			return out;
		}
		
		template<class page_t,int SIZE>
		struct BoardTranspositionTable{//局面置換表
			//開番地法
		private:
			
			const static uint64 DELETED=0xffffffffffffffff;
			const static int REHASH_MAX=5;
			const static int REAL_SIZE=SIZE+8;
			
			int convHash_Index(const uint64 hash){
				return ((uint32)(hash)) % SIZE;
			}
			int procIndex(const int index){
				return index+1;
			}
			
		public:
			page_t page[REAL_SIZE];
			
			uint32 npages;
			
			bool is_over95()const{
				return ( npages > (REAL_SIZE*95)/100 );
			}
			
			void init(){
				memset(page,0,REAL_SIZE*sizeof(page_t));
				
				//Null局面を0~15にセットする
				for(int t=0;t<16;++t){
					page[t].setNull(t);
				}
				npages=16;
			}
			page_t *read(const uint64 hash,bool *is_registed){
				//ポインタを返す
				int index=convHash_Index(hash);
				int trials=0;
				page_t *p=&page[index];
				while(1){
					if( !p->hash ){
						*is_registed=false;
						if( !is_over95() ){
							++npages;
							return p;//登録無し
						}else{//オーバー
							return nullptr;
						}
					}else if( p->hash == DELETED ){//DELETED発見
					}else if( hash == p->hash ){//発見
						*is_registed=true;
						return p;
					}
					if( trials >= REHASH_MAX-1 ){
						*is_registed=false;
						break;
					}
					++trials;
					++p;
				}
				return nullptr;
			}
			
			void print()const{
				for(int i=0;i<REAL_SIZE;++i){
					if( page[i].hash && page[i].hash!=DELETED && page[i].size>100 && page[i].getRange()==4 ){
						CERR<<"node "<<i<<" "<<page[i]<<endl;
					}
				}
			}
		};
		
		BoardTranspositionTable<RecursiveNode,(1<<20)+3 >  bdTT;
		
		struct KillerResult{
			KillerMove km;
			fpn_t ev;
		};
		
		template<class board_t,class node_t>
		KillerResult dive_try_node(board_t& bd,node_t& node,int tdepth,int pdepth,ThreadTools *const tools);
		
		template<class board_t,class move_t>
		fpn_t do_uct(board_t& bd,const move_t& mv,int tdepth,int pdepth,ThreadTools *const tools){
			//与えられた局面と着手からUCTを行う
			int changed;
			const int owner_col=bd.getTurnColor();
			const int end=bd.getEnd();
			changed=bd.makeMove(mv,&tools->ddice);//mvをプレーしてみる
			
			//修正用評価点を計算
			//ルートノードからの直結のみ
			fpn_t ev=(tdepth==TDEPTH_MAX)?eval(bd):0;
			
			if( bd.getTurn()==TURN_LAST || pdepth<=0 ){
				//プレイアウト終了
			}else{//depthが0になるまで
				int pto_turn=cmax(TURN_LAST,bd.getTurn()-pdepth);
				do{
					//局面更新(シミュレーション結果に応じて)
					bd.procTurn(changed);
					
					//自明な局面では評価を返す
					if( end==END_LAST ){
						//最終エンド
						//この時点で必勝決定の場合には、その結果を返す
						int mate=judgeMateStrictly(bd);
						if( mate!=-1 ){
							//勝ち確実な時に適当な手が選ばれにくいように、MATE判定のタイミングが早いほうがいい点が返るようにする
							const fpn_t MATE_TIMING_SCORE = 0.004 * (bd.getTurn() + 1);
							if( mate==BLACK ){
								if( owner_col==BLACK ){
									return ev+MATE_TIMING_SCORE+getTmpEndReward(BLACK,+8);
								}else{
									return ev-MATE_TIMING_SCORE-getTmpEndReward(BLACK,+8);
								}
							}else{
								if( owner_col==WHITE ){
									return ev+MATE_TIMING_SCORE+getTmpEndReward(WHITE,-8);
								}else{
									return ev-MATE_TIMING_SCORE-getTmpEndReward(WHITE,-8);
								}
							}
						}
					}
					if( bd.getTurn()==TURN_LAST ){
						//ラストストーン
						//分かりやすい局面では、シミュレーションを行わずに報酬を確定させる
						if( bd.getEndScore()==0 ){//ハウス内に石がない
							DERR<<"0-1,";
							//白が0か1か選ぶ事が出来る
							if( end>=END_JOBAN ){
								//序盤は0優先
								return ev+getTmpEndReward(owner_col,0);
							}else{
								//0と1の良い方を選ぶ
								if( getTmpEndReward(WHITE,0) > getTmpEndReward(WHITE,-1) ){
									return ev+getTmpEndReward(owner_col,0);
								}else{
									return ev+getTmpEndReward(owner_col,-1);
								}
							}
						}
						
						if( bd.NInHouse[BLACK]==0 ){//ハウス内に黒石がない
							if( bd.NInHouse[WHITE]==0 ){//ハウス内に白石がない
								
							}else{//ハウス内に白石がある
								//白は必ずドロー
								//ガードが邪魔でドロー出来ないことは考慮していない
								DERR<<"easydraw"<<bd.NInHouse[WHITE]<<endl;
								const int sc=-(bd.NInHouse[WHITE]+1);
								return ev+getTmpEndReward(owner_col,sc);
							}
						}
					}
					
					assert( bd.getTurn()>=TURN_LAST );
					
					int col=bd.getTurnColor();
					
					if( tdepth>0 ){
						//ノード登録を探す
						if( bd.isNull() ){
							//空場なら局面認識の必要なし
							ASSERT( bdTT.page[bd.getTurn()].killers>0 ,
								   for(int t=0;t<16;++t){
									   CERR<<t<<","<<bdTT.page[t].killers<<endl;
								   }
								   );
							auto chal=dive_try_node<1>(bd,bdTT.page[bd.getTurn()],tdepth-1,pdepth-1,tools);
							if( owner_col==col ){
								return ev+chal.ev;
							}else{
								return ev-chal.ev;
							}
						}else{
							const int RANGE_MAX=11;
							for(int r=RANGE_MAX;r>=0;--r){
								uint64 node_hash=bd.getHash(r);
								bool f;
								auto *const node=bdTT.read(node_hash,&f);
								if( node!=nullptr ){
									//ノード登録可能
									if( !f ){
										//新規登録
										node->set(bd,r,node_hash);
									}else{
										//CERR<<std::hex<<node_hash<<std::dec<<"already made("<<r<<")"<<endl;
									}
									assert( node!=nullptr );
									ASSERT( node->killers>0, cerr<<node->killers<<endl<<*node<<endl; );
									
									//このノードから潜る
									auto chal=dive_try_node<0>(bd,*node,tdepth-1,pdepth-1,tools);
									if( owner_col==col ){
										return ev+chal.ev;
									}else{
										return ev-chal.ev;
									}
								}
							}
							//置換表参照が不可能だったので、1手通常のプレイアウトとする
						}
					}
					
					
					int ops_col=flipColor(col);
					
					fMoveXY<> mv;
					
					//通常プレイアウトの手を選ぶ
					if( bd.NInHouse[ops_col]>0 ){
						//相手のNo1ストーンを叩く
						const auto& opsNo1=bd.stone[ops_col][0];
						genMiddleHit(opsNo1,&mv);
						DERR<<"HIT "<<opsNo1<<endl;
					}else{
						//センターにドロー
						mv.setSpin(tools->dice.rand()%2);
						genTeeShot(&mv);
					}
					DERR<<col<<" "<<mv<<endl;
					changed=bd.makeMove(mv,&tools->ddice);//mvをプレーしてみる
					
				}while( bd.getTurn()>pto_turn );
			}
			ev+=bd.getEval(owner_col);
			return ev;
		}
	
		template<int IS_NF=0,class board_t,class node_t>
		KillerResult dive_try_node(board_t& bd,node_t& node,int tdepth,int pdepth,ThreadTools *const tools){
			//与えられた候補局面から潜ってプレイアウトを行い、
			//遡って局面評価と着手評価を更新する
			//typedef typename node_t::child_t child_t;
			//typedef typename child_t::move_t move_t;
			
			DERR<<"dive from node "<<&node<<" ("<<node.size<<")"<<endl;
			
			auto *const dice=&tools->ddice;
			
			node.unlock();
			
			if( !IS_NF ){
				if( node.size > static_cast<fpn_t>(N_DIV_TRIALS_LEAFNODE)
				   && node.getRange()>4
				   ){//展開閾値を超えていて、レンジが一定以上
					
					//さらに詳しく展開
					
					
					int next_range=node.getRange()-1;
					uint64 next_hash=bd.getHash( next_range );//レンジを1つ小さくした時のハッシュ値
					
					bool f;
					node_t *pnext=bdTT.read(next_hash,&f);
					if( !f ){//ノードが確保されていない
						if( pnext==nullptr ){//確保出来ない
							goto DO_DIVECHILD;
						}
						//確保出来た
						pnext->succeed(bd,node,next_hash);
					}else{
						if( pnext->getRange() != next_range ){
							//レンジ違いのため別局面と思われる
							//無限ループ回避のため抜ける
							goto DO_DIVECHILD;
						}
					}
					assert( pnext!=nullptr );
					ASSERT( pnext->killers>0, cerr<<pnext->killers<<endl<<*pnext<<endl; );
					
					//次の局面へ
					auto chal=dive_try_node<0>(bd,*pnext,tdepth,pdepth,tools);
					
					node.lock();
					node.feed_killer_eval(chal.km,chal.ev);
					node.unlock();
					return chal;
				}
			}
			
		DO_DIVECHILD:
			//ここで着手の潜りに入る
			//分岐を選ぶ
			int best=0;
			fpn_t best_ucb=node.kchild[0].size>4 ?
			calc_ucb1(node.kchild[0].eval_sum,node.kchild[0].size,node.size,0.25):
			fpn_t(1000000-node.kchild[0].size*8192 + dice->rand()%1024);
			
			for(int d=1;d<node.killers;++d){
				fpn_t ucb=node.kchild[d].size>4 ? 
				calc_ucb1(node.kchild[d].eval_sum,node.kchild[d].size,node.size,0.25):
				fpn_t(1000000-node.kchild[d].size*8192 + dice->rand()%1024);
				
				//CERR<<node.kchild[d].eval_sum/(node.kchild[d].size+0.000001)<<" "<<ucb<<" "<<node.kchild[d].size<<endl;
				
				if( ucb>best_ucb ){
					best=d;
					best_ucb=ucb;
				}
			}
			//CERR<<endl;
			
			ASSERT( 0<=best && best<node.killers,
				   CERR<<best<<", "<<node.killers<<endl;
				   );
			
			//キラー手を実着手形式で生成
			KillerMove kmv=node.kchild[best].mv;
			fMoveXY<> mv;
			genKiller(bd,kmv,&mv);
			fpn_t ev=do_uct(bd,mv,tdepth,pdepth,tools);
			
			KillerResult kr;
			
			kr.km=kmv;
			kr.ev=ev;
			
			node.lock();
			node.feed_killer_eval(kmv,ev);
			node.unlock();
			
			return kr;
		}
	}
	
}