//デジタルカーリング
//モンテカルロ関連
//ルートノードでの処理

namespace DigitalCurling{
	namespace Ayumu{
		
		constexpr int N_DIV_TRIALS_ROOTCHILD=8;//ルート着手分岐閾値
		constexpr fpn_t ATT_RATE_ROOTCHILD=0.995;//減衰率

		
		template<class _move_t,int _N_DIV>
		struct RecursiveDivChild{
			//再帰分岐着手データ
			
			typedef _move_t move_t;
			const static int N_DIV=_N_DIV;
			
			struct DivChild{
				RecursiveDivChild<move_t,N_DIV> *pnext;//分割先のポインタ
				float _size;
				float eval_sum;
				float eval2_sum;
			};
			
			
			uint64 hash;//着手ハッシュ
			DivChild div[N_DIV];
			float _size;
			move_t mv;
			
			int tmpBest;//現時点で最も多くトライした分岐先
			
#ifdef MULTI_THREADING
			volatile int _lock[1];//スピンロック
#endif
			void init(){
				for(int d=0;d<N_DIV;++d){
					div[d]._size=1;
					div[d].eval_sum=0;
					div[d].eval2_sum=0.25;
					div[d].pnext=nullptr;
				}
				_size=N_DIV*1;
				tmpBest=0;
				unlock();
			}
			
			move_t dive_largest()const{
				//評価の大きさを基準に、再帰分割の先端まで到達させて着手を返す
				if( mv.getRange() <= 1 || _size == N_DIV * 1 ){
					//まだこの手でトライしていない
					return mv;
				}else{
					int best=tmpBest;fpn_t best_size=size(best);
					if( div[best].pnext != nullptr ){//展開されている
						return div[best].pnext->dive_largest();
					}else{
						return mv.part(N_DIV,best);
					}
				}
			}
			
			void set(const move_t& amv,uint64 ahash){
				init();
				mv=amv;
				hash=ahash;
			}
			
			move_t getMove()const{return mv;}
			fpn_t size_all()const{return _size;}
			fpn_t size(int d)const{return div[d]._size;}
			fpn_t var(int d)const{
				fpn_t m=mean(d);
				return div[d].eval2_sum/size(d) - m*m + static_cast<fpn_t>(0.0000001);
			}
			fpn_t mean(int d)const{assert(size(d)>0);return div[d].eval_sum/size(d);}
			void feed_eval(int d,fpn_t eval){
				div[d]._size*=ATT_RATE_ROOTCHILD;
				div[d].eval_sum*=ATT_RATE_ROOTCHILD;
				div[d].eval2_sum*=ATT_RATE_ROOTCHILD;
				
				div[d]._size+=1.0;
				div[d].eval_sum+=eval;
				div[d].eval2_sum+=eval*eval;
				
				_size*=sqrt(sqrt(ATT_RATE_ROOTCHILD));
				_size+=1.0;
				
				if( mean(d) > mean(tmpBest) ){
					tmpBest=d;
				}
			}
			
			void dive_print(int depth,fpn_t max_depth)const{
				//末端までの木構造を出力
				CERR<<Space(depth*2)<<mv<<endl;
				if( depth+1 >= max_depth ){
					const double rew=mean(tmpBest) * 10000 ;
					const double sem=sqrt(var(tmpBest)) * 10000;
					const double rewZone[2]={rew-sem , rew+sem};
					CERR<<" : "<<(int)rew<<" ( "<<(int)rewZone[0]<<" ~ "<<(int)rewZone[1]<<" ) ";
					CERR<<size(tmpBest)<<" trials."<<endl;
				}else{
					for(int best=0;best<N_DIV;++best){
						
						//CERR<<Space(depth*2+1)<<mv.part(N_DIV,best);
						
						if( size(best)>0 ){
							const double rew=mean(best) * 10000 ;
							const double sem=sqrt(var(best)) * 10000;
							const double rewZone[2]={rew-sem , rew+sem};
							
							//まず調整済み評価点を表示
							CERR<<" : "<<(int)rew<<" ( "<<(int)rewZone[0]<<" ~ "<<(int)rewZone[1]<<" ) ";
							//プレイアウトの結果を表示。これらの値は未調整。
							CERR<<size(best)<<" trials."<<endl;
							
							if( size(best) >= static_cast<fpn_t>(N_DIV_TRIALS_ROOTCHILD) 
							   && div[best].pnext != nullptr
							   && depth<max_depth
							   ){//展開されている
								div[best].pnext->dive_print(depth+1,max_depth);
							}
						}else{
							CERR<<" no evaluation."<<endl;
						}
					}
				}
			}
			
			//void feed_eval_rec(const move_t& mv,fpn_t eval){
				//再帰分割の先端まで評価を加算する
				//feed_eval()
			//}
			
			void lock(){
#ifdef MULTI_THREADING
				::lock(_lock);
#endif
			}
			void unlock(){
#ifdef MULTI_THREADING
				::unlock(_lock);
#endif
			}
			
		};
		
		template<class page_t,int SIZE>
		struct MoveTranspositionTable{//着手置換表
			//開番地法
		private:
			
			const static uint64 DELETED=0xffffffffffffffff;
			const static int REHASH_MAX=5;
			const static int REAL_SIZE=SIZE+8;
			
			int convHash_Index(const uint64 hash){
				return ((uint32)(hash)) % SIZE;
			}
			int procIndex(const int index){
				return index+1;
			}
			
		public:
			page_t page[REAL_SIZE];
			
			uint32 npages;
			
			bool is_over95()const{
				return ( npages > (REAL_SIZE*95)/100 );
			}
			
			void init(){
				memset(page, 0, REAL_SIZE*sizeof(page_t));
				npages = 0;
			}

			page_t *read(const uint64 hash,bool *is_registed){
				//ポインタを返す
				int index=convHash_Index(hash);
				int trials=0;
				page_t *p=&page[index];
				while(1){
					if( !p->hash ){
						*is_registed=false;
						if( !is_over95() ){
							++npages;
							return p;//登録無し
						}else{//オーバー
							return nullptr;
						}
					}else if( p->hash == DELETED ){//DELETED発見
					}else if( hash == p->hash ){//発見
						*is_registed=true;
						return p;
					}
					if( trials >= REHASH_MAX-1 ){
						*is_registed=false;
						break;
					}
					++trials;
					++p;
				}
				return nullptr;
			}
		};
		
		MoveTranspositionTable<RecursiveDivChild<MoveXY,4>,(1<<17) >  mvTT;
		
		template<class move_t>
		struct RootNode{
			//モンテカルロ法着手決定するルートノード
			
			constexpr static int N_MAX_CHILDS=64;
			
			RecursiveDivChild<move_t,4> child[N_MAX_CHILDS];//候補着手
			int childs;
			std::atomic<uint32> trials;
			std::atomic<int> ex;
			uint64 limit_time;
			
#ifdef MULTI_THREADING
			volatile int _lock[1];//spin lock
#endif
			
			RootNode(){}
			
			const static uint32 MIN_TRIALS_ROOT=16;
			
			void lock(){
#ifdef MULTI_THREADING
				::lock(_lock);
#endif
			}
			void unlock(){
#ifdef MULTI_THREADING
				::unlock(_lock);
#endif
			}
			
			void init(){
				unlock();
				
				ex=0;
				trials=0;
				childs=0;
			}
			
			template<class board_t,class rmove_t>
			void addChild(const board_t& bd,const rmove_t& mv){
				assert( 0<=childs && childs<N_MAX_CHILDS );
				child[childs].set( mv, (bd.isNull()?0ULL:bd.getHash(0))^genHash_Move(mv) );
				++childs;
			}
			
			void set_limit_time(uint64 alimit_time){
				limit_time=alimit_time;
			}
			
			template<class dice_t>
			int choose_root_UCB(dice_t *const dice)const{
				//検討する着手を選ぶ
				
				int bestId=-1;
				
				switch(childs){
					case 0:assert(0);break;
					case 1:bestId=0;break;
					case 2:
						//2つの時は同数(分布サイズ単位)に割り振る
						if( child[0].size_all() < child[1].size_all() ){
							bestId=0;
						}else if( child[0].size_all() > child[1].size_all() ){
							bestId=1;
						}else{
							bestId=dice->rand()%2;
						}
						break;
					default:
						fpn_t bestScore=-99.0;
						fpn_t allSize=trials;
						fpn_t logAS=log(allSize);
						for(int c=0;c<childs;++c){
							int best=child[c].tmpBest;
							fpn_t tmpScore;
							fpn_t size=child[c].size(best);
							if( child[c].size_all() < MIN_TRIALS_ROOT ){
								//最低プレイアウト数をこなしていないものは、大きな値にする
								//ただし最低回数のもののうちどれかがランダムに選ばれるようにする
								tmpScore=(fpn_t)((1U<<16) - ((uint32)child[c].size_all()) + (dice->rand()%(1U<<6)));
							}else{
								assert( size );
								fpn_t ucb=child[c].mean(best) + sqrt(logAS) * exp(5-size*0.07);
								
								tmpScore=ucb;
							}
							if(tmpScore > bestScore){
								bestScore=tmpScore;
								bestId=c;
							}
						}
						break;
				}
				return bestId;
			}
			
			void feed_eval(int i,fpn_t eval){
				child[i].feed_eval(eval);
				++trials;
			}
			
			int choose_best_id()const{
				//最高評価のものを選ぶ
				int best_id=0;
				fpn_t best_score=-99999.0;
				for(int i=0;i<childs;++i){
					int best=child[i].tmpBest;
					fpn_t tmp;
					if( child[i].size(best)>0 ){
						tmp=child[i].mean(best);
					}else{
						tmp=-99999.0;
					}
					if(tmp>best_score){
						best_id=i;
						best_score=tmp;
					}
				}
				return best_id;
			}
			
			//void sort()const{
			//現時点での評価が高い順に並べ替える
			//	std::sort(child,child+nchilds,RootChild::GreaterMean);
			//}
			
#ifdef MONITOR
			void print()const{
				//モンテカルロ解析結果表示
				for(int c=0;c<childs;++c){
					child[c].dive_print(0,0.5);
				}
			}
#endif //MONITOR
		};
		/*
		template<class board_t,class move_t>
		fpn_t do_haetataki(board_t& bd,const move_t& mv,ThreadTools *const tools){
			//与えられた局面と着手からプレイアウトを行う
			//ハエ叩きモードなのでそのまま評価点を出す
			const int owner_col=bd.getTurnColor();
			changed=bd.makeMove(mv,&tools->ddice);//mvをプレーしてみる
			int sc=bd.getTmpScore();
			if( owner_col==BLACK ){
				
		}
		*/
		template<class board_t,class child_t>
		fpn_t dive_try_child(board_t& bd,child_t& c,int cdepth,ThreadTools *const tools){
			//与えられた候補着手から潜ってプレイアウトを行い、
			//遡って局面評価と着手評価を更新する
			typedef typename child_t::move_t move_t;
			
			auto *const dice=&tools->ddice;
			
			const int col=bd.getTurnColor();
			
			c.unlock();
			
			//分岐を選ぶ
			int best=0;
			fpn_t best_ucb=-9999;
			for(int d=0;d<child_t::N_DIV;++d){
				fpn_t ucb=c.size(d)>3 ? 
				calc_ucb1(c.div[d].eval_sum,c.size(d),c.size_all(),2.5)
				//(c.div[d].eval_sum/c.size(d) + sqrt(log(c.size_all())) * exp(13-c.size(d)*0.07))
				:
				fpn_t(1000000-c.size(d)*8192 + dice->rand()%1024);
				if( ucb>best_ucb ){
					best=d;
					best_ucb=ucb;
				}
			}
			assert( 0<=best && best<child_t::N_DIV );
			
			if( c.size(best)>N_DIV_TRIALS_ROOTCHILD 
			   && c.getMove().getRange()>2
			   && cdepth>0
			   ){//展開閾値を超えていて、レンジが3以上、着手展開深さ限度以上
				
				//dで展開
				child_t *pnext=c.div[best].pnext;
				if( pnext == nullptr ){//次の場所がまだ確保されていない
					//ハッシュ値を計算
					move_t nextmv=c.mv.part(child_t::N_DIV,best);
					//uint64 hash=bd.getHash(0)^genHash_Move(nextmv);
					//uint64 next_hash=c.hash;
					//subtrHash_VRange( &next_hash, c.getMove().getRange() );//差分計算
					//CERR<<next_hash<<" "<<((bd.isNull()?0ULL:bd.getHash(0))^genHash_Move(nextmv));getchar();
					
					uint64 next_hash=(bd.isNull()?0ULL:bd.getHash(0))^genHash_Move(nextmv);
					
					assert( next_hash == (bd.isNull()?0ULL:bd.getHash(0))^genHash_Move(nextmv) );
					
					//場所を確保
					bool f;
					pnext=mvTT.read(next_hash,&f);
					//assert( !f );
					
					if( pnext == nullptr ){//確保出来ない
						goto DO_PO;
					}
					c.div[best].pnext=pnext;
					pnext->set(nextmv,next_hash);
				}
				assert( pnext!=nullptr );
				
				fpn_t ev=dive_try_child(bd,*pnext,cdepth-1,tools);
				
				c.lock();
				c.feed_eval(best,ev);
				c.unlock();
				return ev;
			}
		DO_PO:
			//ここでプレイアウトに入って、評価を得る
			//ノーレンジな着手を1つ生成
			move_t mv=c.getMove().part(child_t::N_DIV,best).genNR(dice);
			
			fMoveXY<> fmv;
			convMove(mv,&fmv);
			//CERR<<"playout move : "<<fmv;
			
			fpn_t ev;
			//if( mcf->isHaeTataki() ){
			//	ev=do_haetataki(bd,mv,tools);
			//}else{
				ev=do_uct(bd,mv,3,1024,tools);
			//}
			//CERR<<" result : "<<ev<<endl;
			
			c.lock();
			c.feed_eval(best,ev);
			c.unlock();
			return ev;
		}
		
		template<class root_t>
		int mc_thread(const int thread_id,root_t *const mcf,const ThinkBoard *const bd){
			
			//モンテカルロスレッド
			//マルチスレッディングにしない場合は関数として開く
			
			ClockMS clms;//時計(millisec単位)
			clms.start();
			
			//グローバルに置いてあるこのスレッド用の道具
			auto& tools=thread_tools[thread_id];
			
			ThinkBoard tb;//プレイアウタ空間
			
			int thread_nplayouts=0;//このスレッドのプレイアウト回数
			
			while( !mcf->ex ){//終了判定が出ていない
				
				for(int cnt=0;cnt<16;cnt++){
					DERR<<"playout "<<thread_nplayouts<<".";
					
					int i=mcf->choose_root_UCB(&tools.ddice);//検討する着手を選ぶ
					
					memcpy(&tb,bd,sizeof(ThinkBoard));
					
					dive_try_child(tb,mcf->child[i],1024,&tools);
					mcf->trials++;//atomic命令
					++thread_nplayouts;
				}
				
				if( clms.stop() > mcf->limit_time ){//制限時間オーバー
					mcf->ex=1;//atomic命令
					return 0;
				}
			}
			
			//終了判定が出た
			return 0;
		}
	}
	

}