//デジタルカーリング
//モンテカルロ関連

namespace DigitalCurling{
	namespace Ayumu{
		struct ThinkBoard{
			//思考用の局面表現	
			
			int end;
			int turn;
			int rscore;//このエンドの先手から見た累積得点の差
			int tscore;//現在の石の配置で終了した場合の、このエンドのみの得点
			
			uint64 hash_stones[12];//局面認識ハッシュ値の石成分(レンジ幅分)
			
			//石の位置、先手後手それぞれハウスに近い順
			//最後には番兵として無限遠にある石を置いておく(rを大きい値にするだけ) <-今はやってない
			fPosXYRT<> stone[2][8];
			int NInHouse[2];//ハウス内の個数
			int NActive[2];//有効エリア内の個数
			//int NFG[2];//フリーガードゾーン内の個数
			//int fg_id[2][2];//フリーガードゾーン内の石のid(stoneのインデックス)
			StoneState state[2][8];//石の状態
			
			//局面更新補助変数
			int pushedLast;//最後にpushされた石のid
			
			//シミュレート補助変数
			fpn_t frontY;//もっとも手前側にある石のY座標
			
			void init(){
				for(int c=0;c<2;++c){
					NActive[c]=0;
					NInHouse[c]=0;
					//NFG[c]=0;
					//fg_id[c][0]=fg_id[c][1]=0;
				}
				memset( hash_stones, 0, sizeof(hash_stones));
				memset( state, 0, sizeof(state));
				end=0;turn=0;rscore=0;tscore=0;
				frontY=999;
				pushedLast=-1;
				for(int r=0;r<12;++r){
					hash_stones[r]=0ULL;
				}
			}
			
			void setState(){
				//石の状態を一括設定
				if( NActive[BLACK]+NActive[WHITE] != 0 ){
					
					//状態を設定
					fpn_t front=FY_PA_MAX + F_STONE_RAD;
					for(int i=0;i<NActive[BLACK];++i){
						state[BLACK][i].init();
						state[BLACK][i].setBlack();
						
						const auto& st=stone[BLACK][i];
						if( st.getY() < front ){
							front=st.getY();
						}
						
						addHash_BXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
					}
					for(int i=0;i<NActive[WHITE];++i){
						state[WHITE][i].init();
						state[WHITE][i].setWhite();
						
						const auto& st=stone[WHITE][i];
						if( st.getY() < front ){
							front=st.getY();
						}
						addHash_WXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
					}
					frontY=front;
					
					if( isFreeGuardTurn(turn) ){
						//フリーガードの石を設定
						int c=getTurnColor();
						int oc=flipColor( c );
						for(int i=0;i<NActive[c];++i){
							if( isInFreeGuardZone(stone[c][i]) ){
								state[c][i].setInFreeGuardZone();
								//fg_id[c][NFG[c]]=i;
								//++NFG[c];
							}
						}
						for(int i=0;i<NActive[oc];++i){
							if( isInFreeGuardZone(stone[oc][i]) ){
								state[oc][i].setFreeGuard();
								state[oc][i].setInFreeGuardZone();
								//fg_id[oc][NFG[oc]]=i;
								//++NFG[oc];
							}
						}
					}
					
					tscore=countEndScore();//仮得点
				}else{
					tscore=0;//仮得点
				}
				assert( exam(3) );
			}
			
			void set(const ClientField& cf){
				//ClientFieldの情報(公式のデータ型に近い)を
				//自分の思考用のデータに変換する
				init();
				
				end=cf.to_end - 1 -cf.end;//エンド数は最終を0とする
				turn=cf.to_turn - 1 -cf.turn;//ターン数は最終を0とする
				rscore=(cf.color[0]==0)?cf.rel_score:(-cf.rel_score);
				
				//先手の石
				for(int i=0;i<8;++i){
					fPosXY<> mypos=convPos_Official_Mine( fPosXY<>(cf.pos[i][0][0],cf.pos[i][0][1]) );
					if( isInPlayArea( mypos ) ){
						pushStone(BLACK, mypos.getX(), mypos.getY() );
					}
				}
				//後手の石
				for(int i=0;i<8;++i){
					fPosXY<> mypos=convPos_Official_Mine( fPosXY<>(cf.pos[i][1][0],cf.pos[i][1][1]) );
					if( isInPlayArea( mypos ) ){
						pushStone(WHITE, mypos.getX(), mypos.getY() );
					}
				}
				
				setState();
			}
			
			template<int EXACT=0,class dice_t>
			int makeMove(const MoveXY& mv,dice_t *const adice){
				//石を投げて、盤面を更新する(離散直交ベクトル)
				
				fMoveXY<> ShotVec;
				ShotVec.x = VXtoFVX(mv.getVX()); ShotVec.y = VYtoFVY(mv.getVY()); ShotVec.s = mv.getSpin();
				
				int ret= makeMove( ShotVec, adice );
				
				return ret;
			}
			
			template<int EXACT=0,class dice_t>
			int makeMove(const fMoveXY<>& mv,dice_t *const adice){
				//石を投げて、盤面を更新する(連続直交ベクトル)
				
				int ret;
				//乱数をかける
				fMoveXY<> tmp=mv;
				addRandToMove(&tmp,adice);
				
				//CERR<<tmp<<endl;
				
				if( NActive[0]+NActive[1]==0 ){
					//完全フリー
					if( EXACT ){
						ret=Simulator::simulate( (ThinkBoard*)this, tmp );
					}else{
						static StaticClock c;
						c.start();
						ret=FastSimulator::simulateSoloF( (ThinkBoard*)this, tmp );
						c.stop();
					}
				}else{
					if( EXACT ){
						ret=Simulator::simulate( (ThinkBoard*)this, tmp );
					}else{
						static StaticClock c;
						c.start();
						ret=FastSimulator::simulateF( (ThinkBoard*)this, tmp );
						c.stop();
					}
				}
				
				/*for(int c=0;c<2;++c){
				 CERR<<OutColor(c);
				 for(int i=0;i<NActive[c];++i){
				 CERR<<" ("<<stone[c][i].getX()<<","<<stone[c][i].getY()<<")";
				 }
				 CERR<<endl;
				 }*/
				
				
				
				return ret;
			}
			
			
			void procTurn(int changed=2){
				//情報を1ターン進める
				
				assert( turn>TURN_LAST );//すでにエンドが終了した場合は入らないように
				--turn;
				
				switch( changed ){
						//石配置の変化の度合いによって処理を分ける
					case 0:{
						//局面変化なし
						DERR<<"uc."<<endl;
						if( turn >= TURN_FREEGUARD-1 ){
							if( turn == TURN_FREEGUARD-1 ){
								//フリーガード終了
								int c=getTurnColor();
								/*for(int i=0;i<NActive[c];++i){
									state[c][fg_id[c][i]].resetFreeGuard();
								}
								NFG[0]=NFG[1]=0;*/
								for(int i=0;i<NActive[c];++i){
									state[c][i].resetFreeGuard();
								}
							}else{
								//フリーガード反転
								int c=getTurnColor();
								int oc=flipColor( c );
								/*for(int i=0;i<NFG[c];++i){
									state[c][fg_id[c][i]].resetFreeGuard();
								}
								for(int i=0;i<NFG[oc];++i){
									state[oc][fg_id[oc][i]].setFreeGuard();
								}*/
								for(int i=0;i<NActive[c];++i){
									state[c][i].resetFreeGuard();
								}
								for(int i=0;i<NActive[oc];++i){
									state[oc][i]|=((uint32)state[oc][i] & (1U<<2))<<1;
								}
							}
						}
						//それ以外の操作は必要なし
					}break;
					case 1:
					{
						//石を置いただけ
						DERR<<"put only."<<endl;
						
						int c=getTurnColor();
						int oc=flipColor( c );
						int nn=pushedLast;
						
						state[oc][nn].init();
						state[oc][nn].setColor(oc);
						
						if( turn >= TURN_FREEGUARD-1 ){
							if( turn == TURN_FREEGUARD-1 ){
								//フリーガード終了
								/*for(int i=0;i<NFG[c];++i){
									state[c][fg_id[c][i]].resetFreeGuard();
								}
								NFG[0]=NFG[1]=0;*/
								for(int i=0;i<NActive[c];++i){
									state[c][i].resetFreeGuard();
								}
							}else{
								//フリーガード反転
								
								/*for(int i=0;i<NFG[c];++i){
									state[c][fg_id[c][i]].resetFreeGuard();
								}
								for(int i=0;i<NFG[oc];++i){
									state[oc][fg_id[oc][i]].setFreeGuard();
								}*/
								if( isInFreeGuardZone(stone[oc][nn]) ){
									//fg_id[nc][NFG[nc]]=nn;
									//++NFG[nc];
									state[oc][nn].setInFreeGuardZone();
								}
								
								for(int i=0;i<NActive[c];++i){
									state[c][i].resetFreeGuard();
								}
								for(int i=0;i<NActive[oc];++i){
									state[oc][i]|=((uint32)state[oc][i] & (1U<<2))<<1;
								}

							}
						}
						if( NActive[BLACK]+NActive[WHITE]==1 ){
							for(int i=0;i<12;++i){hash_stones[i]=0ULL;}
						}
						if( oc==BLACK ){
							const auto& st=stone[BLACK][nn];
							addHash_BXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
							frontY=min(frontY,st.getY());
						}else{
							const auto& st=stone[WHITE][nn];
							addHash_WXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
							frontY=min(frontY,st.getY());
						}
						
						tscore=countEndScore();//仮得点
					}break;
					case 2:{
						//衝突によって複雑な変化あり
						//NFG[0]=NFG[1]=0;
						
						//石の状態を設定
						if( NActive[BLACK]+NActive[WHITE] != 0 ){
							
							for(int i=0;i<12;++i){hash_stones[i]=0ULL;}
							
							//状態を設定
							fpn_t front=FY_PA_MAX + F_STONE_RAD;
							for(int i=0;i<NActive[BLACK];++i){
								state[BLACK][i].init();
								state[BLACK][i].setBlack();
								const auto& st=stone[BLACK][i];
								if( st.getY() < front ){
									front=st.getY();
								}            
								addHash_BXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
							}
							for(int i=0;i<NActive[WHITE];++i){
								state[WHITE][i].init();
								state[WHITE][i].setWhite();
								const auto& st=stone[WHITE][i];
								if( st.getY() < front ){
									front=st.getY();
								}
								addHash_WXY(hash_stones,FXtoX( st.getX() ),FYtoY( st.getY() ),0,11);
							}
							/*
							 for(int i=0;i<12;++i){
							 CERR<<hash_stones[i]<<endl;
							 }
							 getchar();
							 */
							
							frontY=front;
							
							if( isFreeGuardTurn(turn) ){
								//フリーガードの石を設定
								int c=getTurnColor();
								int oc=flipColor( c );
								for(int i=0;i<NActive[c];++i){
									if( isInFreeGuardZone(stone[c][i]) ){
										//fg_id[c][NFG[c]]=i;
										//++NFG[c];
										state[c][i].setInFreeGuardZone();
									}
								}
								for(int i=0;i<NActive[oc];++i){
									if( isInFreeGuardZone(stone[oc][i]) ){
										state[oc][i].setInFreeGuardZone();
										state[oc][i].setFreeGuard();//相手の石にはフリーガードフラグ
										//fg_id[oc][NFG[oc]]=i;
										//++NFG[oc];
									}
								}
							}
							
							tscore=countEndScore();//仮得点
						}else{
							tscore=0;//仮得点
						}
					}break;
					default:assert(0);break;
				}
				
				assert( exam(changed) );
			}
			int getTurn()const{return turn;}
			int getEnd()const{return end;}
			int getRelScore()const{return rscore;}
			int getTurnColor()const{
				return TURNtoCOL(turn);
			}
			int getEndScore()const{return tscore;}
			const fPosXYRT<>& getStonePos(int col,int i)const{
				assert( examColor(col) );
				assert( 0<=i && i<8 );
				
				return stone[col][i];
			}
			uint64 getHash(int r)const{
				return hash_stones[r]^getTurn();
			}
			bool isNull()const{return !(NActive[0]+NActive[1]);}
			bool hasAdvantage(int col)const{
				return (col==BLACK)?(getRelScore()>0):(getRelScore()<0);
			}
			bool isTie()const{
				return (getRelScore()==0);
			}
			bool isFinalTurn()const{
				return ((getTurn()==TURN_LAST) && (getEnd()==END_LAST));
			}
			
			void pushStone(int col,fpn_t x,fpn_t y){
				//石を挿入ソートで置いていく
				assert( examColor(col) );
				
				fPosXYRT<> tmp(x,y);
				
				int st;
				if( isInHouse(tmp.getR()) ){//ハウス内
					st=NInHouse[col]-1;
					++NInHouse[col];
				}else{
					st=NActive[col]-1;
				}
				for(int i=st;i>=0;--i){
					if( tmp.getR() >= stone[col][i].getR() ){
						for(int j=NActive[col]-1;j>i;--j){
							stone[col][j+1]=stone[col][j];
						}
						stone[col][i+1]=tmp;
						goto END;
					}
				}
				for(int j=NActive[col]-1;j>=0;--j){
					stone[col][j+1]=stone[col][j];
				}
				stone[col][0]=tmp;
			END:
				++NActive[col];
				assert( exam_nstones(col,turn-1) );
			}
			void pushStoneWithState(int col,fpn_t x,fpn_t y){
				//石を挿入ソートで置いていく
				//状態も一緒に移動させる
				//pushedLastに移動させた石を記録
				assert( examColor(col) );
				
				//printStones();
				
				fPosXYRT<> tmp(x,y);
				
				int st;
				if( isInHouse(tmp.getR()) ){//ハウス内
					st=NInHouse[col]-1;
					++NInHouse[col];
				}else{
					st=NActive[col]-1;
				}
				for(int i=st;i>=0;--i){
					if( tmp.getR() >= stone[col][i].getR() ){
						for(int j=NActive[col]-1;j>i;--j){
							stone[col][j+1]=stone[col][j];
							state[col][j+1]=state[col][j];
						}
						stone[col][i+1]=tmp;
						pushedLast=i+1;
						goto END;
					}
				}
				for(int j=NActive[col]-1;j>=0;--j){
					stone[col][j+1]=stone[col][j];
					state[col][j+1]=state[col][j];
				}
				stone[col][0]=tmp;
				pushedLast=0;
			END:
				++NActive[col];
				//printStones();
				assert( exam_nstones(col,turn-1) );
			}
			void initStonePos(){
				NInHouse[0]=NInHouse[1]=0;
				NActive[0]=NActive[1]=0;
			}
			int countEndScore()const{
				//先手から見た得点を返す
				//CERR<<" ("<<NInHouse[BLACK]<<","<<NInHouse[WHITE]<<")";
				
				int sc;
				if( NInHouse[BLACK]==0 ){
					sc=-NInHouse[WHITE];
				}else if( NInHouse[WHITE]==0 ){
					sc=NInHouse[BLACK];
				}else{
					int i;
					fpn_t da,db;
					da=stone[BLACK][0].getR();db=stone[WHITE][0].getR();
					if( da > db ){//後手がNo1
						sc=-1;
						i=1;
						while(1){
							if( i==NInHouse[WHITE] ){break;}
							db=stone[WHITE][i].getR();
							if( da <= db ){break;}
							--sc;++i;
						}
					}else if( da < db ){//先手がNo1
						sc=1;
						i=1;
						while(1){
							if( i==NInHouse[BLACK] ){break;}
							da=stone[BLACK][i].getR();
							if( da >= db ){break;}
							++sc;++i;
						}
					}else{//同点
						sc=0;
					}
				}
				//CERR<<sc;
				return sc;
			}
			int get2ndScore(int col)const{
				assert( examColor(col) );
				//colの第2ストーン以内にあるops_colの石数+1を返す
				int sc;
				const int ops_col=flipColor(col);
				if( NInHouse[col]<=1 ){//第2ストーンまでハウス内にない
					sc=NInHouse[ops_col]+1;//相手のハウス内の全て+1
				}else{//第2ストーンがハウス内にある
					const fpn_t da=stone[col][2-1].getR();//第2ストーンの距離
					int i=0;
					for(;i<NInHouse[ops_col];++i){
						if( stone[ops_col][i].getR() > da ){break;}
					}
					sc=i+1;
				}
				return sc;
			}
			fpn_t getEval(int col){
				//colから見た局面評価点を返す
				fpn_t ev;
				if( turn==TURN_LAST ){//最終投終了後の場合、エンド末端評価
					int sc=countEndScore();
					ev=getTmpEndReward(col,sc);
				}else{//最終投でない
					//予測得点
					if( TURNtoCOL(turn) == BLACK ){//先手のターン
						int sc2=-get2ndScore(BLACK);//-(先手の第2ストーン内にある後手のストーン＋1)
						ev=getTmpEndReward(col,sc2);
					}else{
						//スコア
						int sc=countEndScore();
						ev=getTmpEndReward(col,sc);
					}
				}
				return ev;
			}
			bool isSymmetry()const{
				//左右対称かどうか
				if( NActive[0]+NActive[1]==0 ){
					return true;//石が無いので対称
				}
				//それ以外は未実装
				return false;
			}
			
			void save(const ThinkBoard& arg){
				//局面をargに戻す
			}
			bool exam_end()const{
				if( !(0<=end && end<MAX_NENDS ) ){
					cerr<<"ThinkBoard : illegal end."<<endl;
					return false;
				}
				return true;
			}
			bool exam_turn()const{
				if( !(0<=turn && turn<16) ){
					cerr<<"ThinkBoard : illegal turn."<<endl;
					return false;
				}
				return true;
			}
			bool exam_fg()const{
				if( isFreeGuardTurn(turn)  ){
					/*if( !(0<=NFG[0] && NFG[0]<=2 && 0<=NFG[1] && NFG[1]<=2) ){
						cerr<<"ThinkBoard : illegal the number of free-guard stones."<<endl;
						return false;
					}*/
					for(int c=0;c<2;++c){
						for(int n=0;n<NActive[c];++n){
							if( isInFreeGuardZone(stone[c][n]) ){
								/*for(int i=0;i<NFG[c];++i){
									if( fg_id[c][i]==n ){
										if( c==getTurnColor() ){
											if( state[c][n].isFreeGuard() ){
												cerr<<"ThinkBoard : turn player's free-guard stone("<<c<<n<<") has free-guard flag."<<endl;
												return false;
											}
										}else{
											if( !state[c][n].isFreeGuard() ){
												cerr<<"ThinkBoard : ops player's free-guard stone("<<c<<n<<") doesn't have free-guard flag."<<endl;
												return false;
											}
										}
										break;
									}
									if( i==NFG[c]-1 ){
										cerr<<"ThinkBoard : free-guard stone("<<c<<n<<") is not in free-guard list."<<endl;
										return false;
									}
								}*/
								
							}else{
								/*for(int i=0;i<NFG[c];++i){
									if( fg_id[c][i]==n ){
										cerr<<"ThinkBoard : non free-guard stone("<<c<<n<<") is in free-guard list."<<endl;
										return false;
									}
								}*/
							}
						}
					}
					return true;				
				}
				return true;//とりあえず適応外のターンの場合は不問としておく
			}
			bool exam_nstones(int col,int t)const{//石数
				if( !(0<=NActive[col] && NActive[col]<=(16-t-col)/2) ){//そのターン開始時点での最大個数以内
					cerr<<"ThinkBoard : illegal number of stones( "<<OutColor(col)<<NActive[col]<<"t"<<t<<" )."<<endl;
					return false;
				}
				return true;
			}
			bool exam_state()const{//石数
				for(int c=0;c<2;++c){
					for(int n=0;n<NActive[c];++n){
						if( c != state[c][n].getColor() ){
							cerr<<"ThinkBoard : illegal stone state( "<<OutColor(c)<<n<<" )."<<endl;
							return false;
						}
					}
				}
				return true;
			}
			bool exam_hash()const{
				//石が無い場合はハッシュ値は正しくなくていいので注意
				if( NActive[BLACK]+NActive[WHITE]>0 ){
					for(int r=0;r<12;++r){
						uint64 hash_tmp=0ULL;
						for(int n=0;n<NActive[BLACK];++n){
							hash_tmp^=genHash_BX(FXtoX(stone[BLACK][n].x),r);
							hash_tmp^=genHash_BY(FYtoY(stone[BLACK][n].y),r);
						}
						for(int n=0;n<NActive[WHITE];++n){
							hash_tmp^=genHash_WX(FXtoX(stone[WHITE][n].x),r);
							hash_tmp^=genHash_WY(FYtoY(stone[WHITE][n].y),r);
						}
						if( hash_stones[r] != hash_tmp ){
							cerr<<"ThinkBoard : illegal hash value( range "<<r<<" ";
							cerr<<std::hex<<hash_stones[r]<<" - "<<hash_tmp<<std::dec<<" ) ."<<endl;
							return false;
						}
					}
				}
				return true;
			}
			bool exam(int timing=3)const{
				//基本データ
				bool bRet=true;
				if( !exam_end() ){bRet=false;}
				if( !exam_turn() ){bRet=false;}
				if( !exam_fg() ){bRet=false;}
				//石数
				if( !exam_nstones(BLACK,turn) ){bRet=false;}
				if( !exam_nstones(WHITE,turn) ){bRet=false;}
				//石の状態
				if( !exam_state() ){bRet=false;}
				//ハッシュ値
				if( !exam_hash() ){bRet=false;}
				
				if( !bRet ){
					switch( timing ){
						case 0:cerr<<"( after throw )"<<endl;break;
						case 1:cerr<<"( after put )"<<pushedLast<<endl;break;
						case 2:cerr<<"( after set )"<<endl;break;
						case 3:cerr<<"( after first set )"<<endl;break;
						default:break;
					}
				}
				return bRet;
			}
			
			void printStones()const{
				CERR<<"B";
				for(int i=0;i<NActive[BLACK];++i){
					CERR<<stone[BLACK][i];
				}
				CERR<<"W";
				for(int i=0;i<NActive[WHITE];++i){
					CERR<<stone[WHITE][i];
				}
				CERR<<endl;
			}
		};
	}
}