//デジタルカーリング
//与えられた条件から、着手型を生成する

namespace DigitalCurling{

	/**************************ノーレンジ着手の生成**************************/
	
	template<class move_t>
	void genPASS(move_t *const mv){
		//パス
		mv->setVX(FVX_MIN);mv->setVY(FVY_MIN);
	}
	
	template<class pos_t,class move_t>
	void genDrawShot(const pos_t& pos,move_t *const mv){
		//ドローショット
		FastSimulator::FXYtoFMV(pos,mv);
	}
	
	template<class move_t>
	void genTeeShot(move_t *const mv){
		//ティーショット
		genDrawShot(fPosXY<>(FX_TEE,FY_TEE),mv);
	}
	
	template<class move_t>
	void genShortTeeShot(move_t *const mv){
		//ティー少し前ショット
		genDrawShot(fPosXY<>(FX_TEE,FY_TEE-0.9),mv);
	}
	
	template<class move_t>
	void genRightSideShot(move_t *const mv){
		//右サイドショット
		genDrawShot(fPosXY<>(FX_TEE + (F_HOUSE_RAD - 2*F_STONE_RAD),FY_TEE),mv);
	}
	
	template<class move_t>
	void genLeftSideShot(move_t *const mv){
		//左サイドショット
		genDrawShot(fPosXY<>(FX_TEE - (F_HOUSE_RAD - 2*F_STONE_RAD),FY_TEE),mv);
	}
	
	template<class move_t>
	void genCenterGuard(move_t *const mv){
		//センターガード
		genDrawShot(fPosXY<>(FX_TEE,FY_TEE-3.7),mv);
	}
	
	template<class move_t>
	void genRightGuard(move_t *const mv){
		//右ガード
		genDrawShot(fPosXY<>(FX_TEE+1.2,FY_TEE-3),mv);
	}
	
	template<class move_t>
	void genLeftGuard(move_t *const mv){
		//左ガード
		genDrawShot(fPosXY<>(FX_TEE-1.2,FY_TEE-3),mv);
	}
	
	template<class pos_t,class move_t>
	void genStrongHit(const pos_t& pos,move_t *const mv,int free_spin=0){
		//指定されたposのストーンを強くヒットする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointF(fPosXY<>(pos.getX(),pos.getY()-0.01),33.0,mv);
	}
	
	template<class pos_t,class move_t>
	void genMiddleHit(const pos_t& pos,move_t *const mv,int free_spin=0){
		//指定されたposのストーンを中程度の強さでヒットする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointGoF(fPosXY<>(pos.getX(),pos.getY()-0.01),5.0,mv);
	}
	
	template<class pos_t,class move_t>
	void genWeakHit(const pos_t& pos,move_t *const mv,int free_spin=0){
		//指定されたposのストーンを軽くヒットする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointGoF(fPosXY<>(pos.getX(),pos.getY()-0.01),0.8,mv);
	}
	
	template<class pos_t,class move_t>
	void genFreeze(const pos_t& pos,move_t *const mv,int free_spin=0){
		//指定されたposのストーンにちょうど到達するショット
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		genDrawShot(fPosXY<>(pos.getX(),pos.getY()+0.01-F_STONE_RAD*2),mv);
	}
	
	/**************************レンジ着手の生成**************************/
	
	template<class rmove_t>
	void genTeeRShot(rmove_t *const mv){
		//ティーショット
		FastSimulator::FXYtoFMV(fPosXY<>(FX_TEE,FY_TEE),mv);
		mv->setRangeVX( F_STEP*2048*2 );
		mv->setRangeVY( F_STEP*2048*2 );
		
		//CERR<<*mv<<endl;
	}
	
	template<class pos_t,class rmove_t>
	void genRStrongHit(const pos_t& pos,rmove_t *const mv,int free_spin=0){
		//指定されたposのストーンをテイクアウトする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointF(fPosXY<>(pos.getX(),pos.getY()-0.01),33.0,mv);
		
		mv->setRangeVX( F_STEP*512*2*2 );
		mv->setRangeVY( F_STEP*512*2*2 );
	}
	template<class pos_t,class rmove_t>
	void genRMiddleHit(const pos_t& pos,rmove_t *const mv,int free_spin=0){
		//指定されたposのストーンをテイクアウトする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointGoF(fPosXY<>(pos.getX(),pos.getY()-0.01),5.0,mv);
		
		mv->setRangeVX( F_STEP*512*2*2 );
		mv->setRangeVY( F_STEP*512*2*2 );
	}
	template<class pos_t,class rmove_t>
	void genRWeakHit(const pos_t& pos,rmove_t *const mv,int free_spin=0){
		//指定されたposのストーンをテイクアウトする
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		FastSimulator::rotateToPassPointGoF(fPosXY<>(pos.getX(),pos.getY()-0.01),0.8,mv);
		
		mv->setRangeVX( F_STEP*512*2*2 );
		mv->setRangeVY( F_STEP*512*2*2 );
	}
	template<class pos_t,class move_t>
	void genRFreeze(const pos_t& pos,move_t *const mv,int free_spin=0){
		//指定されたposのストーンにちょうど到達するショット
		if( !free_spin ){
			if( pos.getX() > FX_TEE ){
				mv->setSpin(Spin::RIGHT);
			}else{
				mv->setSpin(Spin::LEFT);
			}
		}
		genDrawShot(fPosXY<>(pos.getX(),pos.getY()+0.01-F_STONE_RAD*2),mv);
		
		mv->setRangeVX( F_STEP*512*2*2 );
		mv->setRangeVY( F_STEP*512*2*2 );
	}
}