//デジタルカーリング用
//公式提供の関数をベースにしたシミュレーション
//盤面表現の基本型はThinkBoard

namespace DigitalCurling{
	
	b2Body *createBody( fpn_t x, fpn_t y, b2World *const world )
	{
		b2BodyDef bodyDef;
		
		bodyDef.type = b2_dynamicBody;
		bodyDef.position.Set( x, y );//位置
		bodyDef.angle = 0;//角度(円だから関係無し?)
		
		b2Body *body = world->CreateBody( &bodyDef );//ボディの生成
		
		//フィクスチャの作成
		b2FixtureDef fixtureDef;
		
		b2CircleShape dynamicBall;
		dynamicBall.m_radius = F_STONE_RAD;
		
		fixtureDef.shape = &dynamicBall;
		
		fixtureDef.density = F_DENS_STONE;// 密度
		fixtureDef.restitution = F_REST_STONE;// 反発力
		fixtureDef.friction = F_FRIC_STONE;// 摩擦力
		fixtureDef.filter.categoryBits = STONEINFO_CATEGORYBITS;//?
		fixtureDef.filter.maskBits =  STONEINFO_MASKBITS;//?
		
		body->CreateFixture( &fixtureDef );
		
		return body;
	}
	
	// ゲーム状態
	struct b2Board{
		
		//Box2dシミュレーションのための局面表現
		b2World world;
		b2Body *body[16];
		StoneState state[16];//状態
		//int fg;
		
		int	NAsleep;
		int NAwake;
		
		int NMoved;//結果的に動いた石の数
		
		b2Board():
		world( b2Vec2(0,0) ),
		NAsleep(0),
		NAwake(0)
		{}
		
		~b2Board(){
			const int size=NAwake+NAsleep;
			for( int i=0; i<size; i++ ){
				assert( body[i] != nullptr );
				world.DestroyBody( body[i] );
			}
		}
		
		template<class mst_t>
		void setAwakeStone(const mst_t& org, const int col=WHITE){
			//動く石を1つセット
			//偶然速さが0になる場合を否定出来ないので、速さチェックは入れない
			b2Body *newBody=createBody( org.getX(), org.getY(), &world );
			
			newBody->SetLinearVelocity( b2Vec2( org.getVX(), org.getVY() ));//Vx,Vyを設定
			newBody->SetAngularVelocity( org.getW() );//回転を設定
			
			body[ NAwake ] = newBody;
			state[ NAwake ].init();
			state[ NAwake ].setColor(col);
			
			++NAwake;
		}
		
		template<int RINK_ONLY=1,class board_t>
		void setAsleepStones(const board_t& bd){
			//静止している石をセット
			//動く石のセットの後の必要あり
			
			int cnt=NAwake;
			for( int c=0; c<2; ++c ){
				for( int i=0; i<bd.NActive[c]; ++i ){
					
					if( RINK_ONLY ){
						assert( isInPlayArea( bd.getStonePos(c,i) ) );//有効エリアに入っているものしか考慮しない
					}
					fpn_t x=bd.getStonePos(c,i).getX();
					fpn_t y=bd.getStonePos(c,i).getY();
					b2Body *newBody = createBody( x, y, &world );
					
					newBody->SetLinearVelocity( b2Vec2(0.0, 0.0) );//Vx,Vyを設定
					newBody->SetAngularVelocity( 0.0 );
					
					body[ cnt ] = newBody;
					state[ cnt ] = bd.state[c][i];
					
					++cnt;
				}
			}
			NAsleep = cnt-NAwake;
		}
		
		template<int RINK_ONLY=1,class st_t>
		void setAsleepStone(const st_t& org, const StoneState& astate){
			//静止している石を1つセット
			//動く石のセットの後の必要あり
			
			int id=NAwake+NAsleep;
			
			if( RINK_ONLY ){
				assert( isInPlayArea( org.x, org.y ) );//有効エリアに入っているものしか考慮しない
			}
			b2Body *newBody = createBody( org.x, org.y, &world );
			
			newBody->SetLinearVelocity( b2Vec2(0.0, 0.0) );//Vx,Vyを設定
			newBody->SetAngularVelocity( 0.0 );
			
			body[ id ] = newBody;
			state[ id ] = astate;
			
			++NAsleep;
		}
	};
	
	
	namespace Simulator{
		
		//ショットの強さを摩擦力を考慮した値に直す
		template<class vxy_t,class pos_t>
		void DIRtoVEC( const pos_t& arg , vxy_t *const dst ){
			
			fpn_t x=arg.x-FX_THROW;
			fpn_t y=arg.y-FY_THROW;
			
			fpn_t l= sqrt( 2.0f*F_FRIC_STONE ) / sqrt( XYtoR( x, y ) );
			dst->setVX( x*l );
			dst->setVY( y*l );
			//c.stop();
		}
		
		//^の逆関数
		template<class vxy_t,class pos_t>
		void VECtoDIR( const vxy_t& arg , pos_t *const dst ){
			const fpn_t l=XYtoR(arg.x,arg.y)/(2.0f*F_FRIC_STONE);
			dst->x=arg.x * l + FX_THROW;
			dst->y=arg.y * l + FY_THROW;
		}
		
		//停止位置座標からノーレンジ着手を生成
		//スピンはmvにすでに指定されているものとする
		template<class pos_t,class nrmove_t>
		void calcDeparture( const pos_t& pos, nrmove_t *const mv )
		{
			constexpr fpn_t tt = static_cast<fpn_t>(0.0335f);
			constexpr fpn_t k = static_cast<fpn_t>(1.22f);
			
			fPosXY<> tpos;
			
			if( mv->getSpin()!=Spin::RIGHT ){//左回転
				tpos.set( k + pos.x - tt*( pos.y - FY_TEE ), tt*( pos.x - FX_TEE ) + pos.y );
			}else{//右回転
				tpos.set( -k + pos.x + tt*( pos.y - FY_TEE ), -tt*( pos.x - FX_TEE ) + pos.y );
			}
			
			DIRtoVEC( tpos, mv );
		}
		
		//^の逆関数
		//着手から乱数無し時の予測停止位置を計算する
		template<class move_t,class pos_t>
		void calcDestination( const move_t& mv, pos_t *const pos ){
			
			constexpr fpn_t tt = static_cast<fpn_t>(0.0335f);
			constexpr fpn_t tt2p1 = tt*tt + static_cast<fpn_t>(1.0f);
			constexpr fpn_t k = static_cast<fpn_t>(1.22f);
			fPosXY<> dvec;
			
			VECtoDIR( mv, &dvec );
			if( mv.getSpin()!=Spin::RIGHT ){//左回転
				pos->set( (dvec.x - tt*FY_TEE - k + tt*dvec.y + tt*tt*FX_TEE )/tt2p1 ,
						 ( -tt*dvec.x + tt*tt*FY_TEE + k*tt + dvec.y + tt*FX_TEE )/tt2p1 );
			}else{//右回転
				pos->set( (dvec.x + tt*FY_TEE + k - tt*dvec.y + tt*tt*FX_TEE )/tt2p1 ,
						 ( +tt*dvec.x + tt*tt*FY_TEE + k*tt + dvec.y - tt*FX_TEE )/tt2p1 );
			}
		}

		template<class float_t,class vec_t>
		void FrictionStep( float_t fric, const vec_t& vec, float_t angle,vec_t *const dst, float_t *const dstangle )
		{
			//1つの石に摩擦を掛ける
			float_t v=XYtoR(vec.x,vec.y);
			
			if( v > fric ){
				float_t vx_v=vec.x/v;
				float_t vy_v=vec.y/v;
				float_t dx= -fric*vx_v;
				float_t dy= -fric*vy_v;
				if( angle!=0 ){
					float_t d=FANGV_ORG * fric;
					
					if( angle>0 ){
						dx+= d *vy_v;
						dy+= -d *vx_v;
					}else{
						dx+= -d *vy_v;
						dy+= d *vx_v;
					}
				}
				dst->set( vec.x+dx, vec.y+dy );
				*dstangle=angle;
			}else{
				dst->set( 0 , 0 );
				*dstangle=0;
			}
		}
		
		template<class float_t,class vec_t>
		void FrictionInvStep( float_t fric, const vec_t& vec, float_t angle, vec_t *const dst, float_t *const dstangle )
		{
			//FrictionStepの逆変換
			//石が静止しているときだけは、前の状態を生成出来ないが、そのときは速度をY軸の正方向にする
			//この静->動の場合の速さの設定値はおそらく間違っているのだろうが、
			//フレームレートを上げれば無視出来る誤差なので気にしていない
			if( vec.x==0 && vec.y==0 ){
				dst->set( 0, fric/20000 );
				*dstangle=FANGV_ORG;
			}else{
				float_t v=XYtoR(vec.x,vec.y);
				float_t vx_v=vec.x/v;
				float_t vy_v=vec.y/v;
				float_t dx= fric*vx_v;
				float_t dy= fric*vy_v;
				if( angle!=0 ){
					float_t d=FANGV_ORG * fric;
					
					if( angle>0 ){
						dx+= -d *vy_v;
						dy+= d *vx_v;
					}else{
						dx+= d *vy_v;
						dy+= -d *vx_v;
					}
				}
				dst->set( vec.x+dx, vec.y+dy );
				*dstangle=angle;
			}
		}
		
		void friction( fpn_t fric, b2Body *const pbody ){
			//b2Body表現の石に摩擦を掛ける
			fPosXY<> ovec( pbody->GetLinearVelocity().x, pbody->GetLinearVelocity().y ),vec;
			fpn_t ow=pbody->GetAngularVelocity();
			fpn_t w;
			
			FrictionStep<fpn_t,fPosXY<>>( fric, ovec, ow, &vec, &w  );
			pbody->SetLinearVelocity( b2Vec2( vec.x, vec.y) );
			pbody->SetAngularVelocity( w );
		}
		
		void frictionAll( fpn_t fric, b2Board *const pSubGameState )
		{
			//全てのストーンについて、摩擦を考慮して速度ベクトルの値を調整する
			//fricは速度の変化量
			for( int i=0; i<pSubGameState->NAwake; i++ ){
				assert( pSubGameState->body[i]!=nullptr );//除外物体は含めない
				friction( fric, pSubGameState->body[i] );
			}
		}
		
		void step( b2Board *const pb2bd, const fpn_t timeStep ){
			//b2Boardで表現された局面を1ステップ進める
			frictionAll( F_FRIC_STONE * timeStep * 0.5, pb2bd);
			pb2bd->world.Step( timeStep, velocityIterations, positionIterations );
			frictionAll( F_FRIC_STONE * timeStep * 0.5, pb2bd);
		}
		
		template<int RINK_ONLY=1>//リンク外も考慮する
		int MainLoop( b2Board *const pb2bd, const fpn_t timeStep, const int LoopCount = -1 )
		{
			//公式由来のシミュレーションメイン関数
			//有効範囲内の石は、動->静の順で配列に詰めて保持しているので、
			//石が止まったり範囲外になったりした場合には並べ替えを行っている
			
			int iRet;
			
			frictionAll( F_FRIC_STONE * timeStep * 0.5, pb2bd);//最初に0.5ステップ分摩擦を掛ける
			
			for( iRet=0; iRet < LoopCount || LoopCount == -1; iRet++ ){
				
				// 摩擦力の計算
				pb2bd->world.Step( timeStep, velocityIterations, positionIterations );

				//ストーンの除外判定
				//動いている石にのみ判定
				int con=0;
				for( int i=0; i<pb2bd->NAwake; ){
					assert( pb2bd->body[i] != nullptr );
					// ストーンがリンク外に出たら除外する
					b2Vec2 vec = pb2bd->body[i]->GetPosition();
					if( RINK_ONLY && !isOnRink( vec.x, vec.y ) ){//リンク内にない
						//フリーガードゾーンルール適用ターンの時、フリーガード違反を判定
						if( pb2bd->state[i].isFreeGuard() ){
							//フリーガード違反なので、シミュレーション結果を破棄
							DERR<<"freeguard foul"<<endl;
							return 0;
						}
						
						pb2bd->world.DestroyBody( pb2bd->body[i] );
						--pb2bd->NAwake;
						//無くなった分の石を詰める
						
						if( i==pb2bd->NAwake ){
							//インデックス最後の石が消えた
							if( pb2bd->NAsleep!=0 ){
								pb2bd->body[i]=pb2bd->body[ i + pb2bd->NAsleep ];
								pb2bd->state[i]=pb2bd->state[ i + pb2bd->NAsleep ];
							}
						}else{
							//インデックス最後でない石が消えたので、2手間かけて詰める必要あり
							pb2bd->body[i]=pb2bd->body[ pb2bd->NAwake ];
							pb2bd->state[i]=pb2bd->state[ pb2bd->NAwake ];
							if( pb2bd->NAsleep!=0 ){
								pb2bd->body[pb2bd->NAwake]=pb2bd->body[ pb2bd->NAwake + pb2bd->NAsleep ];
								pb2bd->state[pb2bd->NAwake]=pb2bd->state[ pb2bd->NAwake + pb2bd->NAsleep ];
							}
						}
						

					}else{
						b2Vec2 v=pb2bd->body[i]->GetLinearVelocity();
						if( !v.x && !v.y ){
							//新バージョンのbox2dではIsAwake()関数の挙動が違うので、速さによって判定しないとおかしくなる
							
							//動いていた石が静止した
							--pb2bd->NAwake;
							++pb2bd->NAsleep;
							if( i!=pb2bd->NAwake ){
								//インデックス最後でない石が静止した
								std::swap( pb2bd->body[i], pb2bd->body[pb2bd->NAwake] );
								std::swap( pb2bd->state[i], pb2bd->state[pb2bd->NAwake] );
							}
						}else{
							//次のインデックスの石へ
							++i;
						}
					}
				}
				//止まっている石が動き出したか判定
				if( pb2bd->NAsleep!=0 ){
					//インデックス最初のもの
					b2Vec2 v=pb2bd->body[pb2bd->NAwake]->GetLinearVelocity();
					while( v.x || v.y ){
						DERR<<"remove"<<endl;
						//CERR<<pb2bd->body[pb2bd->NAwake]->GetAngularVelocity()<<endl;
						++pb2bd->NAwake;
						--pb2bd->NAsleep;
						if( pb2bd->NAsleep==0 ){goto TO_NEXT;}
					}
					//インデックス最初でないもの
					const int size=pb2bd->NAwake+pb2bd->NAsleep;
					for( int i=pb2bd->NAwake+1;i<size;++i ){
						b2Vec2 v=pb2bd->body[i]->GetLinearVelocity();
						if( v.x || v.y ){
							DERR<<"remove"<<endl;
							std::swap( pb2bd->body[i], pb2bd->body[pb2bd->NAwake] );
							std::swap( pb2bd->state[i], pb2bd->state[pb2bd->NAwake] );
							++pb2bd->NAwake;
							--pb2bd->NAsleep;
						}
					}
				}
				if( pb2bd->NAwake == 0 ){
					//動いている石がなくなった
					//CERR<<"all stones were stopped."<<endl;
					//getchar();
					return iRet;
				}
				
			TO_NEXT:;
				
				frictionAll( F_FRIC_STONE * timeStep, pb2bd );//動摩擦係数×重力加速度×1ステップの時間の摩擦を掛ける
			}
		}
		
		//シミュレーション本体
		template<int RINK_ONLY=1,class board_t,class mst_t>
		int simulateb2d( board_t *const pbd, const mst_t& org, const fpn_t step ){
			
			//box2dを用いたシミュレーションを行う
			
			b2Board b2bd;
			const int LoopCount=-1;

			// ゲーム情報の初期化
			b2bd.setAwakeStone( org, pbd->getTurnColor() );
			b2bd.setAsleepStones<RINK_ONLY>( *pbd );

			// 物理演算
			int ret=MainLoop<RINK_ONLY>( &b2bd, step, LoopCount );
			
			// シミュレーション結果をまとめる
			if( ret!=0 ){
				//石の配置が変化する可能性が高い
				pbd->initStonePos();
				const int size=b2bd.NAwake+b2bd.NAsleep;
				for( int i=0; i<size; ++i ){
					assert( b2bd.body[i] != nullptr );
					const int col=b2bd.state[i].getColor();
					const b2Vec2 vec = b2bd.body[i]->GetPosition();
					if( (!RINK_ONLY) || isInPlayArea( vec.x, vec.y ) ){
						pbd->pushStone(col,vec.x,vec.y);
						DERR<<OutColor(col)<<" ("<<vec.x<<","<<vec.y<<")"<<endl;
					}
				}
			}
			return ret;
		}
		
		template<class board_t,class move_t>
		int simulate( board_t *const pbd, const move_t& mv ){
			
			//着手形式を引数とする場合

			fpn_t w;
			if( mv.isLeftSpin() ){
				w=-FANGV_ORG;
			}else{
				w=FANGV_ORG;
			}
			
			//シミュレーション開始座標と速度
			const fMStone<fpn_t> org( FX_THROW, FY_THROW, mv.getVX(), mv.getVY(), w );
			
			int ret=simulateb2d( pbd, org, g_timeStep );
			return 2;
		}
		
		template<class float_t,class mst_t>
		void stepSolo(mst_t *const mst, const float_t step){
			//衝突が無いとして、1つの石の状態を1ステップ進める
			fPosXY<float_t> ov( mst->getVX(), mst->getVY() ),v;//ストーンの速度
			float_t ow( mst->getW()) ,w;//ストーンの角速度
			fPosXY<float_t> pos( mst->getX(), mst->getY() );
			
			FrictionStep( static_cast<float_t>(F_FRIC_STONE) * step * static_cast<float_t>(0.5), ov, ow, &v, &w );//速度を半ステップ進める
			const float_t dx=v.x*step;const float_t dy=v.y*step;
			mst->x+=dx;mst->y+=dy;//位置更新
			ov=v;ow=w;
			FrictionStep<float_t>( static_cast<float_t>(F_FRIC_STONE) * step * static_cast<float_t>(0.5), ov, ow, &v, &w );//速度を半ステップ進める
			mst->vx=v.x;mst->vy=v.y;
			mst->w=w;
			return;
		}
		
		template<class float_t,class mst_t>
		void invStepSolo(mst_t *const mst, const float_t step){
			//衝突が無いとして、1つの石の状態を1ステップ元に戻す
			fPosXY<float_t> ov( mst->getVX(), mst->getVY() ),v;//ストーンの速度
			float_t ow( mst->getW()) ,w;//ストーンの角速度
			fPosXY<float_t> pos( mst->getX(), mst->getY() );
			FrictionInvStep( static_cast<float_t>(F_FRIC_STONE) * step * static_cast<float_t>(0.5), ov, ow, &v, &w );//速度を半ステップ戻す
			const float_t dx=v.x*step;const float_t dy=v.y*step;
			mst->x-=dx;mst->y-=dy;//位置更新
			ov=v;ow=w;
			FrictionInvStep<float_t>( static_cast<float_t>(F_FRIC_STONE) * step * static_cast<float_t>(0.5), ov, ow, &v, &w );//速度を半ステップ戻す
			mst->vx=v.x;mst->vy=v.y;
			mst->w=w;
			return;
		}
		
		template<class mst_t>
		int simulateSolo(const mst_t& org,mst_t *const dst,const fpn_t step){
			
			//衝突を考えず、1つの石の軌道をシミュレートする
			//精度最優先
			//最終到達位置を記録する
			//ステップは1フレームの時間
			
			fPosXY<> ov( org.getVX(), org.getVY() ),v;//ストーンの速度
			fpn_t ow( org.getW()) ,w;//ストーンの角速度
			fPosXY<> pos( org.getX(), org.getY() );
			
			FrictionStep( F_FRIC_STONE * step * 0.5, ov, ow, &v, &w );
			for(int i=1;;){
				
				++i;
				if( v.y==0 && v.x==0 ){
					dst->setX( pos.x );
					dst->setY( pos.y );
					return 2;
				}
				
				const fpn_t dx=v.x*step;
				const fpn_t dy=v.y*step;
				pos.x+=dx;
				pos.y+=dy;
				
				ov=v;
				ow=w;
				FrictionStep( F_FRIC_STONE * step, ov, ow, &v, &w );//速度更新
			}
			assert(0);
		}
		
		template<class mst_t>
		int simulateSoloEx(const mst_t& org,mst_t *const dst,const fpn_t step){
			
			//衝突を考えず、1つの石の軌道をシミュレートする
			//精度最優先
			//それぞれの時間での位置を記録する
			//フレーム数を返す
			//ステップは1フレームの時間
			
			fPosXY<> ov( org.getVX(), org.getVY() ),v;//ストーンの速度
			fpn_t ow( org.getW() ),w;//ストーンの角速度
			dst[0]=org;
			
			FrictionStep( F_FRIC_STONE * step * 0.5, ov, ow, &v, &w );
			for(int i=1;;){
				
				// 摩擦力の計算
				const fpn_t dx=v.x*step;
				const fpn_t dy=v.y*step;
				
				dst[i].setX( dst[i-1].getX() + dx );
				dst[i].setY( dst[i-1].getY() + dy );
				dst[i].setVX( v.x );
				dst[i].setVY( v.y );
				dst[i].setW( w );
				
				++i;
				if( v.y==0 && v.x==0 ){return 2;}
				
				ov=v;
				ow=w;
				FrictionStep( F_FRIC_STONE * step, ov, ow, &v, &w );//速度更新
			}
			assert(0);
		}
		
		template<class pos_t,class move_t>
		void rotateToPassPoint(const pos_t& pos,move_t *const mv,const fpn_t step){
			//指定された位置を通るように速度ベクトルを回転させる
			
			//ストーンの初期位置を(0,0)とする座標系を使う
			
			//const b2Vec2 tar( pos.x - FX_THROW, FY_THROW - pos.y );
			const fpn_t ltar=XYtoR(pos.x - FX_THROW,pos.y - FY_THROW);
			const fpn_t ttar=atan2(pos.x - FX_THROW,pos.y - FY_THROW);//x,yの順は別にどっちでもいいが
			
			//CERR<<"l "<<ltar<<" theta "<<ttar<<endl;
			
			fPosXY<> otp( 0, 0 ),tp;//ストーンの位置
			fPosXY<> ov( mv->getVX(), mv->getVY() ),v;//ストーンの速度
			fpn_t ow,w;//ストーンの角速度
			
			//回転を設定
			if( mv->isLeftSpin() ){
				ow = FANGV_ORG * -1;
			}else{
				ow = FANGV_ORG;
			}
			
			fpn_t ol=0;
			FrictionStep( F_FRIC_STONE * step * 0.5, ov, ow, &v, &w );
			
			for(int i=0; ; i++){
				
				// 摩擦力の計算
				const fpn_t dx=v.x * step;
				const fpn_t dy=v.y * step;
				tp.x=otp.x+dx;
				tp.y=otp.y+dy;
				
				//CERR<<tp<<endl;
				
				const fpn_t l=XYtoR(tp.x,tp.y);//原点との距離
				
				if( l > ltar ){//目標より遠くに行ってしまったので終了
					//距離の割合から、位置を微調整
					fpn_t a = l - ol;
					fpn_t b = ltar - ol;
					
					assert( a>0 );
					
					const fpn_t s= b/a;
					otp.x+=s * dx;
					otp.y+=s * dy;
					
					const fpn_t n=atan2( otp.x,otp.y )-ttar;//角度差
					
					const fPosXY<> u = fPosXY<>( mv->x, mv->y );//停止時の向きのみ正しい着手ベクトル
					mv->setVX( +cos( n )*u.x + sin( n )*u.y );
					mv->setVY( -sin( n )*u.x + cos( n )*u.y );
					return;
				}
				if( v.y==0 ){return;}
				
				ol=l;
				ov=v;
				ow=w;
				otp=tp;
				FrictionStep( F_FRIC_STONE * step, ov, ow, &v, &w );//速度更新
			}
		}
	}
}

