//デジタルカーリング用
//定跡手の生成

namespace DigitalCurling{
	
	enum Killer{//キラー手番号
		NO=-1,//キラー手ではない事を示す
		PASS=0,
		TEE_SHOT_R,
		TEE_SHOT_L,
		SHORT_TEE_SHOT_R,
		SHORT_TEE_SHOT_L,
		RIGHT_SIDE_SHOT_R,
		RIGHT_SIDE_SHOT_L,
		LEFT_SIDE_SHOT_R,
		LEFT_SIDE_SHOT_L,
		GUARD_CENTER,
		GUARD_RIGHT,
		GUARD_LEFT,
		HITSTRONG_OPS1_R,
		HITSTRONG_OPS1_L,
		HITMIDDLE_OPS1_R,
		HITMIDDLE_OPS1_L,
		HITWEAK_OPS1_R,
		HITWEAK_OPS1_L,
		FREEZE_OPS1_R,
		FREEZE_OPS1_L,
		HITSTRONG_OPS2_R,
		HITSTRONG_OPS2_L,
		HITMIDDLE_OPS2_R,
		HITMIDDLE_OPS2_L,
		HITWEAK_OPS2_R,
		HITWEAK_OPS2_L,
		FREEZE_OPS2_R,
		FREEZE_OPS2_L,
	};
	
	struct KillerMove{
		int type;
		
		constexpr operator int(){return type;}
		
		void set(int at){type=at;}
		void settle(){type|=(1<<16);}//決定手(他の手を読まない)を宣言
		
		int getType()const{return type & 255U;}
		int isSettled()const{return type & (1<<16);}
		
		constexpr KillerMove():type(){}
		constexpr KillerMove(const int& at):type(at){}
		constexpr KillerMove(const KillerMove& ak):type(ak.type){}
	};
	
	template<class killermove_t,class board_t>
	int pushKillerRoot(killermove_t *const mv0,const board_t& bd){
		//与えられた局面で検討しておきたいキラー手を配列に入れていく(ルートノード)
		killermove_t *mv=mv0;
		
		const int col=bd.getTurnColor();
		const int ops_col=flipColor(col);
		const int end=bd.getEnd();
		const int turn=bd.getTurn();
		
		const bool sym=bd.isSymmetry();
		
		//まず詰みを判定
		int mate=judgeMateStrictly(bd);
		if( mate==col ){
			//自分の必勝
#ifdef USE_NULLMOVE
			mv->set(Killer::PASS);
			mv->settle();
			return 1;
#endif
		}else if( mate==ops_col ){
			//相手の必勝
#ifdef USE_NULLMOVE
			mv->set(Killer::PASS);
			mv->settle();
			return 1;
#endif
		}
		
		if( turn==TURN_LAST ){
			if( bd.getEndScore()==0 ){//ブランクエンドに出来る
				if( getTmpEndReward(WHITE,0) >= getTmpEndReward(WHITE,-1) ){
					//ブランクエンドが後手1点より良い(後手にとって)場合
					if( bd.NActive[WHITE]==0 ){//2点以上は取れない
#ifdef USE_NULLMOVE
						mv->set(Killer::PASS);
						mv->settle();
						return 1;
#endif
					}
				}
			}
		}
		if( bd.isNull() ){
			//空場定跡
			if( col==BLACK ){//先手
				if( bd.hasAdvantage(BLACK) ){//得点勝ち
					if( turn==TURN_LAST + 1 ){
						mv->set(Killer::SHORT_TEE_SHOT_R);//2015/3/21 ダウン回避のため定跡追加しました
					}else{
						mv->set(Killer::PASS);
					}
					mv->settle();
					return 1;
				}else if( bd.isTie() ){//同点
					if( end >= END_JOBAN ){
						if( turn!=(TURN_LAST + 1) ){
							mv->set(Killer::PASS);
							mv->settle();
							return 1;
						}else{
							mv->set(Killer::SHORT_TEE_SHOT_R);//2015/3/22 ダウン回避のため定跡追加しました
							mv->settle();
							return 1;
						}
					}else{
						if( turn >= 12 ){
							mv->set(Killer::GUARD_CENTER);
							mv->settle();
							return 1;
						}
					}
				}else{//得点負け
					if( turn >= 12 ){
						mv->set(Killer::GUARD_CENTER);
						mv->settle();
						return 1;
					}
				}
			}else{//後手
				if( bd.hasAdvantage(WHITE) ){//得点勝ち
					mv->set(Killer::RIGHT_SIDE_SHOT_R);
					mv->settle();
					return 1;
				}else if( bd.isTie() ){//同点
					if( bd.isFinalTurn() ){//ゲーム最終投
						//センターにドローで終わり
						mv->set(Killer::TEE_SHOT_R);
						mv->settle();
						return 1;
					}
					if( turn >= 12 ){
						mv->set(Killer::GUARD_RIGHT);
						mv->settle();
						return 1;
					}
				}else{//得点負け
					if( bd.isFinalTurn() ){//ゲーム最終投
						//センターにドローで延長戦
						mv->set(Killer::TEE_SHOT_R);
						mv->settle();
						return 1;
					}
					if( turn >= 12 ){
						mv->set(Killer::GUARD_RIGHT);
						mv->settle();
						return 1;
					}
				}
			}
		}
		if( bd.NInHouse[ops_col]>0 ){
			//No1に当てるショット
			mv->set(Killer::FREEZE_OPS1_R); ++mv;
			if( !sym ){
				mv->set(Killer::FREEZE_OPS1_L); ++mv;
			}
			mv->set(Killer::HITWEAK_OPS1_R); ++mv;
			if( !sym ){
				mv->set(Killer::HITWEAK_OPS1_L); ++mv;
			}
			mv->set(Killer::HITMIDDLE_OPS1_R); ++mv;
			if( !sym ){
				mv->set(Killer::HITMIDDLE_OPS1_L); ++mv;
			}
			mv->set(Killer::HITSTRONG_OPS1_R); ++mv;
			if( !sym ){
				mv->set(Killer::HITSTRONG_OPS1_L); ++mv;
			}
			if( bd.NInHouse[ops_col]>1 ){
				//No2に当てるショット
				mv->set(Killer::FREEZE_OPS2_R); ++mv;
				if( !sym ){
					mv->set(Killer::FREEZE_OPS2_L); ++mv;
				}
				mv->set(Killer::HITWEAK_OPS2_R); ++mv;
				if( !sym ){
					mv->set(Killer::HITWEAK_OPS2_L); ++mv;
				}
				mv->set(Killer::HITMIDDLE_OPS2_R); ++mv;
				if( !sym ){
					mv->set(Killer::HITMIDDLE_OPS2_L); ++mv;
				}
				mv->set(Killer::HITSTRONG_OPS2_R); ++mv;
				if( !sym ){
					mv->set(Killer::HITSTRONG_OPS2_L); ++mv;
				}
			}
		}
		if( turn==TURN_LAST ){
			//ティーショットを生成
			mv->set(Killer::TEE_SHOT_R); ++mv;
			if( !sym ){
				mv->set(Killer::TEE_SHOT_L); ++mv;
			}
		}
		return mv-mv0;
	}
	
	template<class killermove_t,class board_t>
	int pushKiller(killermove_t *const mv0,const board_t& bd){
		//与えられた局面で検討しておきたいキラー手を配列に入れていく
		killermove_t *mv=mv0;
		
		const int col=bd.getTurnColor();
		const bool sym=bd.isSymmetry();
		
		if( bd.NInHouse[flipColor(col)]!=0 ){
			//No1テイクアウトを生成
			mv->set(Killer::HITMIDDLE_OPS1_R); ++mv;
			if( !sym ){
				mv->set(Killer::HITMIDDLE_OPS1_L); ++mv;
			}
			if( bd.NInHouse[flipColor(col)]>1 ){
				
				mv->set(Killer::HITMIDDLE_OPS2_R); ++mv;
				if( !sym ){
					mv->set(Killer::HITMIDDLE_OPS2_L); ++mv;
				}
			}
			//if( bd.getTurn()==TURN_LAST || col==BLACK ){
				//ティーショットを生成
				mv->set(Killer::TEE_SHOT_R); ++mv;
				if( !sym ){
					mv->set(Killer::TEE_SHOT_L); ++mv;
				}
			//}else{
				//サイドショットを生成
			//	mv->set(Killer::RIGHT_SIDE_SHOT_L); ++mv;
			//	if( !sym ){
			//		mv->set(Killer::LEFT_SIDE_SHOT_R); ++mv;
			//	}
			//}
		}else{
			//ティーショットを生成
			mv->set(Killer::TEE_SHOT_R); ++mv;
			if( !sym ){
				mv->set(Killer::TEE_SHOT_L); ++mv;
			}
			//サイドショットを生成
			mv->set(Killer::RIGHT_SIDE_SHOT_L); ++mv;
			if( !sym ){
				mv->set(Killer::LEFT_SIDE_SHOT_R); ++mv;
			}
		}
		
		return mv-mv0;
	}
	
	template<class killermove_t>
	int pushNullKiller(killermove_t *const mv0,int t){
		//ターンtの空場で検討しておきたいキラー手を配列に入れていく
		killermove_t *mv=mv0;
		//パスを生成
		mv->set(Killer::PASS); ++mv;
		//ティーショットを生成
		//mv->set(Killer::TEE_SHOT_R); ++mv;
		
		if( TURNtoCOL(t)==BLACK ){
			//ショートティーショットを生成
			mv->set(Killer::SHORT_TEE_SHOT_R); ++mv;
			//センターガードを生成
			mv->set(Killer::GUARD_CENTER); ++mv;
		}else{
			//サイドショットを生成
			mv->set(Killer::RIGHT_SIDE_SHOT_R); ++mv;
			//サイドガードを生成
			mv->set(Killer::GUARD_RIGHT); ++mv;
		}
		return mv-mv0;
	}
	
	template<class board_t,class killermove_t,class move_t>
	void genKiller(const board_t& bd,const killermove_t& akill,move_t *const dst){
		//ある局面でのキラー手から実際の着手形式を生成(ノーレンジ)
		switch( akill.getType() ){
			case PASS:{
				genPASS(dst);
			}break;
			case TEE_SHOT_R:{
				dst->setSpin(Spin::RIGHT);
				genTeeShot(dst);
			}break;
			case TEE_SHOT_L:{
				dst->setSpin(Spin::LEFT);
				genTeeShot(dst);
			}break;
			case SHORT_TEE_SHOT_R:{
				dst->setSpin(Spin::RIGHT);
				genShortTeeShot(dst);
			}break;
			case SHORT_TEE_SHOT_L:{
				dst->setSpin(Spin::LEFT);
				genShortTeeShot(dst);
			}break;
			case RIGHT_SIDE_SHOT_R:{
				dst->setSpin(Spin::RIGHT);
				genRightSideShot(dst);
			}break;
			case RIGHT_SIDE_SHOT_L:{
				dst->setSpin(Spin::LEFT);
				genRightSideShot(dst);
			}break;
			case LEFT_SIDE_SHOT_R:{
				dst->setSpin(Spin::RIGHT);
				genLeftSideShot(dst);
			}break;
			case LEFT_SIDE_SHOT_L:{
				dst->setSpin(Spin::LEFT);
				genLeftSideShot(dst);
			}break;
			case GUARD_CENTER:{
				dst->setSpin(Spin::RIGHT);
				genCenterGuard(dst);
			}break;
			case GUARD_RIGHT:{
				dst->setSpin(Spin::RIGHT);
				genRightGuard(dst);
			}break;
			case GUARD_LEFT:{
				dst->setSpin(Spin::LEFT);
				genLeftGuard(dst);
			}break;
			case HITMIDDLE_OPS1_R:{
				dst->setSpin(Spin::RIGHT);
				genMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITMIDDLE_OPS1_L:{
				dst->setSpin(Spin::LEFT);
				genMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITMIDDLE_OPS2_R:{
				dst->setSpin(Spin::RIGHT);
				genMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITMIDDLE_OPS2_L:{
				dst->setSpin(Spin::LEFT);
				genMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			default:ASSERT(0,cerr<<akill.getType()<<endl;);
		}
	}
	
	template<class board_t,class killermove_t,class move_t>
	void genRKiller(const board_t& bd,const killermove_t& akill,move_t *const dst){
		//ある局面でのキラー手から実際の着手形式を生成(レンジあり)
		switch( akill.getType() ){
			case PASS:{
				genPASS(dst);
			}break;
			case TEE_SHOT_R:{
				dst->setSpin(Spin::RIGHT);
				genTeeRShot(dst);
			}break;
			case TEE_SHOT_L:{
				dst->setSpin(Spin::LEFT);
				genTeeRShot(dst);
			}break;
			case HITSTRONG_OPS1_R:{
				dst->setSpin(Spin::RIGHT);
				genRStrongHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITSTRONG_OPS1_L:{
				dst->setSpin(Spin::LEFT);
				genRStrongHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITMIDDLE_OPS1_R:{
				dst->setSpin(Spin::RIGHT);
				genRMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITMIDDLE_OPS1_L:{
				dst->setSpin(Spin::LEFT);
				genRMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITWEAK_OPS1_R:{
				dst->setSpin(Spin::RIGHT);
				genRWeakHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITWEAK_OPS1_L:{
				dst->setSpin(Spin::LEFT);
				genRWeakHit(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case FREEZE_OPS1_R:{
				dst->setSpin(Spin::RIGHT);
				genRFreeze(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case FREEZE_OPS1_L:{
				dst->setSpin(Spin::LEFT);
				genRFreeze(bd.getStonePos(flipColor(bd.getTurnColor()),0),dst,1);
			}break;
			case HITSTRONG_OPS2_R:{
				dst->setSpin(Spin::RIGHT);
				genRStrongHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITSTRONG_OPS2_L:{
				dst->setSpin(Spin::LEFT);
				genRStrongHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITMIDDLE_OPS2_R:{
				dst->setSpin(Spin::RIGHT);
				genRMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITMIDDLE_OPS2_L:{
				dst->setSpin(Spin::LEFT);
				genRMiddleHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITWEAK_OPS2_R:{
				dst->setSpin(Spin::RIGHT);
				genRWeakHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case HITWEAK_OPS2_L:{
				dst->setSpin(Spin::LEFT);
				genRWeakHit(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case FREEZE_OPS2_R:{
				dst->setSpin(Spin::RIGHT);
				genRFreeze(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			case FREEZE_OPS2_L:{
				dst->setSpin(Spin::LEFT);
				genRFreeze(bd.getStonePos(flipColor(bd.getTurnColor()),1),dst,1);
			}break;
			default:assert(0);
		}
	}
}
