//デジタルカーリング
//クライアントの主観的情報空間

namespace DigitalCurling{
	
	/*
	 struct ShotLog{
	 //1ショットのログ
	 Positions pos;
	 };
	 
	 struct EndLog{
	 //1エンドのログ
	 };
	 
	 struct GameLog{
	 //1ゲームのログ
	 };
	 */
	struct ClientField{
		//主観的情報、客観的情報を分け隔てなく
		//基本的にサーバーから送られてくる情報の型、数値のまま保存
		//ゆえに、primitive型表現とは値が異なる事が多いので注意
		
		//一応、16エンド戦まで対応
		
		//全ゲームの情報
		int ngames[2];
		int wons[2];
		int drews[2][2];
		int loses[2];
		
		uint16 to_end_score_dist[16][17][2];
		
		
		//固定情報のインデックスは最初のエンドの先攻後攻
		std::string name[2];//名前
		
		int my_first_color;//最初のエンドでの自分の先攻or後攻
		TimeLimit tl[2];
		int end_score[16][2];//各エンドの得点
		int score[2];//合計得点
		int rel_score;//最初のエンドの先攻基準での相対得点

		//現エンドでの情報
		int my_color;
		int color[2];
		
		//エンド
		int end;//現在エンド
		int to_end;//エンド数
		
		//エンド内での手数
		int turn;
		const static int to_turn=16;
		
		//エンド中のプレーヤーのインデックスは先攻後攻なので、
		//エンドごとに対応が異なる
		float pos[8][2][2];//ストーン位置
		
		void initAll(){
			//起動時の初期化
			for(int c=0;c<2;++c){
				wons[c]=0;
				loses[c]=0;
				for(int cc=0;cc<2;++cc){
					drews[c][cc]=0;
				}
			}
		}
		void setTimeLimit(int col,uint64 atime){
			tl[col].set(atime);
		}
		void initGame(){
			//1ゲーム開始前の初期化
			score[0]=score[1]=0;
			for(int e=0;e<16;++e){
				end_score[e][0]=end_score[e][1]=0;
			}
			rel_score=0;
			
			color[0]=0;
			color[1]=1;
			
			end=0;
			my_first_color=-1;//自分の先攻後攻が不明
		}
		void setMyFirstColor(int col){
			assert( examColor(col) );
			my_first_color=col;
			my_color=col;
		}
		void setEnd(int aend){
			end=aend;
		}
		void setToEnd(int aend){
			to_end=aend;
		}
		int getNRemainedEnds()const{
			//残りエンド数(開始していないもの)
			return to_end-end-1;
		}
		int getNRemainedTurns()const{
			//残りエンド数(開始していないもの)
			return 16-turn-1;
		}
		void setName(int p,std::string aname){
			assert( examColor(p) );
			name[p]=aname;
		}
		void setTurn(int aturn){
			turn=aturn;
		}
		bool isMyTurn()const{
			return ( my_color==(turn%2) );
		}
		int getTurnColor()const{
			return (turn%2);
		}
		void initEnd(){
			turn=0;
			for(int i=0;i<8;++i){
				for(int j=0;j<2;++j){
					for(int k=0;k<2;++k){
						pos[i][j][k]=0;
					}
				}
			}
		}
		void setEndScore(int s){
			//+ならばid:0のプレーヤー、-ならばid:1のプレーヤーの得点
			//次エンドが始まってから報告されるようだ
			assert( 0<=end && end<16 );
			if( s!=0 ){
				int fcol=(s>0)?0:1;
				rel_score+=s;
				s=abs(s);
				end_score[end][fcol]=s;
				end_score[end][1-fcol]=0;
				score[fcol]+=s;
			}
			assert( rel_score==score[0]-score[1] );
		}
		void closeEnd(){
			//得点の入り方によって次のエンドの手番を決める
			if( (end_score[end][0]>0 && color[0]==1)
			   || (end_score[end][1]>0 && color[1]==1)
			   ){
				std::swap(color[0],color[1]);
				my_color=1-my_color;
			}
		}
		void closeGame(int res){
			
			int my_score=score[my_first_color];
			int ops_score=score[1-my_first_color];
			
			if( res>0 ){//勝ち
				++wons[ my_first_color ];
			}else if( res<0 ){//負け
				++loses[ my_first_color ];
			}else{//引き分け
				if( my_color==0 ){//延長戦先手
					++drews[ my_first_color ][0];
				}else{//延長戦後手
					++drews[ my_first_color ][1];
				}
			}
			++ngames[my_first_color];
		}
		void closeAll(){
			//結果まとめ
			int wons_sum=0;
			int loses_sum=0;
			int drews_sum[2]={0};
			
			for(int c=0;c<2;++c){
				wons_sum+=wons[c];
				loses_sum+=loses[c];
				for(int cc=0;cc<2;++cc){
					drews_sum[cc]+=drews[c][cc];
				}
			}
			cerr<<"b game "
			<<"w : "<<wons[0]<<" l : "<<loses[0]<<" dw : "<<drews[0][1]<<" dl : "<<drews[0][0]<<endl
			<<"w game "
			<<"w : "<<wons[1]<<" l : "<<loses[1]<<" dw : "<<drews[1][1]<<" dl : "<<drews[1][0]<<endl
			<<"all    "
			<<"w : "<<wons_sum<<" l : "<<loses_sum<<" dw : "<<drews_sum[1]<<" dl : "<<drews_sum[0]<<endl;
		}
		void setRunShot(int col,float x,float y){
			
		}
		
		void setStonePos(int sn,float x,float y){
			pos[sn/2][sn%2][0]=x;
			pos[sn/2][sn%2][1]=y;
		}
		void broadcast()const{
			const std::string player[4]={
				"Lead","Second","Third","Skip",
			};
			
			int ost=turn/2;
			
			//現在状況を実況
			CERR<<"*** END : "<<end+1<<" to "<<to_end<<" ";
			CERR<<" SHOT : "<<turn+1<<" to 16 ( "<<player[ost/2]<<" - "<<(ost%2+1)<<" )";
			CERR<<" ***"<<endl;
			
			//名前の文字数は長い方に固定(最低でも8)
			int name_len=std::max(name[0].size(),name[1].size());
			if( name_len < 8 ){name_len=8;}
			
			//ストーンの色は最初のエンドから固定する
			if( color[0]==0 ){
				CERR<<" "<<std::setw(name_len)<<name[0]<<" "<<std::setw(3)<<score[0]<<" ";
				for(int i=0;i<((17-turn)/2)-1;++i){CERR<<"O";}
				if( turn%2 == 0 ){CERR<<"@";}
				CERR<<endl;
				CERR<<" "<<std::setw(name_len)<<name[1]<<" "<<std::setw(3)<<score[1]<<" ";
				for(int i=0;i<((18-turn)/2)-1;++i){CERR<<"O";}
				if( turn%2 == 1 ){CERR<<"@";}
				CERR<<endl;
			}else{
				CERR<<" "<<std::setw(name_len)<<name[1]<<" "<<std::setw(3)<<score[1]<<" ";
				for(int i=0;i<((17-turn)/2)-1;++i){CERR<<"O";}
				if( turn%2 == 0 ){CERR<<"@";}
				CERR<<endl;
				CERR<<" "<<std::setw(name_len)<<name[0]<<" "<<std::setw(3)<<score[0]<<" ";
				for(int i=0;i<((18-turn)/2)-1;++i){CERR<<"O";}
				if( turn%2 == 1 ){CERR<<"@";}
				CERR<<endl;
			}
		}
		
		
		~ClientField(){
			closeAll();
		}
	};
	
	//公式のデータ構造との変換
	void convField(const ClientField& cf,GAMESTATE *const gs){
		gs->ShotNum=cf.turn;
		gs->CurEnd=cf.end;
		gs->LastEnd=cf.to_end;
		for(int i=0;i<10;++i){
			gs->Score[i]=cf.end_score[i][0]-cf.end_score[i][1];
		}
		gs->WhiteToMove=cf.color[0]^(cf.turn%2);
		for(int i=0;i<8;++i){
			gs->body[i*2][0]=cf.pos[i][0][0];
			gs->body[i*2][1]=cf.pos[i][0][1];
			gs->body[i*2+1][0]=cf.pos[i][1][0];
			gs->body[i*2+1][1]=cf.pos[i][1][1];
		}
	}

}