//デジタルカーリング
//ハッシュ関数

namespace DigitalCurling{
	
	/**************************ハッシュテーブル**************************/
	
	//位置は手番ごとに、(x,y)の2次元それぞれにハッシュ値を与える
	uint64 hash_table_bx[16][2];
	uint64 hash_table_by[16][2];
	uint64 hash_table_wx[16][2];
	uint64 hash_table_wy[16][2];
	
	//着手は(vx,vy,回転)の3次元
	//それにvx,vyの範囲を含めた4次元のそれぞれにハッシュ値を与える
	//回転は現在2値なので、ハッシュ値も0と1とする
	
	uint64 hash_table_vx[16][2];
	uint64 hash_table_vy[16][2];
	//uint64 hash_table_s[2];
	std::array<uint64,16> hash_table_vr;
	
	/**************************石の位置ハッシュ関数**************************/
	
	uint64 genHash_BX(uint32 x,int r=0){
		x>>=r;
		uint64 ret=hash_table_bx[r][x&1U];
		for(int i=r+1;i<16;++i){
			x>>=1;
			ret^=hash_table_bx[i][x&1U];
		}
		return ret;
	}
	uint64 genHash_BY(uint32 y,int r=0){
		y>>=r;
		uint64 ret=hash_table_by[r][y&1U];
		for(int i=r+1;i<16;++i){
			y>>=1;
			ret^=hash_table_by[i][y&1U];
		}
		return ret;
	}
	
	uint64 genHash_WX(uint32 x,int r=0){
		x>>=r;
		uint64 ret=hash_table_wx[r][x&1U];
		for(int i=r+1;i<16;++i){
			x>>=1;
			ret^=hash_table_wx[i][x&1U];
		}
		return ret;
	}
	uint64 genHash_WY(uint32 y,int r=0){
		y>>=r;
		uint64 ret=hash_table_wy[r][y&1U];
		for(int i=r+1;i<16;++i){
			y>>=1;
			ret^=hash_table_wy[i][y&1U];
		}
		return ret;
	}
	
	void addHash_BXY(uint64 *const dst,uint32 x,uint32 y,int rmin,int rmax){
		//1個の石について、広いレンジのハッシュ値をまとめて加算
		assert( 0<=rmin && rmax<16 && rmin<=rmax );
		int i=rmin;
		x>>=rmin;y>>=rmin;
		for(;;){
			uint64 hd=hash_table_bx[i][x&1U] ^ hash_table_by[i][y&1U];
			int j=rmin;
			do{
				dst[j]^=hd;
			}while( j++<i );
			++i;
			x>>=1;y>>=1;
			if( i > rmax ){break;}
		}
		for(;;){
			uint64 hd=hash_table_bx[i][x&1U] ^ hash_table_by[i][y&1U];
			int j=rmin;
			do{
				dst[j]^=hd;
			}while( j++<rmax );
			if( i >=16-1 ){break;}
			++i;
			x>>=1;y>>=1;
		}
	}
	
	void addHash_WXY(uint64 *const dst,uint32 x,uint32 y,int rmin,int rmax){
		//1個の石について、広いレンジのハッシュ値をまとめて加算
		assert( 0<=rmin && rmax<16 && rmin<=rmax );
		int i=rmin;
		x>>=rmin;y>>=rmin;
		for(;;){
			uint64 hd=hash_table_wx[i][x&1U] ^ hash_table_wy[i][y&1U];
			int j=rmin;
			do{
				dst[j]^=hd;
			}while( j++<i );
			++i;
			x>>=1;y>>=1;
			if( i > rmax ){break;}
		}
		for(;;){
			uint64 hd=hash_table_wx[i][x&1U] ^ hash_table_wy[i][y&1U];
			int j=rmin;
			do{
				dst[j]^=hd;
			}while( j++<rmax );
			if( i >=16-1 ){break;}
			++i;
			x>>=1;y>>=1;
		}
	}
	
	/**************************着手成分ハッシュ関数**************************/
	
	uint64 genHash_VX(int c,uint32 x){
		uint64 ret=hash_table_vx[0][x&1U];
		for(int i=1;i<16;++i){
			x>>=1;
			ret^=hash_table_vx[i][x&1U];
		}
		return ret;
	}
	uint64 genHash_VY(int c,uint32 y){
		uint64 ret=hash_table_vy[0][y&1U];
		for(int i=1;i<16;++i){
			y>>=1;
			ret^=hash_table_vy[i][y&1U];
		}
		return ret;
	}
	uint64 genHash_VRange(uint32 r){
		return hash_table_vr[r];
	}
	uint32 genHash_S(uint32 s){
		return s;
	}
	
	void subtrHash_VRange(uint64 *const dst,int org_range){
		//レンジを狭くする
		*dst^=hash_table_vr[org_range];
		*dst^=hash_table_vr[org_range-1];
	}
	
	uint64 genHash_VXY(uint32 xy){
		//下半分がx、上半分がyとする形
		uint64 ret=hash_table_vx[0][xy&1U];
		for(int i=1;i<16;++i){
			xy>>=1;
			ret^=hash_table_vx[i][xy&1U];
		}
		for(int i=0;i<16;++i){
			xy>>=1;
			ret^=hash_table_vy[i][xy&1U];
		}
		return ret;
	}
	
	/**************************着手ハッシュ関数**************************/
	
	uint64 genHash_Move(const MoveXY& mv){
		return genHash_VXY(mv.getVXY())^genHash_S(mv.getSpin())^genHash_VRange(mv.getRange());
	}
	
	void initHash(){
		//ハッシュ関連初期化
		
		const uint32 seed=0x0e4a0fe9;//ここを変えると×
		XorShift64 dice;
		
		dice.srand(seed);
		
		for(int i=0;i<16;++i){
			hash_table_bx[i][0]=dice.rand();
			hash_table_bx[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_by[i][0]=dice.rand();
			hash_table_by[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_wx[i][0]=dice.rand();
			hash_table_wx[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_wy[i][0]=dice.rand();
			hash_table_wy[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_vx[i][0]=dice.rand();
			hash_table_vx[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_vy[i][0]=dice.rand();
			hash_table_vy[i][1]=dice.rand();
		}
		for(int i=0;i<16;++i){
			hash_table_vr[i]=dice.rand();
		}
		
	}
	
}

