//デジタルカーリング
//基本的な定数とデータ構造

namespace DigitalCurling{
	
	/**************************先攻後攻**************************/
	
	//エンド内の先攻後攻
	enum{
		BLACK=0,
		WHITE=1,
	};
	
	int flipColor(int c){return BLACK+WHITE - c;}
	bool examColor(int c){return (c==BLACK || c==WHITE);}
	
	const char *str_color[2]={"BLACK","WHITE"};
	
	struct OutColor{
		int c;
		constexpr OutColor(int ac)
		:c(ac){}
	};
	
	ostream& operator<<(ostream& out,const OutColor& arg){
		out<<str_color[arg.c];return out;
	}
	
	/**************************エンド**************************/
	
	//最終エンドを0とする減数型
	enum{
		END_LAST=0,//最終エンド
		MAX_NENDS=10,//延長除く最大エンド数
		END_JOBAN=5,//このエンドまでを前半と捉える
	};
	
	/**************************エンド内のターン**************************/
	
	//先攻後攻含めて 15~0で表現
	constexpr int TURN_LAST=0;
	constexpr int TURN_FIRST=15;
	constexpr int NTURNS=16;
	constexpr int TURN_FREEGUARD=TURN_FIRST-4+1;//このターンまでFGZルール適用
	
	bool examTurn(int t){return (TURN_LAST<=t && t<=TURN_FIRST);}
	
	int TURNtoCOL(int t){return 1-(t%2);}
	
	bool isFreeGuardTurn(int t){//フリーガードルール適用ターン
		return (t >= TURN_FREEGUARD);
	}
	
	int getEndRemStones(int c,int t){//cがこのエンド中投げられる石数
		return (t+c+1)/2;
	}
	int getGameRemStones(int c,int e,int t){//cがこの試合中投げられる石数
		return e*(NTURNS/2) + getEndRemStones(c,t);
	}
	
	
	/**************************制限時間**************************/
	
	constexpr uint64 TIME_LIMIT_UNLIMITED=0xffffffffffffffff;
	
	struct TimeLimit{
		uint64 limit_ms;
		
		void set(uint64 alim){
			limit_ms=alim;
		}
		void subtr(uint64 aused){
			limit_ms-=aused;
		}
		void add(uint64 aadded){
			limit_ms+=aadded;
		}
		bool isUnlimited()const{
			//時間無制限
			return ( limit_ms == TIME_LIMIT_UNLIMITED );
		}
	};
	
	/**************************石の回転**************************/
	
	//現在、回転は右回り、左回りの2値しかない
	enum Spin{
		RIGHT=0,
		LEFT=1,
	};
	
	/**************************連続座標系でのルール(公式)**************************/
	
	//公式が使用している座標系そのまま
	//現実のカーリングでの1メートルが1に相当
	//諸々の判定関数はこの系に対応していないので、座標変換してから使用すること
	
	constexpr fpn_t FX_OFFICIAL_TEE=2.375f;
	constexpr fpn_t FY_OFFICIAL_TEE=4.88f;
	
	
	/**************************連続座標系でのルール(自分)**************************/
	
	//自分の思考用に座標変換したもの
	
	//石の半径
	constexpr fpn_t F_STONE_RAD=0.145f;
	//ハウスの半径
	constexpr fpn_t F_HOUSE_RAD=1.83f;
	
	//ティー(ハウス中心)の座標
	constexpr fpn_t FX_TEE=0;
	constexpr fpn_t FY_TEE=0;
	
	//プレーエリアの縦横
	constexpr fpn_t F_PA_WIDTH=4.75f;
	constexpr fpn_t F_PA_HEIGHT=8.23f;
	
	//プレーエリアの座標
	constexpr fpn_t FX_PA_MAX=FX_TEE + F_PA_WIDTH/2;
	constexpr fpn_t FY_PA_MAX=FY_TEE + F_HOUSE_RAD;
	constexpr fpn_t FX_PA_MIN=FX_PA_MAX - F_PA_WIDTH;
	constexpr fpn_t FY_PA_MIN=FY_PA_MAX - F_PA_HEIGHT;
	
	//線座標
	constexpr fpn_t FY_HOGLINE=FY_PA_MIN;
	
	//スロー位置
	constexpr fpn_t FX_THROW=FX_TEE;
	constexpr fpn_t FY_THROW=FY_HOGLINE - 30.0f;
	
	//リンクの縦横
	constexpr fpn_t F_RINK_WIDTH=F_PA_WIDTH;
	//constexpr fpn_t F_RINK_HEIGHT=F_PA_HEIGHT;
	
	//リンクの端の座標(公式のものと定義が異なるので注意)
	constexpr fpn_t FX_RINK_MIN=FX_PA_MIN;
	constexpr fpn_t FY_RINK_MIN=FY_THROW - F_STONE_RAD*2;//後ろに投げる事は無いだろう
	constexpr fpn_t FX_RINK_MAX=FX_PA_MAX;
	constexpr fpn_t FY_RINK_MAX=FY_PA_MAX + F_STONE_RAD*2;

	
	//ハウス中心の極座標系
	constexpr fpn_t FR_TEE=0;
	
	//石とリンクの動摩擦
	constexpr fpn_t F_FRIC_STONE=12.009216f;//(=g*mu)
	constexpr fpn_t F_STD_GRAV=9.80665;//重力加速度g
	constexpr fpn_t F_DCOF=F_FRIC_STONE/F_STD_GRAV;//動摩擦係数mu
	
	//石同士の反発係数
	constexpr fpn_t F_REST_STONE=1.0f;
	
	//石の密度(box2dでは0にすると固定物体になる)
	constexpr fpn_t F_DENS_STONE=10.0f;
	
	//石の到達位置のX,Y座標にかかる正規乱数の標準偏差
	constexpr fpn_t F_RAND_SIGMA=0.145f;
	
	//石の初期回転角速度
	constexpr fpn_t FANGV_ORG=0.066696f;
	
	//リンク内
	template<typename float_t>
	bool isOnRink(float_t x,float_t y){
		return ( static_cast<float_t>(FX_RINK_MIN + F_STONE_RAD) < x 
				&& x < static_cast<float_t>(FX_RINK_MAX - F_STONE_RAD) )
		&& ( static_cast<float_t>(FY_RINK_MIN + F_STONE_RAD) < y 
			&& y < static_cast<float_t>(FY_RINK_MAX - F_STONE_RAD) );
	}
	//プレーエリア(石が無効にならない範囲)内
	template<typename float_t>
	bool isInPlayArea(float_t x,float_t y){
		return ( static_cast<float_t>(FX_PA_MIN + F_STONE_RAD) < x 
				&& x < static_cast<float_t>(FX_PA_MAX - F_STONE_RAD) )
		&& ( static_cast<float_t>(FY_RINK_MIN + F_STONE_RAD) < y 
			&& y < static_cast<float_t>(FY_PA_MAX + F_STONE_RAD) );
	}
	//ハウス内
	template<typename float_t>
	bool isInHouse(float_t x,float_t y){
		return ( XYtoR2(x-FX_TEE,y-FY_TEE) < static_cast<float_t>(( F_HOUSE_RAD + F_STONE_RAD )*( F_HOUSE_RAD + F_STONE_RAD )) );
	}
	
	//ティーラインの前
	template<typename float_t>
	bool isThisSideOfTeeLine(float_t y){
		assert( isInPlayArea(FX_TEE,y) );//プレーエリア外のときは考えないだろう
		return ( y < static_cast<float_t>(FY_TEE - F_STONE_RAD) );
	}
	
	//ティーの近傍
	template<typename float_t>
	bool isNearTee(float_t x,float_t y,float_t dist){
		return ( XYtoR2(x-FX_TEE,y-FY_TEE) < dist*dist );
	}
	
	//フリーガードゾーン内
	template<typename float_t,int NOT_IN_HOUSE_CHECKED=0>
	bool isInFreeGuardZone(float_t x,float_t y){
		assert( isInPlayArea(x,y) );
		if( !NOT_IN_HOUSE_CHECKED ){
			if( isInHouse(x,y) ){return false;}
		}
		return isThisSideOfTeeLine(y);
	}
	
	/**************************連続座標系での公式と自分用の変換**************************/
	
	//座標変換
	template<class fpos_t>
	fpos_t convPos_Official_Mine(const fpos_t& arg){
		return fpos_t( (arg.getX()-FX_OFFICIAL_TEE) + FX_TEE, -(arg.getY()-FY_OFFICIAL_TEE) + FY_TEE );	
	}
	
	template<class fpos_t>
	fpos_t convPos_Mine_Official(const fpos_t& arg){
		return fpos_t( (arg.getX()-FX_TEE) + FX_OFFICIAL_TEE, -(arg.getY()-FY_TEE) + FY_OFFICIAL_TEE );	
	}
	
	template<class fmove_t>
	fmove_t convMove_Official_Mine(const fmove_t& arg){
		return fmove_t( arg.getVX(), -arg.getVY(), arg.getSpin() );	
	}
	
	template<class fmove_t>
	fmove_t convMove_Mine_Official(const fmove_t& arg){
		return fmove_t( arg.getVX(), -arg.getVY(), arg.getSpin() );
	}
	
	/**************************離散座標**************************/
	
	constexpr int RESOLUTION_X=15;
	constexpr int RESOLUTION_Y=15;
	
	constexpr uint32 X_MIN=0;
	constexpr uint32 Y_MIN=0;
	constexpr uint32 X_MAX=(1U<<RESOLUTION_X)-1U;
	constexpr uint32 Y_MAX=(1U<<RESOLUTION_Y)-1U;
	constexpr uint32 X_MID=1U<<(RESOLUTION_X-1);
	constexpr uint32 Y_MID=1U<<(RESOLUTION_Y-1);
	
	
	/*************************位置の連続座標と離散座標の変換**************************/
	
	//座標変換
	
	constexpr fpn_t F_STEP=( FX_PA_MAX - FX_PA_MIN )/(double)(1<<RESOLUTION_X);
	constexpr fpn_t F_INVSTEP=(1.0/F_STEP);
	//円形判定のため、STEPはx,y同じ幅である必要がある
	
	constexpr fpn_t FX_XMID=( FX_PA_MIN + FX_PA_MAX )/2.0;
	constexpr fpn_t FY_YMID=( FY_PA_MIN+F_STONE_RAD + FY_PA_MAX+F_STONE_RAD )/2.0;
	
	constexpr fpn_t FX_X0=FX_XMID - F_STEP * (double)X_MID;
	constexpr fpn_t FY_Y0=FY_YMID - F_STEP * (double)Y_MID;
	
	constexpr fpn_t ItoF(uint32 i){
		return (fpn_t)i * F_STEP;
	}
	constexpr fpn_t XtoFX(uint32 x){
		return (fpn_t)x * F_STEP +FX_X0;
	}
	constexpr fpn_t YtoFY(uint32 y){
		return (fpn_t)y * F_STEP +FY_Y0;
	}
	
	constexpr uint32 FtoI(fpn_t f){
		return (uint32)(f*F_INVSTEP);
	}
	constexpr uint32 FXtoX(fpn_t fx){
		return X_MIN + (uint32)((fx-FX_X0)*F_INVSTEP);
	}
	constexpr uint32 FYtoY(fpn_t fy){
		return Y_MIN + (uint32)((fy-FY_Y0)*F_INVSTEP);
	}
	
	constexpr uint32 F2toI2(float f2){
		return (uint32)(f2*static_cast<float>(F_INVSTEP*F_INVSTEP));
	}
	constexpr uint32 F2toI2(double f2){
		return (uint32)(f2*static_cast<double>(F_INVSTEP*F_INVSTEP));
	}
	
	/*************************離散座標でのルール**************************/
	
	constexpr uint32 X_TEE=FXtoX(FX_TEE);
	constexpr uint32 Y_TEE=FYtoY(FY_TEE);
	
	//リンク内
	bool isOnRink(uint32 x,uint32 y){
		return ( FXtoX(FX_RINK_MIN + F_STONE_RAD) < x && x < FXtoX(FX_RINK_MAX - F_STONE_RAD) )
		&& ( FYtoY(FY_RINK_MIN + F_STONE_RAD) < y && y < FYtoY(FY_RINK_MAX - F_STONE_RAD) );
	}
	
	//プレーエリア(石が無効にならない範囲)内
	bool isInPlayArea(uint32 x,uint32 y){
		return ( FXtoX(FX_PA_MIN + F_STONE_RAD) < x && x < FXtoX(FX_PA_MAX - F_STONE_RAD) )
		&& ( FYtoY(FY_PA_MIN + F_STONE_RAD) < y && y < FYtoY(FY_PA_MAX + F_STONE_RAD) );
	}
	
	//ハウス内
	bool isInHouse(uint32 x,uint32 y){
		return ( XYtoR2(x-X_TEE,y-Y_TEE) < F2toI2( (F_HOUSE_RAD + F_STONE_RAD)*(F_HOUSE_RAD + F_STONE_RAD) ) );
	}
	
	//ティーラインの前
	bool isThisSideOfTeeLine(uint32 y){
		assert( isInPlayArea(FXtoX(FX_TEE),y) );//プレーエリア外のときは考えないだろう
		return ( y < FYtoY(FY_TEE - F_STONE_RAD) );
	}
	
	//フリーガードゾーン内
	template<int NOT_IN_HOUSE_CHECKED=0>
	bool isInFreeGuardZone(uint32 x,uint32 y){
		assert( isInPlayArea(x,y) );
		if( !NOT_IN_HOUSE_CHECKED ){
			if( isInHouse(x,y) ){return false;}
		}
		return isThisSideOfTeeLine(y);
	}
	
	
	/**************************連続座標系での盤面情報**************************/
	
	template<typename float_t=fpn_t>
	struct fPosXY{
		//ストーンの位置を表す連続型
		float_t x;
		float_t y;
		
		constexpr fPosXY():
		x(),y(){}
		
		constexpr fPosXY(float_t ax,float_t ay)
		:x(ax),y(ay){}
		
		void set(float_t ax,float_t ay){
			x=ax;y=ay;
		}
		
		void setX(float_t ax){x=ax;}
		void setY(float_t ay){y=ay;}
		
		float_t getX()const{return x;}
		float_t getY()const{return y;}
	};
	
	template<typename float_t>
	ostream& operator<<(ostream& out,const fPosXY<float_t>& arg){
		out<<"("<<arg.getX()<<","<<arg.getY()<<")";
		return out;
	}
	
	template<typename float_t=fpn_t>
	struct fPosXYRT{
		//ストーンの位置を表す連続型
		float_t x;
		float_t y;
		float_t r;
		float_t t;
		
		constexpr fPosXYRT():
		x(),y(),r(),t(){}
		
		constexpr fPosXYRT(float_t ax,float_t ay)
		:x(ax),y(ay),r(XYtoR(ax-FX_TEE,ay-FY_TEE)),t(XYtoT(ax-FX_TEE,ay-FY_TEE)){}
		
		void set(float_t ax,float_t ay){
			x=ax;y=ay;
			r=XYtoR(ax-FX_TEE,ay-FY_TEE);
			t=XYtoT(ax-FX_TEE,ay-FY_TEE);
		}
		
		float_t getX()const{return x;}
		float_t getY()const{return y;}
		float_t getR()const{return r;}
		float_t getT()const{return t;}
	};
	
	bool isOnBoard(float x,float y){
		return (x || y);
	}
	/*
	 struct fPoss{
	 //全てのストーンの位置(+情報)を表す型
	 fPosXY<> pos[2][8];
	 int nstones[2];
	 void init(){
	 nstones[0]=nstones[1]=0;
	 memset(pos,0,2*8*sizeof(fPosXY));
	 }
	 void addStone(int col,const fPosXY& apos){
	 pos[col][nstones[col]]=apos;
	 nstones[col]++;
	 assert( exam() );
	 }
	 
	 bool exam(){
	 //石数の確認
	 //石同士が被っていないことの確認
	 return true;
	 }
	 };*/
	
	template<typename float_t>
	ostream& operator<<(ostream& out,const fPosXYRT<float_t>& arg){
		out<<"("<<arg.getX()<<","<<arg.getY()<<")";
		return out;
	}
	
	/*************************離散座標系での盤面情報**************************/
	
	constexpr uint32 STONE_MASK_X=(1U<<16)-1U;
	constexpr uint64 STONE_MASK_COLOR=1ULL<<32;
	
	struct StoneXY{
		//離散ストーン位置(直交座標)+追加情報
		uint64 st;
		
		constexpr StoneXY():st(){}
		constexpr StoneXY(uint64 ast):st(ast){}
		constexpr StoneXY(const StoneXY& ast):st(ast.st){}
		
		constexpr operator uint64()const{return st;}
		
		//成分取り出し
		uint32 getX()const{return ((uint32)st) & STONE_MASK_X;}
		uint32 getY()const{return (((uint32)st)>>16) & STONE_MASK_X;}
		uint32 getXY()const{return (uint32)st;}
		uint32 getCol()const{return (uint32)(st>>32);}
		
		//性質
		uint64 isWhite()const{return st & STONE_MASK_COLOR;}	
		bool isBlack()const{return !(st & STONE_MASK_COLOR);}
		
		//分割
		/*template<int N>StoneXY div(int n);//N分割のn番目を計算する
		 
		 template<>Move div<4>(int n){//4分割
		 assert( 0<=n && n<4 );
		 assert( getRange()>0 );
		 
		 uint32 r=1U<<getRange();
		 uint32 half_r=r>>1;//範囲の半分
		 
		 uint64 tmp=mv;
		 switch(n){
		 case 0:tmp+=half_r|(half_r<<16);
		 case 1:tmp-=(half_r<<16)-half_r;
		 case 2:tmp+=(half_r<<16)-half_r;
		 case 3:tmp-=half_r|(half_r<<16);
		 default:assert(0);
		 }
		 tmp-=(1ULL<<32);//rangeを1減らす
		 
		 return Move(tmp);
		 }*/
	};
	
	struct StoneRT{
		//離散ストーン位置(極座標)
		uint64 st;
	};
	
	
	/**************************位置座標の基本アルゴリズム**************************/
	
	template<class pos_t>
	bool isInPlayArea(const pos_t& pos){
		return isInPlayArea( pos.getX(), pos.getY() );
	}
	
	template<class pos_t>
	bool isInHouse(const pos_t& pos){
		return isInHouse( pos.getX(), pos.getY() );
	}
	
	template<>
	bool isInHouse<float>(const float& r){
		return ( r < static_cast<float>(F_HOUSE_RAD + F_STONE_RAD) );
	}
	template<>
	bool isInHouse<double>(const double& r){
		return ( r < static_cast<double>(F_HOUSE_RAD + F_STONE_RAD) );
	}
	
	template<>
	bool isInHouse(const fPosXYRT<float>& pos){
		return isInHouse<float>((float)pos.getR());//すでにRが計算されているので
	}
	template<>
	bool isInHouse(const fPosXYRT<double>& pos){
		return isInHouse<double>((double)pos.getR());//すでにRが計算されているので
	}
	
	template<class pos_t>
	bool isNearTee(const pos_t& pos,float dist){
		return isNearTee(pos.getX(),pos.getY(),dist);
	}
	
	template<>
	bool isNearTee(const float& r,float dist){
		return ( r < dist);
	}
	
	template<>
	bool isNearTee(const double& r,float dist){
		return ( r < dist );
	}
	
	template<>
	bool isNearTee(const fPosXYRT<float>& pos,float dist){
		return isNearTee(pos.getR(),dist);//すでにRが計算されているので
	}
	
	template<>
	bool isNearTee(const fPosXYRT<double>& pos,float dist){
		return isNearTee(pos.getR(),dist);//すでにRが計算されているので
	}
	
	template<class pos_t,int NOT_IN_HOUSE_CHECKED=0>
	bool isInFreeGuardZone(const pos_t& pos){
		return isInFreeGuardZone<NOT_IN_HOUSE_CHECKED>(pos.getX(),pos.getY());
	}
	
	template<int NOT_IN_HOUSE_CHECKED=0>
	bool isInFreeGuardZone(const fPosXYRT<float>& pos){
		if( !NOT_IN_HOUSE_CHECKED ){
			if( isInHouse<float>(pos.getR()) ){return false;}//すでにRが計算されているので
		}
		return isThisSideOfTeeLine<float>(pos.getY());
	}
	
	template<int NOT_IN_HOUSE_CHECKED=0>
	bool isInFreeGuardZone(const fPosXYRT<double>& pos){
		if( !NOT_IN_HOUSE_CHECKED ){
			if( isInHouse<double>(pos.getR()) ){return false;}//すでにRが計算されているので
		}
		return isThisSideOfTeeLine<double>(pos.getY());
	}
	
	//2つの石がめり込んでいる
	template<class pos_t>
	bool isCavingIn(const pos_t& p0,const pos_t& p1){
		return ( XYtoR2(p0.x-p1.x,p0.y-p1.y) < 4.0 * F_STONE_RAD * F_STONE_RAD );
	}
	
	
	/**************************連続座標系での速度ベクトルによる着手表現**************************/
	
	//着手を存在させる範囲
	constexpr fpn_t FVX_MIN=-3.1;
	constexpr fpn_t FVX_MAX=+3.1;
	constexpr fpn_t FVY_MIN=+27.4;
	constexpr fpn_t FVY_MAX=+33.6;
	
	constexpr fpn_t FVX_MID=(FVX_MIN+FVX_MAX)/2;
	constexpr fpn_t FVY_MID=(FVY_MIN+FVY_MAX)/2;
	
	constexpr fpn_t FV_RANGE_MAX=(FVX_MAX-FVX_MIN)/2;
	
	//速さの範囲(NullMoveは別)
	constexpr fpn_t FV_MIN=FVY_MIN;
	constexpr fpn_t FV_MAX=XYtoR(FVX_MAX,FVY_MAX);
	
	template<typename float_t=fpn_t>
	struct fMoveXY{
		//ノーレンジ着手
		//速度ベクトルのxy成分と回転
		float_t x;
		float_t y;
		int s;
		
		float_t getVX()const{return x;}
		float_t getVY()const{return y;}
		int getSpin()const{return s;}
		bool isLeftSpin()const{return (s!=Spin::RIGHT);}
		
		float_t calcV()const{return XYtoR(x,y);}//速さ
		float_t calcT()const{return XYtoT(x,y);}//角度
		float_t calcCosT()const{return XYtoCosT(x,y);}
		float_t calcSinT()const{return XYtoSinT(x,y);}
		
		void set(float_t ax,float_t ay,int as){
			x=ax;y=ay;s=as;
		}
		void setVX(float_t ax){x=ax;}
		void setVY(float_t ay){y=ay;}
		void setSpin(int as){s=as;}
		
		void addVX(float_t ax){x+=ax;}
		void addVY(float_t ay){y+=ay;}
		
		constexpr fMoveXY():x(),y(),s(){}
		constexpr fMoveXY(float_t ax,float_t ay,int as)
		:x(ax),y(ay),s(as){}
	};
	
	template<typename float_t>
	std::ostream& operator<<(std::ostream& out,const fMoveXY<float_t>& arg){
		out<<"fMV < vx:"<<arg.getVX()<<" vy:"<<arg.getVY()<<" s:"<<arg.getSpin()<<" >";
		return out;
	}
	
	template<typename float_t=fpn_t>
	struct fRMoveXY{
		//速度ベクトルのxy成分と回転、その範囲
		float_t x;
		float_t y;
		int s;
		float_t rx;
		float_t ry;
		
		
		float_t getVX()const{return x;}
		float_t getVY()const{return y;}
		int getSpin()const{return s;}
		float_t getRangeVX()const{return rx;}
		float_t getRangeVY()const{return ry;}
		bool isLeftSpin()const{return (s!=Spin::RIGHT);}
		
		float_t calcV()const{return XYtoR(x,y);}//速さ
		float_t calcT()const{return XYtoT(x,y);}//角度
		float_t calcCosT()const{return XYtoCosT(x,y);}
		float_t calcSinT()const{return XYtoSinT(x,y);}
		
		void setVX(float_t ax){x=ax;}
		void setVY(float_t ay){y=ay;}
		void setSpin(int as){s=as;}
		void setRangeVX(float_t arx){rx=arx;}
		void setRangeVY(float_t ary){ry=ary;}
		
		void addVX(float_t ax){x+=ax;}
		void addVY(float_t ay){y+=ay;}
		
		void setWR(int as){//全範囲着手
			x=FVX_MID;y=FVY_MID;
			s=as;
			rx=FV_RANGE_MAX;ry=FV_RANGE_MAX;
		}
		
		//分割
		fRMoveXY<float_t> part4(int n){
			//4分割のn番目を計算する
			assert( 0<=n && n<4 );
			
			fMoveXY<float_t> ret;
			ret.s=s;
			ret.rx=rx/2;ret.ry=ry/2;
			switch(n){
				case 0:ret.x=x-ret.rx;ret.y=y-ret.ry;break;
				case 1:ret.x=x-ret.rx;ret.y=y+ret.ry;break;
				case 2:ret.x=x+ret.rx;ret.y=y-ret.ry;break;
				case 3:ret.x=x+ret.rx;ret.y=y+ret.ry;break;
				default:assert(0);break;
			}
			return ret;
		}
		
		template<class ddice_t>
		fMoveXY<float_t> genNR(ddice_t *const dice)const{
			//乱数を掛けて、範囲0(no range)の着手を1つ返す
			fMoveXY<float_t> ret;
			ret.s=s;
			ret.x=x+dice->drand() * rx;
			ret.y=y+dice->drand() * ry;
			return ret;
		}
	};
	
	template<typename float_t>
	std::ostream& operator<<(std::ostream& out,const fRMoveXY<float_t>& arg){
		out<<"fMV < vx:"<<arg.getVX()<<" vy:"<<arg.getVY()<<" vr:"<<arg.getRangeVX()<<" s:"<<arg.getSpin()<<" >";
		return out;
	}
	
	//目的座標から着手型への変換
	
	//const fMove FMOVE_FREE_DRAW={}
	
	/**************************離散座標系での着手表現**************************/
	
	constexpr int RESOLUTION_VX=15;
	constexpr int RESOLUTION_VY=15;
	constexpr int RESOLUTION_VRANGE=15;
	constexpr int RESOLUTION_SPIN=1;
	
	constexpr uint32 VX_MIN=0;
	constexpr uint32 VY_MIN=0;
	constexpr uint32 VX_MAX=(1U<<RESOLUTION_VX)-1U;
	constexpr uint32 VY_MAX=(1U<<RESOLUTION_VY)-1U;
	constexpr uint32 VX_MID=1U<<(RESOLUTION_VX-1);
	constexpr uint32 VY_MID=1U<<(RESOLUTION_VY-1);
	
	//constexpr uint32 V_RANGE_MAX=(1U<<RESOLUTION_VRANGE);
	
	constexpr uint32 MOVE_MASK_X=(1U<<16)-1U;
	constexpr uint32 MOVE_MASK_Y=MOVE_MASK_X<<16;
	constexpr uint64 MOVE_MASK_RANGE=((1ULL<<16)-1ULL)<<48;
	constexpr uint64 MOVE_MASK_SPIN=1ULL<<32;
	
	
	struct MoveXY{
		//レンジなし、あり両用の離散着手型(直交座標)
		uint64 mv;
		
		constexpr MoveXY():mv(){}
		constexpr MoveXY(uint64 amv):mv(amv){}
		constexpr MoveXY(const MoveXY& amv):mv(amv.mv){}
		
		constexpr operator uint64()const{return mv;}
		
		void setNR(uint32 ax,uint32 ay,uint64 s){//ノーレンジ着手をセット
			mv= ax | (ay<<16) | (s<<32);
		}
		
		//成分取り出し
		uint32 getVX()const{return ((uint32)mv) & MOVE_MASK_X;}
		uint32 getVY()const{return (((uint32)mv)>>16) & MOVE_MASK_X;}
		uint32 getVXY()const{return (uint32)mv;}
		uint32 getSpin()const{return ((uint32)(mv>>32))&1U;}
		uint32 getRange()const{return (uint32)(mv>>48);}
		
		uint64 anyRange()const{return mv & MOVE_MASK_RANGE;}
		
		void setWR(uint32 as){//全範囲着手
			mv=VX_MID | (VY_MID<<16) | (((uint64)as)<<32) | ((uint64)(RESOLUTION_VRANGE)<<48);
		}
		
		void setNullMove(){//ヌルムーブ
			setNR(VX_MIN,VY_MAX,1);
		}
		
		//分割
		MoveXY part(int N,int n)const{
			ASSERT( getRange()>0, cerr<<getRange()<<endl; );
			assert( 0<=n && n<N );
			uint64 tmp=mv;
			
			switch(N){
				case 4:
				{
					uint32 r=1U<<(getRange()-1U);
					uint32 half_r=r>>1;//範囲の半分
					
					switch(n){
						case 0:tmp+=half_r|(half_r<<16);break;
						case 1:tmp-=(half_r<<16)-half_r;break;
						case 2:tmp+=(half_r<<16)-half_r;break;
						case 3:tmp-=half_r|(half_r<<16);break;
						default:assert(0);break;
					}
					tmp-=(1ULL<<48);//rangeを1減らす
					break;
				}
				default:assert(0);break;
			}
			return MoveXY(tmp);
		}
		
		template<class dice_t>
		MoveXY genNR(dice_t *const dice)const{
			//乱数を掛けて、範囲0(no range)の着手を1つ返す
			MoveXY ret( mv & (~MOVE_MASK_RANGE) );
			uint32 range=(1U<<(getRange()))>>1;
			uint32 r=dice->rand();
			
			uint32 josu=(range<<1)+1U;
			
			ret.mv-=range+(range<<16);
			ret.mv+=(r&MOVE_MASK_X)%josu;
			ret.mv+=((r>>16)%josu)<<16;
			
			return ret;
		}
	};
	
	std::ostream& operator<<(std::ostream& out,const MoveXY& arg){
		out<<"MV< vx:"<<arg.getVX()<<" vy:"<<arg.getVY()<<" vr:"<<arg.getRange()<<" s:"<<arg.getSpin()<<" >";
		return out;
	}
	
	/**************************着手の連続座標と離散座標の変換**************************/
	
	
	constexpr fpn_t FV_STEP=( FV_RANGE_MAX * 2 )/(double)(VX_MAX-VX_MIN+1);
	constexpr fpn_t FV_INVSTEP=(1.0/FV_STEP);
	//速度演算のため、STEPはx,y同じ幅である必要がある
	
	constexpr fpn_t VtoFV(uint32 i){
		return ((fpn_t)i) * FV_STEP;
	}
	constexpr fpn_t VXtoFVX(uint32 x){
		return ((fpn_t)x) * FV_STEP +FVX_MIN;
	}
	constexpr fpn_t VYtoFVY(uint32 y){
		return ((fpn_t)y) * FV_STEP +FVY_MIN;
	}
	
	template<typename float_t>
	constexpr uint32 FVtoV(float_t f){
		return (uint32)(f*FV_INVSTEP);
	}
	template<typename float_t>
	constexpr uint32 FVXtoVX(float_t fx){
		return VX_MIN + (uint32)((fx-FVX_MIN)*FV_INVSTEP);
	}
	template<typename float_t>
	constexpr uint32 FVYtoVY(float_t fy){
		return VY_MIN + (uint32)((fy-FVY_MIN)*FV_INVSTEP);
	}
	template<typename float_t>
	constexpr uint32 FV2toV2(float_t f2){
		return (uint32)(f2*static_cast<float_t>(FV_INVSTEP*FV_INVSTEP));
	}
	
	template<class movexy_t>
	void convMove(const MoveXY& arg,movexy_t *const dst){
		dst->x=VXtoFVX(arg.getVX());
		dst->y=VYtoFVY(arg.getVY());
		dst->s=arg.getSpin();
	}
	
	template<typename movexy_t>
	void convMove(const movexy_t& arg,MoveXY *const dst){
		dst->mv=(uint64)( FVXtoVX(arg.getVX() ) | (FVYtoVY(arg.getVY())<<16) )
		| (((uint64)arg.getSpin())<<32);
	}
	
	template<typename float_t>
	void convRMove(const MoveXY& arg,fRMoveXY<float_t> *const dst){
		dst->x=VXtoFVX(arg.getVX());
		dst->y=VYtoFVY(arg.getVY());
		dst->rx=VtoFV((1U<<arg.getRange())>>1);
		dst->ry=VtoFV((1U<<arg.getRange())>>1);
		dst->s=arg.getSpin();
	}
	
	template<typename float_t>
	void convRMove(const fRMoveXY<float_t>& arg,MoveXY *const dst){
		dst->mv=(uint64)( FVXtoVX(arg.getVX() ) | (FVYtoVY(arg.getVY())<<16) )
		| ((uint64(bsr64((uint64)FVtoV(arg.getRangeVX()))))<<48) | (((uint64)arg.getSpin())<<32);
	}
	
	/**************************石の瞬間パラメータ**************************/
	
	template<typename float_t=fpn_t>
	struct fMStone{
		//動いている石のパラメータ
		float_t x;
		float_t y;
		float_t vx;
		float_t vy;
		float_t w;
		
		void set(float_t ax,float_t ay,float_t avx,float_t avy,float_t aw){
			x=ax;y=ay;vx=avx;vy=avy;w=aw;
		}
		
		void setX(float_t ax){x=ax;}
		void setY(float_t ay){y=ay;}
		void setVX(float_t avx){vx=avx;}
		void setVY(float_t avy){vy=avy;}
		void setW(float_t aw){w=aw;}
		
		float_t getX()const{return x;}
		float_t getY()const{return y;}
		float_t getVX()const{return vx;}
		float_t getVY()const{return vy;}
		float_t getW()const{return w;}
		
		constexpr fMStone():
		x(),y(),vx(),vy(),w(){}
		
		constexpr fMStone(float_t ax,float_t ay,float_t avx,float_t avy,float_t aw):
		x(ax),y(ay),vx(avx),vy(avy),w(aw){}
	};
		
	
	
	/**************************石の状態**************************/
	
	struct StoneState : public BitSet32{
		typedef uint32 data_t;
		
		//色
		void setColor(uint32 col){BitSet32::dat|=col;}
		void setBlack()const{}
		void setWhite(){set(0);}
		data_t isWhite()const{return test(0);}
		data_t getColor()const{return get(0);}
		
		//フリーガード
		void setFreeGuard(){set(3);}
		void resetFreeGuard(){reset(3);}
		data_t isFreeGuard()const{return test(3);}
		//フリーガード予備
		void setInFreeGuardZone(){set(2);}
		void resetInFreeGuardZone(){reset(2);}
		data_t isInFreeGuardZone()const{return test(2);}
		
		void init(){reset();}
	};
	
	/**************************primitive型の基本値初期化**************************/
	
	void initPrim(){
		
#ifdef MONITOR
		CERR<<"V->FV STEP : "<<FV_STEP<<endl;
#endif
	}
	
}